/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* Testing solving time-harmonic Maxwell equations with an analytic solution.
 *
 *	curl(1/mu curl) E - k^2 E = J
 *
 * $Id: maxwell.c,v 1.85 2007/04/02 14:42:58 zlb Exp $
 */

#include "phg.h"
#include <math.h>
#include <stdlib.h>

static FLOAT kk = -1.0;		/* kk = k^2 */

static double
elapsed_time(GRID *g, BOOLEAN flag, double mflops)
/* returns and prints elapsed time since last call to this function */
{
    static double t0 = 0.0;
    double et, tt[3];
    size_t mem;
 
    phgGetTime(tt);
    et = tt[2] - t0;
    t0 = tt[2];

    mem = phgMemoryUsage(g, NULL);

    if (flag) {
	if (mflops <= 0)
	    phgPrintf("[%0.4lgMB %0.4lfs]\n", mem / (1024.0 * 1024.0), et);
	else
	    phgPrintf("[%0.4lgMB %0.4lfs %0.4lgGF]\n", mem / (1024.0 * 1024.0),
			 et, mflops*1e-3);
    }

    return et;
}

static void
func_u(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
{
    *(value++) = x * y * z * (x - 1.0) * (y - 1.0) * (z - 1.0) *
		(x - .5) * (y - .5) * (z - .5);
    *(value++) = Sin(2.0 * M_PI * x) * Sin(2.0 * M_PI * y) *
		Sin(2.0 * M_PI * z);
    *(value++) = (1.0 - Exp(x)) * (M_E - Exp(x)) * (M_E - Exp(2.0 * x)) *
		(1.0 - Exp(y)) * (M_E - Exp(y)) * (M_E - Exp(2.0 * y)) *
		(1.0 - Exp(z)) * (M_E - Exp(z)) * (M_E - Exp(2.0 * z));
    return;
}

static void
func_f(FLOAT x, FLOAT y, FLOAT z, FLOAT *values)
{
    FLOAT u[3];

    func_u(x, y, z, u);

    *(values++) = (-1. - kk) * u[0] +
	4.0 * Cos(2.0 * M_PI * x) * M_PI * M_PI * Cos(2.0 * M_PI * y) *
		Sin(2.0 * M_PI * z) -
	2.0 * x * z * (x - 1.0) * (z - 1.0) * (x - 0.5) *
		(y - 0.5) * (z - 0.5) - 2.0 * x * z * (x - 1.0) * (y - 1.0) *
		(z - 1.0) * (x - 0.5) * (z - 0.5) - 2.0 * x * y * z *
		(x - 1.0) * (z - 1.0) * (x - 0.5) * (z - 0.5) -
	2.0 * x * y * (x - 1.0) * (y - 1.0) * (x - 0.5) * (y - 0.5) *
		(z - 0.5) - 2.0 * x * y * (x - 1.0) * (y - 1.0) * (z - 1.0) *
		(x - 0.5) * (y - 0.5) - 2.0 * x * y * z * (x - 1.0) *
		(y - 1.0) * (x - 0.5) * (y - 0.5) +
	Exp(x) * (M_E - Exp(x)) * (M_E - Exp(2.0 * x)) * (1.0 - Exp(y)) *
		(M_E - Exp(y)) * (M_E - Exp(2.0 * y)) * Exp(z) *
		(M_E - Exp(z)) * (M_E - Exp(2.0 * z)) +
	Exp(x) * (M_E - Exp(x)) * (M_E - Exp(2.0 * x)) * (1.0 - Exp(y)) *
		(M_E - Exp(y)) * (M_E - Exp(2.0 * y)) * (1.0 - Exp(z)) *
		Exp(z) * (M_E - Exp(2.0 * z)) +
	2.0 * Exp(x) * (M_E - Exp(x)) * (M_E - Exp(2.0 * x)) * (1.0 - Exp(y)) *
		(M_E - Exp(y)) * (M_E - Exp(2.0 * y)) * (1.0 - Exp(z)) *
		(M_E - Exp(z)) * Exp(2.0 * z) +
	(1.0 - Exp(x)) * Exp(x) * (M_E - Exp(2.0 * x)) * (1.0 - Exp(y)) *
		(M_E - Exp(y)) * (M_E - Exp(2.0 * y)) * Exp(z) *
		(M_E - Exp(z)) * (M_E - Exp(2.0 * z)) +
	(1.0 - Exp(x)) * Exp(x) * (M_E - Exp(2.0 * x)) * (1.0 - Exp(y)) *
		(M_E - Exp(y)) * (M_E - Exp(2.0 * y)) * (1.0 - Exp(z)) *
		Exp(z) * (M_E - Exp(2.0 * z)) +
	2.0 * (1.0 - Exp(x)) * Exp(x) * (M_E - Exp(2.0 * x)) *
		(1.0 - Exp(y)) * (M_E - Exp(y)) * (M_E - Exp(2.0 * y)) *
		(1.0 - Exp(z)) * (M_E - Exp(z)) * Exp(2.0 * z) +
	2.0 * (1.0 - Exp(x)) * (M_E - Exp(x)) * Exp(2.0 * x) *
		(1.0 - Exp(y)) * (M_E - Exp(y)) * (M_E - Exp(2.0 * y)) *
		Exp(z) * (M_E - Exp(z)) * (M_E - Exp(2.0 * z)) +
	2.0 * (1.0 - Exp(x)) * (M_E - Exp(x)) * Exp(2.0 * x) *
		(1.0 - Exp(y)) * (M_E - Exp(y)) * (M_E - Exp(2.0 * y)) *
		(1.0 - Exp(z)) * Exp(z) * (M_E - Exp(2.0 * z)) +
	4.0 * (1.0 - Exp(x)) * (M_E - Exp(x)) * Exp(2.0 * x) *
		(1.0 - Exp(y)) * (M_E - Exp(y)) * (M_E - Exp(2.0 * y)) *
		(1.0 - Exp(z)) * (M_E - Exp(z)) * Exp(2.0 * z) +
	x * y * z * (x - 1.0) * (y - 1.0) * (z - 1.0) * (x - 0.5) *
		(y - 0.5) * (z - 0.5);

    *(values++) = (-1. - kk) * u[1] +
	(1.0 - Exp(x)) * (M_E - Exp(x)) * (M_E - Exp(2.0 * x)) * Exp(y) *
		(M_E - Exp(y)) * (M_E - Exp(2.0 * y)) * Exp(z) *
		(M_E - Exp(z)) * (M_E - Exp(2.0 * z)) +
	(1.0 - Exp(x)) * (M_E - Exp(x)) * (M_E - Exp(2.0 * x)) * Exp(y) *
		(M_E - Exp(y)) * (M_E - Exp(2.0 * y)) * (1.0 - Exp(z)) *
		Exp(z) * (M_E - Exp(2.0 * z)) +
	2.0 * (1.0 - Exp(x)) * (M_E - Exp(x)) * (M_E - Exp(2.0 * x)) *
		Exp(y) * (M_E - Exp(y)) * (M_E - Exp(2.0 * y)) *
		(1.0 - Exp(z)) * (M_E - Exp(z)) * Exp(2.0 * z) +
	(1.0 - Exp(x)) * (M_E - Exp(x)) * (M_E - Exp(2.0 * x)) *
		(1.0 - Exp(y)) * Exp(y) * (M_E - Exp(2.0 * y)) * Exp(z) *
		(M_E - Exp(z)) * (M_E - Exp(2.0 * z)) +
	(1.0 - Exp(x)) * (M_E - Exp(x)) * (M_E - Exp(2.0 * x)) *
		(1.0 - Exp(y)) * Exp(y) * (M_E - Exp(2.0 * y)) *
		(1.0 - Exp(z)) * Exp(z) * (M_E - Exp(2.0 * z)) +
	2.0 * (1.0 - Exp(x)) * (M_E - Exp(x)) * (M_E - Exp(2.0 * x)) *
		(1.0 - Exp(y)) * Exp(y) * (M_E - Exp(2.0 * y)) *
		(1.0 - Exp(z)) * (M_E - Exp(z)) * Exp(2.0 * z) +
	2.0 * (1.0 - Exp(x)) * (M_E - Exp(x)) * (M_E - Exp(2.0 * x)) *
		(1.0 - Exp(y)) * (M_E - Exp(y)) * Exp(2.0 * y) * Exp(z) *
		(M_E - Exp(z)) * (M_E - Exp(2.0 * z)) +
	2.0 * (1.0 - Exp(x)) * (M_E - Exp(x)) * (M_E - Exp(2.0 * x)) *
		(1.0 - Exp(y)) * (M_E - Exp(y)) * Exp(2.0 * y) *
		(1.0 - Exp(z)) * Exp(z) * (M_E - Exp(2.0 * z)) +
	4.0 * (1.0 - Exp(x)) * (M_E - Exp(x)) * (M_E - Exp(2.0 * x)) *
		(1.0 - Exp(y)) * (M_E - Exp(y)) * Exp(2.0 * y) *
		(1.0 - Exp(z)) * (M_E - Exp(z)) * Exp(2.0 * z) +
	8.0 * Sin(2.0 * M_PI * x) * Sin(2.0 * M_PI * y) *
		Sin(2.0 * M_PI * z) * M_PI * M_PI +
	z * (x - 1.0) * (y - 1.0) * (z - 1.0) * (x - 0.5) * (y - 0.5) *
		(z - 0.5) + x * z * (y - 1.0) * (z - 1.0) * (x - 0.5) *
		(y - 0.5) * (z - 0.5) +
	x * z * (x - 1.0) * (y - 1.0) * (z - 1.0) * (y - 0.5) * (z - 0.5) +
	y * z * (x - 1.0) * (z - 1.0) * (x - 0.5) * (y - 0.5) * (z - 0.5) +
	x * y * z * (z - 1.0) * (x - 0.5) * (y - 0.5) * (z - 0.5) +
	x * y * z * (x - 1.0) * (z - 1.0) * (y - 0.5) * (z - 0.5) +
	y * z * (x - 1.0) * (y - 1.0) * (z - 1.0) * (x - 0.5) * (z - 0.5) +
	x * y * z * (y - 1.0) * (z - 1.0) * (x - 0.5) * (z - 0.5) +
	x * y * z * (x - 1.0) * (y - 1.0) * (z - 1.0) * (z - 0.5) +
	Sin(2.0 * M_PI * x) * Sin(2.0 * M_PI * y) * Sin(2.0 * M_PI * z);

    *(values++) = (-1. - kk) * u[2] +
	y * (x - 1.0) * (y - 1.0) * (z - 1.0) * (x - 0.5) * (y - 0.5) *
		(z - 0.5) +
	x * y * (y - 1.0) * (z - 1.0) * (x - 0.5) * (y - 0.5) * (z - 0.5) +
	x * y * (x - 1.0) * (y - 1.0) * (z - 1.0) * (y - 0.5) * (z - 0.5) +
	y * z * (x - 1.0) * (y - 1.0) * (x - 0.5) * (y - 0.5) * (z - 0.5) +
	x * y * z * (y - 1.0) * (x - 0.5) * (y - 0.5) * (z - 0.5) +
	x * y * z * (x - 1.0) * (y - 1.0) * (y - 0.5) * (z - 0.5) +
	y * z * (x - 1.0) * (y - 1.0) * (z - 1.0) * (x - 0.5) * (y - 0.5) +
	x * y * z * (y - 1.0) * (z - 1.0) * (x - 0.5) * (y - 0.5) +
	x * y * z * (x - 1.0) * (y - 1.0) * (z - 1.0) * (y - 0.5) +
	Exp(x) * (M_E - Exp(x)) * (M_E - Exp(2.0 * x)) * (1.0 - Exp(y)) *
		(M_E - Exp(y)) * (M_E - Exp(2.0 * y)) * (1.0 - Exp(z)) *
		(M_E - Exp(z)) * (M_E - Exp(2.0 * z)) -
	2.0 * Exp(x) * Exp(x) * (M_E - Exp(2.0 * x)) * (1.0 - Exp(y)) *
		(M_E - Exp(y)) * (M_E - Exp(2.0 * y)) * (1.0 - Exp(z)) *
		(M_E - Exp(z)) * (M_E - Exp(2.0 * z)) -
	4.0 * Exp(x) * (M_E - Exp(x)) * Exp(2.0 * x) * (1.0 - Exp(y)) *
		(M_E - Exp(y)) * (M_E - Exp(2.0 * y)) * (1.0 - Exp(z)) *
		(M_E - Exp(z)) * (M_E - Exp(2.0 * z)) +
	(1.0 - Exp(x)) * Exp(x) * (M_E - Exp(2.0 * x)) * (1.0 - Exp(y)) *
		(M_E - Exp(y)) * (M_E - Exp(2.0 * y)) * (1.0 - Exp(z)) *
		(M_E - Exp(z)) * (M_E - Exp(2.0 * z)) -
	4.0 * (1.0 - Exp(x)) * Exp(x) * Exp(2.0 * x) * (1.0 - Exp(y)) *
		(M_E - Exp(y)) * (M_E - Exp(2.0 * y)) * (1.0 - Exp(z)) *
		(M_E - Exp(z)) * (M_E - Exp(2.0 * z)) +
	4.0 * (1.0 - Exp(x)) * (M_E - Exp(x)) * Exp(2.0 * x) * (1.0 - Exp(y)) *
		(M_E - Exp(y)) * (M_E - Exp(2.0 * y)) * (1.0 - Exp(z)) *
		(M_E - Exp(z)) * (M_E - Exp(2.0 * z)) + (1.0 - Exp(x)) *
		(M_E - Exp(x)) * (M_E - Exp(2.0 * x)) * Exp(y) *
		(M_E - Exp(y)) * (M_E - Exp(2.0 * y)) * (1.0 - Exp(z)) *
		(M_E - Exp(z)) * (M_E - Exp(2.0 * z)) - 2.0 * (1.0 - Exp(x)) *
		(M_E - Exp(x)) * (M_E - Exp(2.0 * x)) * Exp(y) * Exp(y) *
		(M_E - Exp(2.0 * y)) * (1.0 - Exp(z)) * (M_E - Exp(z)) *
		(M_E - Exp(2.0 * z)) - 4.0 * (1.0 - Exp(x)) * (M_E - Exp(x)) *
		(M_E - Exp(2.0 * x)) * Exp(y) * (M_E - Exp(y)) * Exp(2.0 * y) *
		(1.0 - Exp(z)) * (M_E - Exp(z)) * (M_E - Exp(2.0 * z)) +
	(1.0 - Exp(x)) * (M_E - Exp(x)) * (M_E - Exp(2.0 * x)) *
		(1.0 - Exp(y)) * Exp(y) * (M_E - Exp(2.0 * y)) *
		(1.0 - Exp(z)) * (M_E - Exp(z)) * (M_E - Exp(2.0 * z)) -
	4.0 * (1.0 - Exp(x)) * (M_E - Exp(x)) * (M_E - Exp(2.0 * x)) *
		(1.0 - Exp(y)) * Exp(y) * Exp(2.0 * y) * (1.0 - Exp(z)) *
		(M_E - Exp(z)) * (M_E - Exp(2.0 * z)) +
	4.0 * (1.0 - Exp(x)) * (M_E - Exp(x)) * (M_E - Exp(2.0 * x)) *
		(1.0 - Exp(y)) * (M_E - Exp(y)) * Exp(2.0 * y) *
		(1.0 - Exp(z)) * (M_E - Exp(z)) * (M_E - Exp(2.0 * z)) +
	4.0 * Sin(2.0 * M_PI * x) * Cos(2.0 * M_PI * y) * M_PI * M_PI *
		Cos(2.0 * M_PI * z) + (1.0 - Exp(x)) * (M_E - Exp(x)) *
		(M_E - Exp(2.0 * x)) * (1.0 - Exp(y)) * (M_E - Exp(y)) *
		(M_E - Exp(2.0 * y)) * (1.0 - Exp(z)) * (M_E - Exp(z)) *
		(M_E - Exp(2.0 * z));
    return;
}

static void
func_curl_u(FLOAT x, FLOAT y, FLOAT z, FLOAT *value)
/*  u = x y z (x-1.0)(y-1.0)(z-1.0)(x-0.5)(y-0.5)(z-0.5)
    v = Sin[2 Pi x] Sin[2 Pi y] Sin[2 Pi z]
    w = (1. - Exp[x])(E - Exp[x])(E - Exp[2. x]) (1. - Exp[y])(E - Exp[y])(E - Exp[2. y]) (1. - Exp[z])(E - Exp[z])(E - Exp[2. z])
    CForm[Simplify[D[w,y] - D[v,z]]]
    CForm[Simplify[D[u,z] - D[w,x]]]
    CForm[Simplify[D[v,x] - D[u,y]]] */
{
    (void)func_curl_u;

    *(value++) = -2.*Exp(2.*y)*(1. - Exp(x))*(M_E - Exp(x))*(M_E - Exp(2.*x))*
     (1. - Exp(y))*(M_E - Exp(y))*(1. - Exp(z))*(M_E - Exp(z))*
     (M_E - Exp(2.*z)) - Exp(y)*(1. - Exp(x))*(M_E - Exp(x))*
     (M_E - Exp(2.*x))*(1. - Exp(y))*(M_E - Exp(2.*y))*
     (1. - Exp(z))*(M_E - Exp(z))*(M_E - Exp(2.*z)) - 
    Exp(y)*(1. - Exp(x))*(M_E - Exp(x))*(M_E - Exp(2.*x))*
     (M_E - Exp(y))*(M_E - Exp(2.*y))*(1. - Exp(z))*(M_E - Exp(z))*
     (M_E - Exp(2.*z)) - 2*M_PI*Cos(2*M_PI*z)*Sin(2*M_PI*x)*Sin(2*M_PI*y);

    *(value++) = 2.*Exp(2.*x)*(1. - Exp(x))*(M_E - Exp(x))*(1. - Exp(y))*
     (M_E - Exp(y))*(M_E - Exp(2.*y))*(1. - Exp(z))*(M_E - Exp(z))*
     (M_E - Exp(2.*z)) + Exp(x)*(1. - Exp(x))*(M_E - Exp(2.*x))*
     (1. - Exp(y))*(M_E - Exp(y))*(M_E - Exp(2.*y))*(1. - Exp(z))*
     (M_E - Exp(z))*(M_E - Exp(2.*z)) + 
    Exp(x)*(M_E - Exp(x))*(M_E - Exp(2.*x))*(1. - Exp(y))*
     (M_E - Exp(y))*(M_E - Exp(2.*y))*(1. - Exp(z))*(M_E - Exp(z))*
     (M_E - Exp(2.*z)) + (-1. + x)*(-0.5 + x)*x*(-1. + y)*(-0.5 + y)*y*
     (-1. + z)*(-0.5 + z) + (-1. + x)*(-0.5 + x)*x*(-1. + y)*(-0.5 + y)*y*
     (-1. + z)*z + (-1. + x)*(-0.5 + x)*x*(-1. + y)*(-0.5 + y)*y*(-0.5 + z)*z;

    *(value++) =
   -((-1. + x)*(-0.5 + x)*x*(-1. + y)*(-0.5 + y)*(-1. + z)*(-0.5 + z)*z) - 
    (-1. + x)*(-0.5 + x)*x*(-1. + y)*y*(-1. + z)*(-0.5 + z)*z - 
    (-1. + x)*(-0.5 + x)*x*(-0.5 + y)*y*(-1. + z)*(-0.5 + z)*z + 
    2*M_PI*Cos(2*M_PI*x)*Sin(2*M_PI*y)*Sin(2*M_PI*z);

    return;
}

static void
build_linear_system(SOLVER *solver, DOF *u_h, DOF *f_h)
{
    int N = u_h->type->nbas;	/* number of basis functions in an element */
    GRID *g = u_h->g;
    SIMPLEX *e;
    int i, j;
    FLOAT A[N][N], B[N], buffer[N];
    INT I[N];

    assert(u_h->dim == 1);

    ForAllElements(g, e) {	/* \PHGindex{ForAllElements} */
	/* compute \int \curl\phi_i \cdot \curl\phi_j making use of symmetry */
	for (i = 0; i < N; i++) {
	    I[i] = phgSolverMapE2L(solver, 0, e, i);
	    for (j = 0; j <= i; j++)
		A[j][i] = A[i][j] =
		    phgQuadCurlBasDotCurlBas(e, u_h, i, u_h, j, QUAD_DEFAULT) -
		    kk * phgQuadBasDotBas(e, u_h, i, u_h, j, QUAD_DEFAULT);
	}

	/* compute local matrix and RHS */
	for (i = 0; i < N; i++) {	/* loop on basis functions */
	    if (phgDofDirichletBC(u_h, e, i, func_u, 
				buffer, B + i, DOF_PROJ_CROSS)) {
		phgSolverAddMatrixEntries(solver, 1, I + i, N, I, buffer); 
	    }
	    else {
		/* right hand side: \int f * phi_i */
#if 0
		B[i] = phgQuadFuncDotBas(e, func_f, u_h, i, QUAD_DEFAULT);
#else
		B[i] = phgQuadDofDotBas(e, f_h, u_h, i, QUAD_DEFAULT);
#endif
		phgSolverAddMatrixEntries(solver, 1, I + i, N, I, A[i]); 
	    }
	}
	phgSolverAddRHSEntries(solver, N, I, B);
    }
}

static FLOAT
estimate_error(DOF *u_h, DOF *f_h, DOF *curl_u_h, DOF *error)
{
    GRID *g = u_h->g;
    SIMPLEX *e;
    DOF *jmp1, *jmp2, *res, *div, *tmp;

    tmp = phgDofCopy(f_h, NULL, u_h->type, "f+k^2u_h");
    phgDofAXPY(kk, u_h, &tmp);			/* tmp = f_h + k^2 u_h */
    jmp1 = phgQuadFaceJump(curl_u_h, DOF_PROJ_CROSS, NULL, QUAD_DEFAULT);
    jmp2 = phgQuadFaceJump(tmp, DOF_PROJ_DOT, NULL, QUAD_DEFAULT);
    res = phgDofCurl(curl_u_h, NULL, NULL, "residual");
    phgDofAXPBY(1.0, tmp, -1.0, &res);		/* residual */
    div = phgDofDivergence(tmp, NULL, NULL, "div(f+k^2 u_h)");
    phgDofFree(&tmp);
    ForAllElements(g, e) {
	int i;
	FLOAT eta, h;
	FLOAT diam = phgGeomGetDiameter(g, e);
	e->mark = 0;		/* clear refinement mmark */
	eta = 0.0;
	for (i = 0; i < NFace; i++) {
	    INT fno;
	    if (e->bound_type[i] == DIRICHLET || e->bound_type[i] == NEUMANN)
		continue;	/* boundary face */
	    h = phgGeomGetFaceDiameter(g, e, i);
	    fno = e->faces[i];
	    eta += (*DofFaceData(jmp1, fno) + *DofFaceData(jmp2, fno)) * h;
	}
	eta += (phgQuadDofDotDof(e, res, res, QUAD_DEFAULT) +
		phgQuadDofDotDof(e, div, div, QUAD_DEFAULT)) * diam * diam;
	*DofElementData(error, e->index) = Sqrt(eta);
    }
    phgDofFree(&jmp1);
    phgDofFree(&jmp2);
    phgDofFree(&res);
    phgDofFree(&div);
    return phgDofNormInftyVec(error);
}

#define TEST_KK0 0

#if TEST_KK0
static void
build_poisson(SOLVER *solver, DOF *u_h, DOF *f_h)
{
    int N = u_h->type->nbas;    /* number of basis functions in an element */
    int dim = u_h->dim;         /* DOF dimension */
    GRID *g = u_h->g;
    SIMPLEX *e;
    int i, j, k;
    FLOAT A[N][N], B[N][dim], buffer[N], *row;
    INT I[N][dim], J[dim][N];

    ForAllElements(g, e) {
	/* compute \int \grad\phi_i \cdot \grad\phi_j making use of symmetry */
	for (i = 0; i < N; i++) {
	    for (k = 0; k < dim; k++)
		I[i][k] = J[k][i] = phgSolverMapE2L(solver, 0, e, i * dim + k);
	    for (j = 0; j <= i; j++)
		A[j][i] = A[i][j] =
		    phgQuadGradBasDotGradBas(e, u_h, i, u_h, j, QUAD_DEFAULT);
	}

	/* loop on basis functions */
	for (i = 0; i < N; i++) {
	    if (phgDofDirichletBC(u_h, e, i, NULL,
				  buffer, B[i], DOF_PROJ_NONE)) {
		row = buffer;
	    }
	    else if (phgDofGetElementBoundaryType(u_h, e, i * dim) & NEUMANN) {
		/* TODO */
		row = NULL;
		phgError(1, "%s:%d: unimplemented.\n", __FILE__, __LINE__);
	    }
	    else {
		/* right hand side: \int f * phi_i */
		phgQuadDofTimesBas(e, f_h, u_h, i, QUAD_DEFAULT, B[i]);
		row = A[i];
	    }
	    for (k = 0; k < dim; k++)
		phgSolverAddMatrixEntries(solver, 1, I[i] + k, N, J[k], row); 
	}
	phgSolverAddRHSEntries(solver, N * dim, I[0], B[0]);
    }
}

static void
poisson(DOF *f, DOF **phi)
/* solves the poisson equation -Laplace(phi) = f */
{
    GRID *g = f->g;
    SOLVER *solver;

    if (f->type->order <= 2)
	*phi = phgDofNew(g, DOF_Pn[f->type->order + 2], 1, "tmp", DofNoAction);
    else
	*phi = phgDofNew(g, DOF_HBn[f->type->order + 2], 1, "tmp", DofNoAction);

    /* save current options */
    phgOptionsPush();
    /* change default solver to PETSc */
    phgOptionsSetKeyword("default_solver", "petsc");
    solver = phgSolverCreate(SOLVER_DEFAULT, *phi, NULL);
    build_poisson(solver, *phi, f);
    phgSolverSolve(solver, TRUE, *phi, NULL);
    phgSolverDestroy(&solver);
    /* restore options */
    phgOptionsPop();
}
#endif	/* TEST_KK0 */

int
main(int argc, char *argv[])
{
    SIMPLEX *e;
    GRID *g;
    DOF *u_h, *u, *f_h, *error, *curl, *tmp;
    SOLVER *solver;
    FLOAT L2_err, Hcurl_err, a, b, thres, max_error;
    size_t mem, mem_peak;

    static char *fn = "../test/cube1o.dat";
    static char *vtk = NULL;
    static INT pre_refines = 0;
    static INT depth = 3;	/* <=0 => uniformly refine -depth or 3 times */
    static INT mem_max = 400;	/* max. memory */
    static FLOAT tol = 1e-3;	/* convergence tolerance */
    static BOOLEAN use_poisson = FALSE;	/* use Poisson matrices in AMS */

    phgOptionsRegisterTitle("\nProgram specific options", NULL);
    phgOptionsRegisterFilename("mesh_file", "Mesh file", (char **) &fn);
    phgOptionsRegisterFilename("vtk_file", "VTK file", (char **) &vtk);
    phgOptionsRegisterInt("pre_refines", "Pre-refines", &pre_refines);
    phgOptionsRegisterInt("refine_depth", "Refinement depth (<=0 means "
			  "uniformly refine -depth or 3 times)", &depth);
    phgOptionsRegisterInt("mem_max", "Max memory (MB)", &mem_max);
    phgOptionsRegisterFloat("tol", "Convergence tolerance", &tol);
    phgOptionsRegisterFloat("kk", "The value of k^2", &kk);
    phgOptionsRegisterNoArg("use_poisson", "Use Poisson matrices in AMS",
				&use_poisson);

    phgOptionsPreset("-default_dof_type ND1");
    if (SOLVER_HYPRE != NULL)
	phgOptionsPreset("-default_solver hypre -hypre_pc ams");

    phgInit(&argc, &argv);
    phgOptionsShowUsed();
   if (depth == 0)
	depth = -3;

    if (kk == 0. && SOLVER_HYPRE != NULL)
	phgOptionsSetNoArg("hypre_ams_singular", TRUE);

    g = phgNewGrid(-1);
    if (!phgImport(g, fn, FALSE))
	phgError(1, "can't read file \"%s\".\n", fn);

    phgRefineAllElements(g, pre_refines);

    u_h = phgDofNew(g, DOF_DEFAULT, 1, "u_h", DofInterpolation/*DofNoAction*/);
    phgDofSetDataByValue(u_h, 0.0);
    u = phgDofNew(g, DOF_DEFAULT, 1, "u", func_u);
    f_h = phgDofNew(g, /*DOF_DEFAULT, 1*/DOF_ANALYTIC, 3, "f_h", func_f);

    error = phgDofNew(g, DOF_P0, 1, "error estimates", DofNoAction);

    while (TRUE) {
	elapsed_time(g, FALSE, 0.);
	phgPrintf("\n------ %d DOF, %d elements, mesh LIF = %lg\n",
		  DofGetDataCountGlobal(u_h), g->nleaf_global, (double)g->lif);

	if (phgBalanceGrid(g, 1.2, 1, NULL, 0.)) {
	    phgPrintf("------ Repartition mesh: nprocs = %d, LIF = %lg ",
			    g->nprocs, (double)g->lif);
	    elapsed_time(g, TRUE, 0.);
	}

	phgPrintf("Set up linear solver: ");
	solver = phgSolverCreate(SOLVER_DEFAULT, u_h, NULL);
	phgPrintf("solver LIF = %lg ", (double)solver->mat->rmap->lif);
	elapsed_time(g, TRUE, 0.);

	phgPrintf("Build linear system ");
	if (kk != 0.0) {
	    build_linear_system(solver, u_h, f_h);
	}
	else {
	    /* singular problem, ensure compatibility condition of RHS */
	    curl = phgDofNew(g, DOF_DEFAULT, 1, "curl_u", func_curl_u);
	    tmp = phgDofCurl(curl, NULL, NULL, NULL);
	    phgDofFree(&curl);
	    build_linear_system(solver, u_h, tmp);
	    phgDofFree(&tmp);
	}
	elapsed_time(g, TRUE, 0.);

	if (use_poisson) {
	    phgPrintf("Build Poisson-matrices ");
	    phgSolverHypreAMSSetConstantPoisson(solver, 1.0, -kk);
	    elapsed_time(g, TRUE, 0.);
	}

	phgPrintf("Solve linear system: ");
	phgSolverSolve(solver, TRUE, u_h, NULL);
	phgPrintf("nits=%d, resid=%0.4lg ", solver->nits,
			(double)solver->residual);
	elapsed_time(g, TRUE, 0.);
	phgSolverDestroy(&solver);

#if TEST_KK0
	if (kk == 0.0) {
	    DOF *tmp1;
	    tmp = phgDofCopy(u, NULL, NULL, NULL);
	    phgDofAXPY(-1.0, u_h, &tmp);
	    tmp1 = phgDofDivergence(tmp, NULL, NULL, NULL);
	    phgDofFree(&tmp);
	    /* solve Poisson equation -Laplace(tmp) = tmp1 */
	    poisson(tmp1, &tmp);
	    phgDofFree(&tmp1);
	    tmp1 = phgDofGradient(tmp, NULL, NULL, NULL);
	    phgDofFree(&tmp);
	    phgDofAXPY(1.0, tmp1, &u_h);
	    phgDofFree(&tmp1);
	}
#endif	/* TEST_KK0 */

	phgPrintf("Errors: ");
	tmp = phgDofCopy(u, NULL, NULL, NULL);
	phgDofAXPY(-1.0, u_h, &tmp);
	phgPrintf("L2=%0.3le, ", (double)(L2_err = phgDofNormL2(tmp)));
#if 0
	phgDofFree(&tmp);
	curl = phgDofNew(g, DOF_DEFAULT, 1, "curl_u", func_curl_u);
	tmp = phgDofCurl(curl, NULL, NULL, NULL);
	phgDofCurl(u_h, &curl, NULL, NULL);
	phgDofAXPY(-1.0, curl, &tmp);
#else
	curl = phgDofCurl(tmp, NULL, NULL, NULL);
	phgDofFree(&tmp);
	tmp = curl;
	curl = phgDofCurl(u_h, NULL, NULL, NULL);
#endif
	phgPrintf("Hcurl=%0.3le, ", (double)(Hcurl_err = phgDofNormL2(tmp)));
	phgDofFree(&tmp);
	max_error = estimate_error(u_h, f_h, curl, error);
	phgPrintf("Est=%0.3le ", phgDofNormL2Vec(error));
	elapsed_time(g, TRUE, 0.);
	phgPrintf("Norms: L2=%0.3lg, Hcurl=%0.3lg\n",
			(double)phgDofNormL2(u_h), (double)phgDofNormL2(curl));
	mem = phgMemoryUsage(g, &mem_peak);
	a = mem / (1024.0 * 1024.0);
	b = mem_peak / (1024.0 * 1024.0);
	phgPrintf("Memory usage: current %0.4lgMB, peak %0.4lgMB\n",
		(double)a, (double)b);
	if (Hcurl_err < tol || mem_peak >= 1024 * (size_t)mem_max * 1024) {
	    if (vtk != NULL) {
		phgDofFree(&f_h);
		phgRefineAllElements(g, 0);
		tmp = phgDofCopy(u, NULL, u_h->type, "u-u_h");
		phgDofAXPY(-1.0, u_h, &tmp);
		phgPrintf("Export mesh to \"%s\" ",
			  phgExportVTK(g, vtk, u_h, tmp, u, NULL));
		phgDofFree(&tmp);
		elapsed_time(g, TRUE, 0.);
	    }
	    phgDofFree(&curl);
	    break;
	}

	phgPrintf("Refine mesh ");
	if (depth <= 0) {
	    phgRefineAllElements(g, -depth);
	}
	else {
#if 0
	    thres = 1e-3 / Sqrt((double) g->nleaf_global);
#else
	    thres = 0.5 * max_error;
#endif
	    ForAllElements(g, e)
		if (*DofElementData(error, e->index) >= thres)
		    e->mark = depth;
	    phgRefineMarkedElements(g);
	}
	phgDofFree(&curl);
	elapsed_time(g, TRUE, 0.);
    }
    phgDofFree(&u_h);
    phgDofFree(&f_h);
    phgDofFree(&u);
    phgDofFree(&error);

    phgFreeGrid(&g);
    phgFinalize();
    return 0;
}
