/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: matvec.h,v 1.63 2007/05/10 03:24:31 zlb Exp $ */

#ifndef PHG_MATVEC_H

/** Note: MAT_OP_N=notrans, MAT_OP_T=trans, MAT_OP_D=diagonal
 *	  may add MAT_OP_L=lower triangular part, and
 *	  MAT_OP_U=upper triangular part in the future */
typedef enum {MAT_OP_N, MAT_OP_T, MAT_OP_D} MAT_OP;

typedef struct MAT_ MAT;

typedef struct VEC_ {
    INT 	magic;		/* magic number */
    struct MAP_	*map;
    FLOAT	*data;
    FLOAT	*offp_data;	/* data in the range [nlocal, localsize) */
    MAT		*mat;		/* the vec is associated (the RHS) with mat */
    int		nvec;		/* number of vectors */
    BOOLEAN	assembled;	/* FALSE ==> needs assembly */
} VEC;

/* The mat-vec multiply function prototype for matrix-free matrices */
typedef int (*MV_FUNC)(MAT_OP op, struct MAT_ *A, VEC *x, VEC *y);

/* struct for a matrix row */
typedef struct {
    FLOAT	*data;		/* data, FLOAT[size] */
    INT		*cols;		/* global col indices, INT[size] */ 
    INT		ncols;		/* number of nonzero cols */
    INT		alloc;		/* allocated size of 'data' and 'cols' */
} MAT_ROW;

/* struct for saving factorized boundary equations, generated by
 * diagonalize_local_bdry_equations() */
typedef struct {
    FLOAT	*LU;	/* LU factorization */
    int		*pvt;	/* pivots */
    INT		*map;	/* indices of the unknowns */
    int		n;	/* number of equations */
} BDRY_EQNS;

/* Note: use PHG_XXXX instead of MAT_XXXX to avoid conflicts with PETSc */
typedef enum {
    PHG_UNPACKED	= 0,
    PHG_PACKED		= 1,
    PHG_DENSE		= 2,	/* for use in the future */
    PHG_MATRIX_FREE	= 3,
    PHG_DESTROYED	= 4	/* matrix entries have been destroyed */
} MAT_TYPE;

struct MAT_ {
    INT		magic;		/* magic number */
    struct MAP_	*rmap, *cmap;	/* row and column maps */
    MAT_ROW	*rows;  
    FLOAT	*diag;		/* diagonal data, used for scaling */
    MAT_TYPE	type;

    /* pointers for packed data */
    INT		*packed_cols;	/* packed column indices */
    FLOAT	*packed_data;	/* packed matrix data */
    size_t	nnz_d, nnz_o;

    /* localsize, O2Gmap and ordering are for off-proc columns in the matrix */
    INT		localsize;

    BOOLEAN	handle_bdry_eqns; /* whether process the bdry equations */

    /* members for matrix-free matrices */
    MV_FUNC	mv_func;
    void	**mv_data;

    /* members for marking boundary columns */
    INT		bdry_localsize, *bdry;

    /* members for saving factorized local bdry equations */
    BDRY_EQNS	*bdry_eqns;	/* list of local boundary equations */
    INT		bdry_eqns_count, bdry_eqns_allocated;

    /* member for saving bdry columns */
    MAT_ROW	*bdry_rows;	/* list of (saved) boundary columns */

    int		ref_count;	/* reference count */
    BOOLEAN	assembled;
};

#define PHG_DECIDE -1

#ifdef __cplusplus
extern "C" {
#endif

void phgMatSetupDiagonal(MAT *mat);
void phgMatPack(MAT *mat);
void phgMatUnpack(MAT *mat);
MAT *phgMatAXPBY(FLOAT a, MAT *x, FLOAT b, MAT **py);

#define phgMatCopy(src, dest)	phgMatAXPBY(1.0, src, 0.0, dest)

VEC *phgVecCreate(GRID *g, INT m, INT M, int nvec);
void phgVecDestroy(VEC **vec);
VEC *phgVecCopy(VEC *src, VEC **dest);
VEC *phgVecAXPBY(FLOAT a, VEC *x, FLOAT b, VEC **y);

VEC *phgMatVec(MAT_OP op, FLOAT alpha, MAT *A, VEC *x, FLOAT beta, VEC **y);
#define phgMatMultVec(mat, vec, result) \
	 phgMatVec(MAT_OP_N, (FLOAT)1.0, mat, vec, (FLOAT)0.0, result)
    
FLOAT phgVecDot(VEC *x, int which_x, VEC *y, int which_y, FLOAT *res);
FLOAT phgVecNorm2(VEC *src, int which, FLOAT *res);
FLOAT phgVecNorm1(VEC *src, int which, FLOAT *res);
FLOAT phgVecNormInfty(VEC *src, int which, FLOAT *res);

MAT *phgMatDup(MAT *src);
void phgMatDumpMATLAB(MAT *A, const char *vn, const char *fn, BOOLEAN dense);
VEC *phgVecRandomize(VEC *v, long int seed);

MAT *phgMatCreate(GRID *g, INT m, INT M);
MAT *phgMatCreateNonSquare(GRID *g, INT m, INT M, INT n, INT N);
MAT *phgMatCreateMatrixFree(GRID *g, INT m, INT M,
				MV_FUNC mv_func, void *mv_data, ...);
MAT *phgMatCreateMatrixFreeNonSquare(GRID *g, INT m, INT M, INT n, INT N,
				MV_FUNC mv_func, void *mv_data, ...);
void phgMatFreeMatrix(MAT *mat);
void phgMatDestroy(MAT **Mat);

struct MAP_ *phgMatGetRowMap(MAT *mat);
struct MAP_ *phgMatGetColumnMap(MAT *mat);

void phgMatAssemble(MAT *mat);
void phgVecAssemble(VEC *vec);

void phgMatAddEntry_(MAT *mat, INT row, INT col, FLOAT value,
			BOOLEAN row_is_global, BOOLEAN col_is_global);
#define phgMatAddEntry(mat, row, col, value) \
	phgMatAddEntry_(mat, row, col, value, FALSE, FALSE)
#define phgMatAddGlobalEntry(mat, row, col, value) \
	phgMatAddEntry_(mat, row, col, value, TRUE, TRUE)
#define phgMatAddLGEntry(mat, row, col, value) \
	phgMatAddEntry_(mat, row, col, value, FALSE, TRUE)
#define phgMatAddGLEntry(mat, row, col, value) \
	phgMatAddEntry_(mat, row, col, value, TRUE, FALSE)

void phgMatAddEntries_(MAT *mat, INT nrows, INT *rows, INT ncols, INT *cols,
			FLOAT *values, BOOLEAN rows_are_global,
			BOOLEAN cols_are_global);
#define phgMatAddEntries(mat, nrows, rows, ncols, cols, values) \
	phgMatAddEntries_(mat, nrows, rows, ncols, cols, values, FALSE, FALSE)
#define phgMatAddGlobalEntries(mat, nrows, rows, ncols, cols, values) \
	phgMatAddEntries_(mat, nrows, rows, ncols, cols, values, TRUE, TRUE)
#define phgMatAddLGEntries(mat, nrows, rows, ncols, cols, values) \
	phgMatAddEntries_(mat, nrows, rows, ncols, cols, values, FALSE, TRUE)
#define phgMatAddGLEntries(mat, nrows, rows, ncols, cols, values) \
	phgMatAddEntries_(mat, nrows, rows, ncols, cols, values, TRUE, FALSE)

void phgVecAddEntry(VEC *vec, int which, INT index, FLOAT value);
void phgVecAddEntries(VEC *vec, int which, INT n, INT *indices, FLOAT *values);

#define phgMat2Solver	phgSolverMat2Solver

MAT *phgMatRemoveBoundaryEntries(MAT *mat);
int phgMatEigenSolve(MAT *A, MAT *B, int n, int which, FLOAT tau,
		     int *nit, FLOAT *evals, struct MAP_ *map, DOF **u, ...);

int phgMatDumpCSR(MAT *mat, const char *fn);
int phgVecDump(VEC *vec, const char *fn);

#ifdef __cplusplus
}
#endif

#define PHG_MATVEC_H
#endif
