/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: dof-utils.c,v 1.49 2007/04/21 02:37:41 zlb Exp $ */

#include "phg.h"
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <math.h>
#include <stdarg.h>

FLOAT *
phgDofEval_(DOF *dof, SIMPLEX *e, const FLOAT lambda[], const FLOAT *basvalues,
	    int bas_stride, FLOAT *values, FLOAT **data)
/* evaluates DOF value at the barycentric coordinates 'lambda'.
 * 'values' is an array of DofDim(dof) entries provided
 * by the calling function containing the evaluated values.
 *
 * returns 'values'
 *
 * 'basvalues', if not NULL, points to the array of values of the basis
 * functions at 'lambda' evaluated previously ('lambda' not used in this case)
 *
 * Values of the basis functions are stored as basvalues[nbas][bas_stride],
 * if bas_stride == 0 then it is set to type->dim.
 *
 * 'data', if not NULL, points to an array of pointers pointing to the DOF
 * data of the element (NVert + NEdge + NFace + 1 entries). This is mainly
 * for use by the interpolation functions.
 */
{
    int i, j, k, l, n, nvalues;
    FLOAT *u;
    const FLOAT *v, *a, *b;
    int dim = dof->dim, dim0;

    if (dof->type == DOF_CONSTANT) {
	memcpy(values, dof->data, dim * sizeof(*values));
	return values;
    }
    else if (dof->type == DOF_ANALYTIC) {
	FLOAT x, y, z;
	if (lambda == NULL || dof->userfunc == NULL) {
	    phgError(1, "%s:%d, unexpected error.\n", __FILE__, __LINE__);
	}
	phgGeomLambda2XYZ(dof->g, e, lambda, &x, &y, &z);
	dof->userfunc(x, y, z, values);
	return values;
    }

    dim0 = dof->type->dim;

    if (dof->type->BasFuncs == NULL)
	phgError(1, "%s:%d, function 'BasFuncs' unvailable in \"%s\"!\n",
		 __FILE__, __LINE__, dof->type->name);

    /* get basis function values, adjust bas_stride -= type->dim */
    if (basvalues == NULL) {
	a = b = dof->type->BasFuncs(dof, e, 0, -1, lambda);
	bas_stride = 0;
    }
    else {
	a = b = basvalues;
	if (bas_stride <= 0)
	    bas_stride = 0;
	else
	    bas_stride -= dim0;
    }

    nvalues = dim * dim0;
    for (i = 0; i < nvalues; i++)
	values[i] = 0.0;

    if ((n = dof->type->np_vert) > 0) {
	for (k = 0; k < NVert; k++) {
	    v = (data == NULL) ? DofVertexData(dof, e->verts[k]) : data[k];
	    for (j = 0; j < n; j++) {
		u = values;
		for (i = 0; i < dim; i++, v++) {
		    a = b;
		    for (l = 0; l < dim0; l++) {
			*(u++) += (*(a++)) * (*v);
		    }
		}
		b = a + bas_stride;
	    }
	}
    }

    if ((n = dof->type->np_edge) > 1) {
	int M[n];
	for (k = 0; k < NEdge; k++) {
	    v = (data == NULL) ? DofEdgeData(dof, e->edges[k]) :
		data[NVert + k];
	    phgDofMapEdgeData(dof->type, e, k, M);
	    for (j = 0; j < n; j++) {
		u = values;
		for (i = 0; i < dim; i++) {
		    a = b;
		    for (l = 0; l < dim0; l++) {
			*(u++) += (*(a++)) * v[M[j] * dof->dim + i];
		    }
		}
		b = a + bas_stride;
	    }
	}
    }
    else if (n > 0) {
	for (k = 0; k < NEdge; k++) {
	    v = (data == NULL) ? DofEdgeData(dof, e->edges[k]) :
		data[NVert + k];
	    for (j = 0; j < n; j++) {
		u = values;
		for (i = 0; i < dim; i++, v++) {
		    a = b;
		    for (l = 0; l < dim0; l++) {
			*(u++) += (*(a++)) * (*v);
		    }
		}
		b = a + bas_stride;
	    }
	}
    }

    if ((n = dof->type->np_face) > 1) {
	int M[n];
	for (k = 0; k < NFace; k++) {
	    v = (data == NULL) ? DofFaceData(dof, e->faces[k]) :
		data[NVert + NEdge + k];
	    phgDofMapFaceData(dof->type, e, k, M);
	    for (j = 0; j < n; j++) {
		u = values;
		for (i = 0; i < dim; i++) {
		    a = b;
		    for (l = 0; l < dim0; l++) {
			*(u++) += (*(a++)) * v[M[j] * dof->dim + i];
		    }
		}
		b = a + bas_stride;
	    }
	}
    }
    else if (n > 0) {
	for (k = 0; k < NFace; k++) {
	    v = (data == NULL) ? DofFaceData(dof, e->faces[k]) :
		data[NVert + NEdge + k];
	    for (j = 0; j < n; j++) {
		u = values;
		for (i = 0; i < dim; i++, v++) {
		    a = b;
		    for (l = 0; l < dim0; l++) {
			*(u++) += (*(a++)) * (*v);
		    }
		}
		b = a + bas_stride;
	    }
	}
    }

    if ((n = dof->type->np_elem) > 0) {
	v = (data == NULL) ? DofElementData(dof, e->index) :
	    data[NVert + NEdge + NFace];
	for (j = 0; j < n; j++) {
	    u = values;
	    for (i = 0; i < dim; i++, v++) {
		a = b;
		for (l = 0; l < dim0; l++) {
		    *(u++) += (*(a++)) * (*v);
		}
	    }
	    b = a + bas_stride;
	}
    }

    return values;
}

static void
get_bas_funcs(DOF *src, DOF *dof, SIMPLEX *e, FLOAT *values)
/* computes values of basis funcs of 'src' at nodal points of 'dof'.
 * values points to an array of
	dof->type->nbas * src->type->nbas * src->type->dim
 * entries to hold the results */
{
    int count, j, k, i0, i1, i2;
    FLOAT *points, *v, *p;
    FLOAT lambda[Dim + 1] = { 0., 0., 0., 0. };

    count = src->type->nbas * src->type->dim;
    k = dof->type->nbas * count;
    v = values;

    points = dof->type->points;
    if (dof->type->np_vert > 0) {
	for (k = 0; k < NVert; k++) {
	    p = points;
	    for (j = 0; j < dof->type->np_vert; j++) {
		lambda[k] = *p;
		memcpy(v, src->type->BasFuncs(src, e, 0, -1, lambda),
		       count * sizeof(*values));
		p += 1;
		v += count;
	    }
	    lambda[k] = 0.;
	}
	points = p;
    }

    if (dof->type->np_edge > 0) {
	for (k = 0; k < NEdge; k++) {
	    p = points;
	    i0 = GetEdgeVertex(k, 0);
	    i1 = GetEdgeVertex(k, 1);
	    for (j = 0; j < dof->type->np_edge; j++) {
		lambda[i0] = p[0];
		lambda[i1] = p[1];
		memcpy(v, src->type->BasFuncs(src, e, 0, -1, lambda),
		       count * sizeof(*values));
		p += 2;
		v += count;
	    }
	    lambda[i0] = lambda[i1] = 0.;
	}
	points = p;
    }

    if (dof->type->np_face > 0) {
	for (k = 0; k < NFace; k++) {
	    p = points;
	    i0 = GetFaceVertex(k, 0);
	    i1 = GetFaceVertex(k, 1);
	    i2 = GetFaceVertex(k, 2);
	    for (j = 0; j < dof->type->np_face; j++) {
		lambda[i0] = p[0];
		lambda[i1] = p[1];
		lambda[i2] = p[2];
		memcpy(v, src->type->BasFuncs(src, e, 0, -1, lambda),
		       count * sizeof(*values));
		p += 3;
		v += count;
	    }
	    lambda[i0] = lambda[i1] = lambda[i2] = 0.;
	}
	points = p;
    }

    if (dof->type->np_elem > 0) {
	for (j = 0; j < dof->type->np_elem; j++) {
	    memcpy(v, src->type->BasFuncs(src, e, 0, -1, points),
		   count * sizeof(*values));
	    points += 4;
	    v += count;
	}
    }
}

static FLOAT *
get_bas_gradient(DOF *src, DOF *dof, SIMPLEX *e)
/* computes gradient of basis funcs of 'src' (with respect to lambda)
 * at nodal points of 'dof'.
 * returns a pointer to a buffer containing the computed values,
 * which should be freed by the calling functions */
{
    int grad_count, j, k, i0, i1, i2;
    FLOAT *points, *grad_values, *g, *p;
    FLOAT lambda[Dim + 1] = { 0., 0., 0., 0. };

    assert(dof->type->dim == src->type->dim);	/* to be implemented later */

    assert(src->type->invariant == TRUE);

    grad_count = src->type->nbas * (Dim + 1) * src->type->dim;
    k = dof->type->nbas * grad_count;
    g = grad_values = phgAlloc(k * sizeof(*grad_values));

    points = dof->type->points;
    if (dof->type->np_vert > 0) {
	for (k = 0; k < NVert; k++) {
	    p = points;
	    for (j = 0; j < dof->type->np_vert; j++) {
		lambda[k] = *p;
		memcpy(g, src->type->BasGrads(src, e, 0, -1, lambda),
		       grad_count * sizeof(*grad_values));
		p += 1;
		g += grad_count;
	    }
	    lambda[k] = 0.;
	}
	points = p;
    }

    if (dof->type->np_edge > 0) {
	for (k = 0; k < NEdge; k++) {
	    p = points;
	    i0 = GetEdgeVertex(k, 0);
	    i1 = GetEdgeVertex(k, 1);
	    for (j = 0; j < dof->type->np_edge; j++) {
		lambda[i0] = p[0];
		lambda[i1] = p[1];
		memcpy(g, src->type->BasGrads(src, e, 0, -1, lambda),
		       grad_count * sizeof(*grad_values));
		p += 2;
		g += grad_count;
	    }
	    lambda[i0] = lambda[i1] = 0.;
	}
	points = p;
    }

    if (dof->type->np_face > 0) {
	for (k = 0; k < NFace; k++) {
	    p = points;
	    i0 = GetFaceVertex(k, 0);
	    i1 = GetFaceVertex(k, 1);
	    i2 = GetFaceVertex(k, 2);
	    for (j = 0; j < dof->type->np_face; j++) {
		lambda[i0] = p[0];
		lambda[i1] = p[1];
		lambda[i2] = p[2];
		memcpy(g, src->type->BasGrads(src, e, 0, -1, lambda),
		       grad_count * sizeof(*grad_values));
		p += 3;
		g += grad_count;
	    }
	    lambda[i0] = lambda[i1] = lambda[i2] = 0.;
	}
	points = p;
    }

    if (dof->type->np_elem > 0) {
	for (j = 0; j < dof->type->np_elem; j++) {
	    memcpy(g, src->type->BasGrads(src, e, 0, -1, points),
		   grad_count * sizeof(*grad_values));
	    points += 4;
	    g += grad_count;
	}
    }

    return grad_values;
}

static FLOAT *
get_grad_lambda(DOF *dof, SIMPLEX *e, const FLOAT lambda[],
		const FLOAT *gradbas)
/* evaluates the gradient with respect to lambda at the barycentric
 * coordinates 'lambda'. returns a pointer to a buffer containing the
 *		[dof->dim][dof->type->dim][Dim+1]
 * evaluated values, which should be freed by the calling function.
 *
 * 'gradbas', if not NULL, points to the array of gradients of the basis
 * functions at 'lambda' evaluated previously ('lambda' not used in this case)
 */
{
    int i, j, k, l, m, dim;
    const FLOAT *b;
    FLOAT *v, *grad_lambda, *grad, *g;

    dim = DofDim(dof);
    grad = grad_lambda = phgAlloc(dim * (Dim + 1) * sizeof(*grad_lambda));
    if (gradbas == NULL) {
	assert(lambda != NULL);
	gradbas = dof->type->BasGrads(dof, e, 0, -1, lambda);
    }

    for (i = 0; i < dim * (Dim + 1); i++)
	grad_lambda[i] = 0.0;

    for (l = 0; l < Dim + 1; l++) {

	b = gradbas++;

	if (dof->type->np_vert > 0) {
	    for (k = 0; k < NVert; k++) {
		v = DofVertexData(dof, e->verts[k]);
		for (j = 0; j < dof->type->np_vert; j++) {
		    g = grad;
		    for (i = 0; i < dof->dim; i++) {
			for (m = 0; m < dof->type->dim; m++) {
			    *g += b[m * (Dim + 1)] * (*v);
			    g += Dim + 1;
			}
			v++;
		    }
		    b += dof->type->dim * (Dim + 1);
		}
	    }
	}

	if (dof->type->np_edge > 1) {
	    for (k = 0; k < NEdge; k++) {
		int M[dof->type->np_edge];
		v = DofEdgeData(dof, e->edges[k]);
		phgDofMapEdgeData(dof->type, e, k, M);
		for (j = 0; j < dof->type->np_edge; j++) {
		    g = grad;
		    for (i = 0; i < dof->dim; i++) {
			for (m = 0; m < dof->type->dim; m++) {
			    *g += b[m * (Dim + 1)] * v[M[j] * dof->dim + i];
			    g += Dim + 1;
			}
		    }
		    b += dof->type->dim * (Dim + 1);
		}
	    }
	}
	else if (dof->type->np_edge > 0) {
	    for (k = 0; k < NEdge; k++) {
		v = DofEdgeData(dof, e->edges[k]);
		for (j = 0; j < dof->type->np_edge; j++) {
		    g = grad;
		    for (i = 0; i < dof->dim; i++) {
			for (m = 0; m < dof->type->dim; m++) {
			    *g += b[m * (Dim + 1)] * (*v);
			    g += Dim + 1;
			}
			v++;
		    }
		    b += dof->type->dim * (Dim + 1);
		}
	    }
	}

	if (dof->type->np_face > 1) {
	    for (k = 0; k < NFace; k++) {
		int M[dof->type->np_face];
		v = DofFaceData(dof, e->faces[k]);
		phgDofMapFaceData(dof->type, e, k, M);
		for (j = 0; j < dof->type->np_face; j++) {
		    g = grad;
		    for (i = 0; i < dof->dim; i++) {
			for (m = 0; m < dof->type->dim; m++) {
			    *g += b[m * (Dim + 1)] * v[M[j] * dof->dim + i];
			    g += Dim + 1;
			}
		    }
		    b += dof->type->dim * (Dim + 1);
		}
	    }
	}
	else if (dof->type->np_face > 0) {
	    for (k = 0; k < NFace; k++) {
		v = DofFaceData(dof, e->faces[k]);
		for (j = 0; j < dof->type->np_face; j++) {
		    g = grad;
		    for (i = 0; i < dof->dim; i++) {
			for (m = 0; m < dof->type->dim; m++) {
			    *g += b[m * (Dim + 1)] * (*v);
			    g += Dim + 1;
			}
			v++;
		    }
		    b += dof->type->dim * (Dim + 1);
		}
	    }
	}

	if (dof->type->np_elem > 0) {
	    v = DofElementData(dof, e->index);
	    for (j = 0; j < dof->type->np_elem; j++) {
		g = grad;
		for (i = 0; i < dof->dim; i++) {
		    for (m = 0; m < dof->type->dim; m++) {
			*g += b[m * (Dim + 1)] * (*v);
			g += Dim + 1;
		    }
		    v++;
		}
		b += dof->type->dim * (Dim + 1);
	    }
	}

	grad++;
    }

    return grad_lambda;
}

FLOAT *
phgDofEvalGradient(DOF *dof, SIMPLEX *e, const FLOAT lambda[],
		   const FLOAT *gradbas, FLOAT *values)
/* evaluates the gradient at the barycentric coordinates 'lambda'.
 * 'values' is an array, provided by the calling function, of
 * [DofDim(dof)][Dim] entries containing the evaluated values.
 *
 * returns 'values'
 *
 * 'gradbas', if not NULL, points to the array of gradients of the basis
 * functions at 'lambda' evaluated previously ('lambda' not used in this case)
 */
{
    int i, j, k, nvalues;
    FLOAT *grad_lambda, *g;
    const FLOAT (*J)[Dim + 1];

    if (SpecialDofType(dof->type))
	phgError(1,
		 "%s:%d, cannot evaluate gradient of special DOF types.\n");

    if (dof->type->BasGrads == NULL)
	phgError(1, "%s:%d, function 'BasGrads' unvailable in \"%s\"!\n",
		 __FILE__, __LINE__, dof->type->name);

    nvalues = DofDim(dof);
    J = (void *)(phgGeomGetJacobian(dof->g, e));

    /* get gradient with respect to lambda */
    grad_lambda = get_grad_lambda(dof, e, lambda, gradbas);

    /* multiply grad_lambda by the Jacobian to get grad_xyz */
    for (i = 0; i < nvalues * Dim; i++)
	values[i] = 0.0;

    g = grad_lambda;
    for (k = 0; k < Dim + 1; k++) {
	for (j = 0; j < Dim; j++) {
	    for (i = 0; i < nvalues; i++) {
		values[i * Dim + j] += g[i * (Dim + 1) + k] * J[k][j];
	    }
	}
    }
    phgFree(grad_lambda);

    return values;
}

FLOAT *
phgDofEvalDivergence(DOF *dof, SIMPLEX *e, const FLOAT lambda[],
		     const FLOAT *gradbas, FLOAT *values)
/* evaluates the divergence at the barycentric coordinates 'lambda'.
 * 'values' is an array, provided by the calling function, of
 * [DofDim(dof) / Dim] entries containing the evaluated values.
 *
 * returns 'values'
 *
 * 'gradbas', if not NULL, points to the array of gradients of the basis
 * functions 'lambda' evaluated previously ('lambda' not used in this case)
 */
{
    int i, j, k, dim;
    FLOAT *grad_lambda, *g, d;
    FLOAT (*J)[Dim + 1];

    if (SpecialDofType(dof->type))
	phgError(1,
		 "%s:%d, can't evaluate divergence of special DOF types.\n");

    assert(DofDim(dof) % Dim == 0);

    if (dof->type->BasGrads == NULL)
	phgError(1, "%s:%d, function 'BasGrads' unvailable in \"%s\"!\n",
		 __FILE__, __LINE__, dof->type->name);

    J = (void *)(phgGeomGetJacobian(dof->g, e));

    /* get gradient with respect to lambda */
    grad_lambda = get_grad_lambda(dof, e, lambda, gradbas);

    /* multiply grad_lambda with columns of the Jacobian to compute div_xyz */
    dim = DofDim(dof) / Dim;
    g = grad_lambda;
    for (i = 0; i < dim; i++) {
	d = 0.0;
	for (j = 0; j < Dim; j++)
	    for (k = 0; k < Dim + 1; k++)
		d += g[(i * Dim + j) * (Dim + 1) + k] * J[k][j];
	values[i] = d;
    }

    phgFree(grad_lambda);

    return values;
}

FLOAT *
phgDofEvalCurl(DOF *dof, SIMPLEX *e, const FLOAT lambda[],
	       const FLOAT *gradbas, FLOAT *values)
/* evaluates the Curl at the barycentric coordinates 'lambda'.
 * 'values' is an array, provided by the calling function, of
 * [DofDim(dof) / Dim] entries containing the evaluated values.
 *
 * returns 'values'
 *
 * 'gradbas', if not NULL, points to the array of gradients of the basis
 * functions 'lambda' evaluated previously ('lambda' not used in this case)
 */
{
    int i, j, k, m, dim;
    FLOAT *grad_lambda, *grad_xyz, *g, d;
    FLOAT (*J)[Dim + 1];

    if (SpecialDofType(dof->type))
	phgError(1, "%s:%d, can't evaluate curl of special DOF types.\n");

    assert(DofDim(dof) % Dim == 0);

    if (dof->type->BasGrads == NULL)
	phgError(1, "%s:%d, function 'BasGrads' unvailable in \"%s\"!\n",
		 __FILE__, __LINE__, dof->type->name);

    J = (void *)(phgGeomGetJacobian(dof->g, e));

    /* get gradient with respect to lambda */
    grad_lambda = get_grad_lambda(dof, e, lambda, gradbas);

    /* multiply grad_lambda with columns of the Jacobian to compute grad_xyz */
    grad_xyz = phgAlloc(DofDim(dof) * Dim * sizeof(*grad_xyz));

    dim = DofDim(dof) / Dim;
    g = grad_lambda;
    for (i = 0; i < dim; i++) {
	for (m = 0; m < Dim; m++)
	    for (j = 0; j < Dim; j++) {
		d = 0.0;
		for (k = 0; k < Dim + 1; k++)
		    d += g[m * (Dim + 1) + k] * J[k][j];
		grad_xyz[m * Dim + j] = d;
	    }
	g += Dim * (Dim + 1);
    }
    /* compute curl_xyz using grad_xyz */
    g = grad_xyz;
    for (i = 0; i < dim; i++) {
	*(values++) = g[7] - g[5];
	*(values++) = g[2] - g[6];
	*(values++) = g[3] - g[1];
	g += (Dim * Dim);
    }

    phgFree(grad_lambda);
    phgFree(grad_xyz);

    return values;
}

static DOF *cache_dof;
static FLOAT *cache_buf = NULL;
static int cache_dim = 0;

static void
grad_func(DOF *dof, SIMPLEX *e, int bno, const FLOAT *lambda, FLOAT *values)
/* temperary Init function used by phgDofGradient */
{
    phgDofEvalGradient(cache_dof, e, lambda, cache_buf + bno * cache_dim,
				values);
}

DOF *
phgDofGradient_(DOF *src, DOF **dest_ptr, DOF_TYPE *newtype,
		const char *name, const char *srcfile, int srcline)
/* returns a new DOF which is the gradient of 'src' */
{
    char *s = NULL;
    DOF *dest = (dest_ptr == NULL ? NULL : *dest_ptr);

    MagicCheck(DOF, dest)

    if (dest != NULL)
	phgDofFree(&dest);

    if (newtype == NULL)
	newtype = src->type->grad_type;

    if (newtype == NULL)
	phgError(1, "phgDofGradient: gradient type for DOF \"%s\" undefined.\n",
		 src->type->name);

    /* create the new DOF */
    if (name != NULL) {
	s = phgAlloc(strlen(name) + 1);
	strcpy(s, name);
    }
    else {
	s = phgAlloc(strlen(src->name) + 1 + 6);
	sprintf(s, "grad(%s)", src->name);
    }
    dest = phgDofNew_(src->g, newtype, (Dim * DofDim(src)) / newtype->dim, s,
		     DofNoAction, srcfile, srcline);
    dest->DB_mask = src->DB_mask;
    phgFree(s);

    if (dest_ptr != NULL)
	*dest_ptr = dest;

    if (src->data == NULL)
	return dest;

    if (src->type->invariant == TRUE && dest->type->points != NULL) {
	cache_buf = get_bas_gradient(src, dest, src->g->roots);
	cache_dim = src->type->nbas * (Dim + 1) * src->type->dim;
    }
    cache_dof = src;

    phgDofSetDataByFunction_(dest, NULL, grad_func, NULL);

    if (cache_buf != NULL) {
	phgFree(cache_buf);
	cache_buf = NULL;
    }
    cache_dim = 0;

    return dest;
}

static void
div_func(DOF *dof, SIMPLEX *e, int bno, const FLOAT *lambda, FLOAT *values)
/* temperary Init function used by phgDofDivergence */
{
    phgDofEvalDivergence(cache_dof, e, lambda, cache_buf + bno * cache_dim,
			 values);
}

DOF *
phgDofDivergence_(DOF *src, DOF **dest_ptr, DOF_TYPE *newtype,
			const char *name, const char *srcfile, int srcline)
/* returns a new DOF which is the divergence of 'src' */
{
    char *s = NULL;
    DOF *dest = (dest_ptr == NULL ? NULL : *dest_ptr);

    MagicCheck(DOF, dest)

    if (dest != NULL)
	phgDofFree(&dest);

    assert(DofDim(src) % (Dim * src->type->grad_type->dim) == 0);

    if (newtype == NULL)
	newtype = src->type->grad_type;

    if (newtype == NULL)
	phgError(1, "%s:%d, gradient type for DOF \"%s\" undefined.\n",
		 __FILE__, __LINE__, src->type->name);

    /* create the new DOF */
    if (name != NULL) {
	s = phgAlloc(strlen(name) + 1);
	strcpy(s, name);
    }
    else {
	s = phgAlloc(strlen(src->name) + 1 + 6);
	sprintf(s, "div(%s)", src->name);
    }
    dest = phgDofNew_(src->g, newtype, DofDim(src) / (Dim * newtype->dim), s,
		     DofNoAction, srcfile, srcline);
    dest->DB_mask = src->DB_mask;
    phgFree(s);

    if (dest_ptr != NULL)
	*dest_ptr = dest;

    if (src->data == NULL)
	return dest;

    if (src->type->invariant == TRUE && dest->type->points != NULL) {
	cache_buf = get_bas_gradient(src, dest, src->g->roots);
	cache_dim = src->type->nbas * (Dim + 1) * src->type->dim;
    }
    cache_dof = src;

    phgDofSetDataByFunction_(dest, NULL, div_func, NULL);

    if (cache_buf != NULL) {
	phgFree(cache_buf);
	cache_buf = NULL;
    }
    cache_dim = 0;

    return dest;
}

static void
curl_func(DOF *dof, SIMPLEX *e, int bno, const FLOAT *lambda, FLOAT *values)
/* temperary Init function used by phgDofCurl */
{
    phgDofEvalCurl(cache_dof, e, lambda, cache_buf + bno * cache_dim, values);
}

DOF *
phgDofCurl_(DOF *src, DOF **dest_ptr, DOF_TYPE *newtype,
		const char *name, const char *srcfile, int srcline)
{
    char *s;
    int dof_dim;
    DOF *dest = (dest_ptr == NULL ? NULL : *dest_ptr);

    MagicCheck(DOF, dest)

    if (dest != NULL)
	phgDofFree(&dest);

    if (newtype == NULL)
	newtype = src->type->grad_type;

    if (newtype == NULL)
	phgError(1, "%s:%d, gradient type for DOF \"%s\" undefined.\n",
		 __FILE__, __LINE__, src->type->name);

    dof_dim = DofDim(src);
    assert(dof_dim % newtype->dim == 0 && dof_dim % Dim == 0);
    dof_dim /= newtype->dim;

    /* create the new DOF */
    if (name != NULL) {
	s = phgAlloc(strlen(name) + 1);
	strcpy(s, name);
    }
    else {
	s = phgAlloc(strlen(src->name) + 1 + 6);
	sprintf(s, "curl(%s)", src->name);
    }
    dest = phgDofNew_(src->g, newtype, dof_dim, s, DofNoAction,
						srcfile, srcline);
    dest->DB_mask = src->DB_mask;
    phgFree(s);

    if (dest_ptr != NULL)
	*dest_ptr = dest;

    if (src->data == NULL)
	return dest;

#if 1
    if (newtype->InitFunc == phgDofInitFuncPoint) {
	/* simpler code when grad_type is a point function */
	FLOAT *g, *a;
	INT k, n;
	DOF *grad;

	grad = phgDofGradient(src, NULL, NULL, NULL);
	g = DofData(grad);
	a = DofData(dest);
	n = DofGetDataCount(dest);
	for (k = 0; k < n; k += Dim) {
	    *(a++) = g[7] - g[5];
	    *(a++) = g[2] - g[6];
	    *(a++) = g[3] - g[1];
	    g += (Dim * Dim);
	}
	phgDofFree(&grad);
	return dest;
    }
#endif

    if (src->type->invariant == TRUE &&
	dest->type->InitFunc == phgDofInitFuncPoint) {
	cache_buf = get_bas_gradient(src, dest, src->g->roots);
	cache_dim = src->type->nbas * (Dim + 1) * src->type->dim;
    }
    cache_dof = src;

    phgDofSetDataByFunction_(dest, NULL, curl_func, NULL);

    if (cache_buf != NULL) {
	phgFree(cache_buf);
	cache_buf = NULL;
    }
    cache_dim = 0;
    cache_dof = NULL;

    return dest;
}

BOOLEAN
phgDofDirichletBC(DOF *u, SIMPLEX *e, int bas_index, DOF_USER_FUNC func,
		  FLOAT mat[], FLOAT rhs[], DOF_PROJ proj)
/* computes BC using either interpolation or L2-projection on the boundary.
 *
 *   0) bas_index = (element DOF index / u->dim);
 *
 *   1) This function turns the Dirichlet BC at a given DOF into an equation.
 *	If u->dim == 1, the equation is given by:
 *		\sum_{0 <= j < u->type->nbas} mat[j] u[j] = *rhs
 *	where the coefficients mat[j] = <\phi_j, \phi_{bas_index}>, and
 *	the right hand side *rhs = <func, \phi_{bas_index}>, in which <.,.>
 *	denotes the inner product on an edge or on a face, depending on the
 *	geometric type of the DOF.
 *
 *	Note: if enforce_L2 == FALSE and u->type->points != NULL, direct
 *	evaluation (interpolation) is used, i.e.:
 *		mat[bas_index] = 1.0, mat[j] = 0 for j != bas_index
 *		*rhs = the function value at DOF location.
 *	If enforce_L2 == TRUE, then L2 projection is enforced.
 *
 *   2) u->type->nbas coefficients are returned in mat[], which are independent
 *	of u->dim, and u->dim values are returned in rhs[] (if rhs != NULL).
 *
 *	The user program is resposible for duplicating the coefficients in
 *	mat[] according to u->dim when assembling the global matrix.
 *
 *   3) 'proj' must be either DOF_PROJ_NONE or DOF_PROJ_CROSS, the latter case
 *	requires u->type->dim == Dim and is a special case for the Dirichlet
 *	BC of the Maxwell equations: u \cros n = g \cross n
 *
 *   4) func == NULL means func(x) == 0
 */
{
    BOOLEAN enforce_L2 = FALSE;
    GRID *g = u->g;
    DOF_TYPE *type = u->type;
    int i, j, k, n0, n1, index, v0, v1, v2;
    int N = type->nbas;
    const FLOAT *b, *qp, *qw, *p;
    FLOAT *q, lambda[Dim + 1], w0, x, y, z;
    FLOAT f[DofDim(u)], w[type->dim];
    QUAD *quad;
    BTYPE btype;
    /* flags indicating which basis functions to use in the L2 projection */
    unsigned char flags[N];

    /* Warning: if 'func' calls phgDofEval with the same DOF_TYPE, then the
	values from BasFuncs are destroyed after calling 'func' */
    assert(proj == DOF_PROJ_NONE || type->dim == Dim);

    btype = phgDofGetElementBoundaryType(u, e, bas_index * u->dim);
    if (!(btype & u->DB_mask))
	return FALSE;

    for (i = 0; i < N; i++)
	mat[i] = 0.0;

#if 0
    /* only evaluate BC when the submesh is the owner */
    if (rhs != NULL && !(btype & OWNER)) {
	for (i = 0; i < u->dim * u->type->dim; i++)
	    rhs[i] = 0.;
	return TRUE;
    }
#endif

    if (!enforce_L2 && proj == DOF_PROJ_NONE && type->points != NULL) {
	mat[bas_index] = 1.0;
	if (rhs == NULL)
	    return TRUE;
	if (func == NULL) {
	    bzero(rhs, u->dim * u->type->dim * sizeof(*rhs));
	}
	else {
	    p = phgDofGetElementCoordinates(u, e, bas_index * u->dim);
	    func(p[0], p[1], p[2], rhs);
	}
	return TRUE;
    }

    /* TODO: up to now with all basis functions, bdry DOFs can be
     * evaluated by calling InitFunc which only needs func values
     * at the boundary, moreover phgDofDirichletBC is called
     * in proper order. Thus, direct evaluation of bdry DOFs by
     * calling InitFunc is possible and is faster (when np_xxxx > 1,
     * the values can be cached for the next few calls) */

    /* Note: use different quadrature orders for different basis
     * functions may save some (marginal) computations */

    switch (phgDofGetElementGeometryType(u, bas_index * u->dim, &index)) {
	case VERTEX:
	    assert(proj == DOF_PROJ_NONE);
	    lambda[0] = lambda[1] = lambda[2] = lambda[3] = 0.0;
	    lambda[index] = 1.0;
	    n0 = index * u->type->np_vert;
	    n1 = n0 + u->type->np_vert;
	    b = type->BasFuncs(u, e, n0, n1, lambda);
	    if (type->dim == 1) {
		/* faster code for type->dim == 1 */
		w0 = b[bas_index - n0];
		for (i = n0; i < n1; i++)
		    mat[i] = w0 * *(b++);
		if (rhs == NULL)
		    break;
		if (func == NULL) {
		    bzero(rhs, u->dim * sizeof(*rhs));
		}
		else {
		    phgGeomLambda2XYZ(g, e, lambda, &x, &y, &z);
		    func(x, y, z, rhs);
		    for (i = 0; i < u->dim; i++)
			rhs[i] *= w0;
		}
	    }
	    else {
		p = b + (bas_index - n0) * type->dim;
		for (i = n0; i < n1; i++) {
		    w0 = *(b++) * p[0];
		    for (j = 1; j < type->dim; j++)
			w0 += *(b++) * p[j];
		    mat[i] = w0;
		}
		if (rhs == NULL)
		    break;
		if (func == NULL) {
		    bzero(rhs, u->dim * u->type->dim * sizeof(*rhs));
		}
		else {
		    phgGeomLambda2XYZ(g, e, lambda, &x, &y, &z);
		    func(x, y, z, q = f);
		    for (i = 0; i < u->dim; i++) {
			w0 = *(q++) * p[0];
			for (j = 1; j < type->dim; j++)
			    w0 += *(q++) * p[j];
			*(rhs++) = w0;
		    }
		}
	    }
	    break;
	case EDGE:
	    /* L2 projection on the edge */
	    if (rhs != NULL)
		for (i = 0; i < u->dim; i++)
		    rhs[i] = 0.0;
	    GetEdgeVertices(e, index, v0, v1);
	    quad = phgQuadGetQuad1D(2 * type->order);
	    lambda[0] = lambda[1] = lambda[2] = lambda[3] = 0.0;
	    qp = quad->points;
	    qw = quad->weights;
	    memset(flags, 0, N);
	    /* basis functions on the vertices of the current edge */
	    if (type->np_vert > 0) {
		memset(flags + v0 * type->np_vert, 1, type->np_vert);
		memset(flags + v1 * type->np_vert, 1, type->np_vert);
	    }
	    /* basis functions on the current edge */
	    n0 = type->np_vert * NVert + type->np_edge * index;
	    memset(flags + n0, 1, type->np_edge);
	    for (k = 0; k < quad->npoints; k++) {
		lambda[v0] = *(qp++);
		lambda[v1] = *(qp++);
		b = type->BasFuncs(u, e, 0, -1, lambda);
		w0 = *(qw++);
		if (type->dim == 1) {	/*-----------------------------------*/
		    /* faster code for type->dim == 1 */
		    w0 *= b[bas_index];
		    for (i = 0; i < N; i++) {
			if (flags[i] == 0)
			    continue;
			mat[i] += b[i] * w0;
		    }
		    if (func != NULL && rhs != NULL) {
			phgGeomLambda2XYZ(g, e, lambda, &x, &y, &z);
			func(x, y, z, f);
			for (i = 0; i < u->dim; i++)
			    rhs[i] += f[i] * w0;
		    }
		} else if (proj == DOF_PROJ_CROSS) {	/*-------------------*/
		    /* Dirichlet BC for Maxwell equations: n x u = n x g */
		    /* Bdry eqn:  <t . u, t . phi_j> = <t . g, t . phg_j>,
		     * where t is the edge vector */
		    COORD *c0, *c1;
		    FLOAT t0, t1, t2;
		    c0 = g->verts + e->verts[v0];
		    c1 = g->verts + e->verts[v1];
		    t0 = (*c1)[0] - (*c0)[0];
		    t1 = (*c1)[1] - (*c0)[1];
		    t2 = (*c1)[2] - (*c0)[2];
		    q = (void *)(b + bas_index * Dim);
		    w0 *= t0 * q[0] + t1 * q[1] + t2 * q[2];
		    for (i = 0; i < N; i++, b += Dim) {
			if (flags[i] == 0)
				continue;
			mat[i] += w0 * (t0 * b[0] + t1 * b[1] + t2 * b[2]);
		    }
		    if (func != NULL && rhs != NULL) {
			phgGeomLambda2XYZ(g, e, lambda, &x, &y, &z);
			func(x, y, z, f);
			for (i = 0, q = f; i < u->dim; i++, q += Dim)
			    rhs[i] += w0 * (t0 * q[0] + t1 * q[1] + t2 * q[2]);
		    }
		}
		else {	/*---------------------------------------------------*/
		    p = b + bas_index * type->dim;
		    for (j = 0; j < type->dim; j++)
			w[j] = *(p++) * w0;
		    for (i = 0; i < N; i++) {
			if (flags[i] == 0) {
			    b += type->dim;
			    continue;
			}
			w0 = mat[i];
			for (j = 0; j < type->dim; j++)
			    w0 += *(b++) * w[j];
			mat[i] = w0;
		    }	/* for i */
		    if (func != NULL && rhs != NULL) {
			phgGeomLambda2XYZ(g, e, lambda, &x, &y, &z);
			func(x, y, z, q = f);
			for (i = 0; i < u->dim; i++) {
			    w0 = rhs[i];
			    for (j = 0; j < type->dim; j++)
				w0 += *(q++) * w[j];
			    rhs[i] = w0;
			}
		    }
		}	/*---------------------------------------------------*/
	    }	/* for k */
#if 0
	    /* multiply by edge length */
	    {
		COORD *c0, *c1;
		c0 = g->verts + e->verts[v0];
		c1 = g->verts + e->verts[v1];
		x = (*c1)[0] - (*c0)[0];
		y = (*c1)[1] - (*c0)[1];
		z = (*c1)[2] - (*c0)[2];
		w = Sqrt(x * x + y * y + z * z);
	    }
	    for (i = 0; i < N; i++)
		mat[i] *= w;
	    if (rhs == NULL)
		break;
	    for (i = 0; i < u->dim; i++)
		rhs[i] *= w;
#endif
	    break;
	case FACE:
	    /* L2 projection on the face */
	    if (rhs != NULL)
		for (i = 0; i < u->dim; i++)
		    rhs[i] = 0.0;
	    GetFaceVertices(e, index, v0, v1, v2);
	    quad = phgQuadGetQuad2D(2 * type->order);
	    lambda[index] = 0.0;
	    qp = quad->points;
	    qw = quad->weights;
	    memset(flags, 0, N);
	    /* basis functions on the vertices of the current face */
	    if (type->np_vert > 0) {
		memset(flags + v0 * type->np_vert, 1, type->np_vert);
		memset(flags + v1 * type->np_vert, 1, type->np_vert);
		memset(flags + v2 * type->np_vert, 1, type->np_vert);
	    }
	    /* basis functions on the edges of the current face */
	    n0 = type->np_vert * NVert;
	    if (type->np_edge > 0) {
		i = GetEdgeNo(v0, v1);
		memset(flags + n0 + i * type->np_edge, 1, type->np_edge);
		i = GetEdgeNo(v0, v2);
		memset(flags + n0 + i * type->np_edge, 1, type->np_edge);
		i = GetEdgeNo(v1, v2);
		memset(flags + n0 + i * type->np_edge, 1, type->np_edge);
	    }
	    /* basis functions on the current face */
	    n0 += type->np_edge * NEdge + type->np_face * index;
	    memset(flags + n0, 1, type->np_face);
	    for (k = 0; k < quad->npoints; k++) {
		lambda[v0] = *(qp++);
		lambda[v1] = *(qp++);
		lambda[v2] = *(qp++);
		b = type->BasFuncs(u, e, 0, -1, lambda);
		w0 = *(qw++);
		if (type->dim == 1) {
		    /* faster code for type->dim == 1 */
		    w0 *= b[bas_index];
		    for (i = 0; i < N; i++) {
			if (flags[i] == 0)
			    continue;
			mat[i] += b[i] * w0;
		    }
		    if (func != NULL && rhs != NULL) {
			phgGeomLambda2XYZ(g, e, lambda, &x, &y, &z);
			func(x, y, z, f);
			for (i = 0; i < u->dim; i++)
			    rhs[i] += f[i] * w0;
		    }
		}
		else {
		    const FLOAT *n = phgGeomGetFaceOutNormal(g, e, index);
		    if (proj == DOF_PROJ_CROSS) {
			/* Dirichlet BC for Maxwell equations: n x u = n x g */
			/* Bdry eqn:  <u x n, phi_j x n> = <g x n, phg_j x n> */

			/* compute basis x n */
			/* Warning: overriding static buffer in BasFuncs */
			for (i = 0, q = (void *)b; i < N; i++, q += Dim) {
			    x = q[1] * n[2] - q[2] * n[1];
			    y = q[2] * n[0] - q[0] * n[2];
			    z = q[0] * n[1] - q[1] * n[0];
			    q[0] = x;
			    q[1] = y;
			    q[2] = z;
			}
		    }
		    p = b + bas_index * type->dim;
		    for (j = 0; j < type->dim; j++)
			w[j] = *(p++) * w0;
		    for (i = 0; i < N; i++) {
			if (flags[i] == 0) {
			    b += type->dim;
			    continue;
			}
			w0 = mat[i];
			for (j = 0; j < type->dim; j++)
			    w0 += *(b++) * w[j];
			mat[i] = w0;
		    }
		    if (func != NULL && rhs != NULL) {
			phgGeomLambda2XYZ(g, e, lambda, &x, &y, &z);
			func(x, y, z, q = f);
			if (proj == DOF_PROJ_CROSS) {
			    /* compute g x n */
			    for (i = 0; i < u->dim; i++, q += Dim) {
				x = q[1] * n[2] - q[2] * n[1];
				y = q[2] * n[0] - q[0] * n[2];
				z = q[0] * n[1] - q[1] * n[0];
				q[0] = x;
				q[1] = y;
				q[2] = z;
			    }
			    q = f;
			}
			for (i = 0; i < u->dim; i++) {
			    w0 = rhs[i];
			    for (j = 0; j < type->dim; j++)
				w0 += *(q++) * w[j];
			    rhs[i] = w0;
			}
		    }
		}
	    }
#if 0
	    /* multiply by face area */
	    w = phgGeomGetFaceArea(g, e, index);
	    for (i = 0; i < N; i++)
		mat[i] *= w;
	    if (rhs == NULL)
		break;
	    for (i = 0; i < u->dim; i++)
		rhs[i] *= w;
#endif
	    break;
	case ELEMENT:
	    phgError(1, "%s:%d: unimplemented.\n", __FILE__, __LINE__);
	    break;
    }

    return TRUE;	/* to make gcc happy */
}

/*--------------------------- BLAS-like functiones ----------------------*/

static FLOAT *bas;
static size_t bas_count;

static void
func(DOF *dof, SIMPLEX *e, int bno, const FLOAT *lambda, FLOAT *values)
{
    Unused(bno);
    phgDofEval_(cache_dof, e, lambda, bas, 0, values, NULL);
    bas += bas_count;
}

DOF *
phgDofCopy_(DOF *src, DOF **dest_ptr, DOF_TYPE *newtype,
	    const char *name, const char *srcfile, int srcline)
/* returns a new DOF whose type is 'newtype', and whose values are evaluated
 * using 'src' (a copy of src). */
{
    GRID *g = src->g;
    SIMPLEX *e;
    FLOAT w = 1.0, *basvalues = NULL, *a, *d, *buffer = NULL;
    int i, k, dim, count = 0, nvalues;
    INT n;
    DOF *wgts = NULL;
    DOF *dest = (dest_ptr == NULL ? NULL : *dest_ptr);
    BYTE *flags0, *flags;
    char *auto_name;

    MagicCheck(DOF, dest)

    if (dest != NULL)
	phgDofFree(&dest);

    if (newtype == NULL)
	newtype = src->type;

    assert(!SpecialDofType(newtype));

    dim = DofDim(src);
    assert(dim % newtype->dim == 0);

    /* assign a default name to help debugging */
    if ((auto_name = (void *)name) == NULL) {
#if 0
	auto_name = phgAlloc(strlen(src->name) + 8 + 1);
	sprintf(auto_name, "copy of %s", src->name);
#else
	auto_name = strdup(src->name);
#endif
    }
    dest = phgDofNew_(g, newtype, dim / (newtype == NULL ? 1 : newtype->dim),
		     auto_name, DofNoAction, srcfile, srcline);
    dest->DB_mask = src->DB_mask;

    if (dest_ptr != NULL)
	*dest_ptr = dest;

    if (auto_name != name)
	phgFree(auto_name);
    dest->userfunc = DofNoAction;
    if (newtype == src->type && dest->data != NULL) {
	/* simply duplicate the data */
	size_t size = (newtype == NULL) ? dest->dim : DofGetDataCount(dest);
	memcpy(dest->data, src->data, sizeof(FLOAT) * size);
	return dest;
    }

    if (src->type == DOF_ANALYTIC) {
	phgDofSetDataByFunction(dest, src->userfunc);
	return dest;
    }

    if (newtype->BasFuncs == NULL)
	phgError(1, "phgDofCopy: basis funcs for DOF type \"%s\" undefined.\n",
		 newtype->name);

    if (!SpecialDofType(src->type) && newtype->points != NULL) {
	count = src->type->nbas * src->type->dim;
	basvalues = phgAlloc(newtype->nbas * count * sizeof(*basvalues));

	if (src->type->invariant == TRUE && dest->type->points != NULL)
	    get_bas_funcs(src, dest, src->g->roots, basvalues);
    }

    flags0 = phgCalloc((newtype->np_vert > 0 ? g->nvert : 0)
		       + (newtype->np_edge > 0 ? g->nedge : 0)
		       + (newtype->np_face > 0 ? g->nface : 0),
		       sizeof(*flags0));

    if (!SpecialDofType(src->type) && src->type->continuity < 0) {
	static DOF_TYPE DOF_WGTS = {DofCache,
	    "Weights", NULL, NULL, NULL, NULL, NULL,
	    phgDofInitFuncPoint, NULL, NULL,
	    FALSE, FALSE, -1, 0, 0, -1, 1, 0, 0, 0, 0
	};
	DOF_WGTS.np_vert = (newtype->np_vert > 0) ? 1 : 0;
	DOF_WGTS.np_edge = (newtype->np_edge > 0) ? 1 : 0;
	DOF_WGTS.np_face = (newtype->np_face > 0) ? 1 : 0;
	DOF_WGTS.nbas = DOF_WGTS.np_vert * NVert + DOF_WGTS.np_edge * NEdge +
	    DOF_WGTS.np_face * NFace;
	if (DOF_WGTS.nbas > 0) {
	    /* Other cases will be implemented later when really needed */
	    wgts = phgDofNew(g, &DOF_WGTS, 1, "weights", DofNoAction);
	    phgDofSetDataByValue(wgts, 0.0);
	    phgDofSetDataByValue(dest, 0.0);
	    /* allocate buffer for storing weighted vertex/edge/face data */
	    if ((n = DofGetDataCount(dest) - DofGetElementDataCount(dest)) > 0)
		buffer = phgCalloc(n, sizeof(*buffer));
	}
    }

    nvalues = dest->dim;
    cache_dof = src;
    bas_count = count;
    ForAllElements(g, e) {
	if (wgts != NULL) {
#if 0
	    /* use element volume as weight */
	    w = phgGeomGetVolume(g, e);
#else
	    /* use 1 as weight */
	    w = 1.0;
#endif
	}

	if (!SpecialDofType(src->type) && src->type->invariant == FALSE
	    && newtype->points != NULL)
	    get_bas_funcs(src, dest, e, basvalues);
	bas = basvalues;
	flags = flags0;

	if (newtype->np_vert > 0) {
	    for (k = 0; k < NVert; k++) {
		if (flags[n = e->verts[k]] && wgts == NULL) {
		    bas += count * newtype->np_vert;
		    continue;
		}
		flags[n] = TRUE;
		a = DofVertexData(dest, n);
		newtype->InitFunc(dest, e, VERTEX, k, NULL, func, NULL, a,
					NULL);
		if (wgts != NULL) {
		    d = buffer + (a - DofData(dest));
		    for (i = 0; i < nvalues * newtype->np_vert; i++)
			*(d++) += *(a++) * w;
		    *DofVertexData(wgts, n) += w;
		}
	    }
	    flags += g->nvert;
	}

	if (newtype->np_edge > 0) {
	    for (k = 0; k < NEdge; k++) {
		if (flags[n = e->edges[k]] && wgts == NULL) {
		    bas += count * newtype->np_edge;
		    continue;
		}
		flags[n] = TRUE;
		a = DofEdgeData(dest, n);
		newtype->InitFunc(dest, e, EDGE, k, NULL, func, NULL, a,
					NULL);
		if (wgts != NULL) {
		    d = buffer + (a - DofData(dest));
		    for (i = 0; i < nvalues * newtype->np_edge; i++)
			*(d++) += *(a++) * w;
		    *DofEdgeData(wgts, n) += w;
		}
	    }
	    flags += g->nedge;
	}

	if (newtype->np_face > 0) {
	    for (k = 0; k < NFace; k++) {
		if (flags[n = e->faces[k]] && wgts == NULL) {
		    bas += count * newtype->np_face;
		    continue;
		}
		flags[n] = TRUE;
		a = DofFaceData(dest, n);
		newtype->InitFunc(dest, e, FACE, k, NULL, func, NULL, a,
					NULL);
		if (wgts != NULL) {
		    d = buffer + (a - DofData(dest));
		    for (i = 0; i < nvalues * newtype->np_face; i++)
			*(d++) += *(a++) * w;
		    *DofFaceData(wgts, n) += w;
		}
	    }
	}

	if (newtype->np_elem > 0) {
	    a = DofElementData(dest, e->index);
	    newtype->InitFunc(dest, e, ELEMENT, 0, NULL, func, NULL, a,
					NULL);
	}
    }

    phgFree(basvalues);
    phgFree(flags0);

    if (wgts == NULL)
	return dest;

    if ((n = DofGetDataCount(dest) - DofGetElementDataCount(dest)) > 0) {
	memcpy(DofData(dest), buffer, n * sizeof(*buffer));
	phgFree(buffer);
    }

    a = DofData(dest);
    d = DofData(wgts);

    if (newtype->np_vert > 0) {
	k = nvalues * newtype->np_vert;
	for (n = 0; n < g->nvert; n++) {
	    if ((w = *(d++)) == 0.0) {
		a += k;
	    }
	    else if (k == 1) {
		*(a++) /= w;
	    }
	    else {
		w = 1.0 / w;
		for (i = 0; i < k; i++)
		    *(a++) *= w;
	    }
	}
    }

    if (newtype->np_edge > 0) {
	k = nvalues * newtype->np_edge;
	for (n = 0; n < g->nedge; n++) {
	    if ((w = *(d++)) == 0.0) {
		a += k;
	    }
	    else if (k == 1) {
		*(a++) /= w;
	    }
	    else {
		w = 1.0 / w;
		for (i = 0; i < k; i++)
		    *(a++) *= w;
	    }
	}
    }

    if (newtype->np_face > 0) {
	k = nvalues * newtype->np_face;
	for (n = 0; n < g->nface; n++) {
	    if ((w = *(d++)) == 0.0) {
		a += k;
	    }
	    else if (k == 1) {
		*(a++) /= w;
	    }
	    else {
		w = 1.0 / w;
		for (i = 0; i < k; i++)
		    *(a++) *= w;
	    }
	}
    }

    phgDofFree(&wgts);

    return dest;
}

FLOAT
phgDofNormL2Vec(DOF *x)
/* vector L2 norm */
{
    INT i, n;
    FLOAT *d = x->data;
    FLOAT a, b;

    n = DofGetDataCount(x);

    a = 0.0;
    for (i = 0; i < n; i++, d++) {
	if (!DofIsOwner(x, i))
	    continue;
	b = *d;
	a += b * b;
    }

    return Sqrt(a);
}

FLOAT
phgDofNormL1Vec(DOF *x)
/* vector L1 norm */
{
    INT i, n;
    FLOAT *d = x->data;
    FLOAT a;

    n = DofGetDataCount(x);

    a = 0.0;
    for (i = 0; i < n; i++, d++)
	if (DofIsOwner(x, i))
	    a += Fabs(*d);

    return a;
}

FLOAT
phgDofNormInftyVec(DOF *x)
{
    INT i, n;
    FLOAT *d = x->data;
    FLOAT a, b;

    n = DofGetDataCount(x);

    a = 0.0;
    for (i = 0; i < n; i++, d++) {
	if (!DofIsOwner(x, i))
	    continue;		/* skip unused vertex data */
	b = Fabs(*d);
	if (a < b)
	    a = b;
    }

    return a;
}

FLOAT
phgDofNormL2(DOF *u)
{
    SIMPLEX *e;
    FLOAT norm;

    norm = 0.0;
    ForAllElements(u->g, e)
	norm += phgQuadDofDotDof(e, u, u, QUAD_DEFAULT);

    norm = Sqrt(norm);

    return norm;
}

FLOAT
phgDofNormH1(DOF *u)
{
    SIMPLEX *e;
    FLOAT norm;
    DOF *grad_u = phgDofGradient(u, NULL, NULL, NULL);

    norm = 0.0;
    ForAllElements(u->g, e)
	norm += phgQuadDofDotDof(e, u, u, QUAD_DEFAULT) +
		phgQuadDofDotDof(e, grad_u, grad_u, QUAD_DEFAULT);

    norm = Sqrt(norm);

    phgDofFree(&grad_u);

    return norm;
}

void
phgDofRandomize(DOF *u, long int seed)
/* Note: seed == 0 ==> no seed */
{
    INT j, n;
    FLOAT *data;
    BTYPE b;

    MagicCheck(DOF, u);

    if (seed != 0)
	srand48(seed + u->g->rank);
    n = DofGetDataCount(u);
    data = DofData(u);
    for (j = 0; j < n; j++) {
	b = phgDofGetBoundaryType(u, j);
	if (b == UNREFERENCED || (b & u->DB_mask))
	    *(data++) = 0.0;
	else
	    *(data++) = 2.0 * drand48() - 1.0;
    }
}

static DOF *dofa, *dofb, *dofc;
static FLOAT *cache_bufb = NULL, *cache_bufc = NULL;
static int cache_dimb = 0, cache_dimc = 0;
static int M0, N0, K0, Kstart, Ksize, incA;
static FLOAT *bufa, *bufb, alpha0;
static MAT_OP ta, tb;

static void
mm_func(DOF *dof, SIMPLEX *e, int bno, const FLOAT *lambda, FLOAT *values)
/* temperary Init function used by phgDofMM */
{
    int i, j, k, m, k0;
    FLOAT tmp, *A, *B, *C;

    if (dofa != NULL)
	phgDofEval_(dofa, e, lambda, cache_buf + bno * cache_dim, 0,
		    bufa, NULL);
    phgDofEval_(dofb, e, lambda, cache_bufb + bno * cache_dimb, 0, bufb, NULL);
    if (dofc->type->InitFunc != phgDofInitFuncPoint) {
	phgDofEval_(dofc, e, lambda, cache_bufc + bno * cache_dimc, 0,
			values, NULL);
    }
    else for (; TRUE;) {
	INT loc;
	int map[8];
	assert(dofc->type->np_edge < 8 && dofc->type->np_face < 8);
	/* faster code: don't need to go through phgDofEval */
	if (bno < (m = dofc->type->np_vert * NVert)) {
	    loc = (e->verts[bno / dofc->type->np_vert] * dofc->type->np_vert +
		   bno % dofc->type->np_vert) * dofc->dim;
	    memcpy(values, dofc->data_vert + loc, dofc->dim * sizeof(*values));
	    break;
	}
	bno -= m;
	if (bno < (m = dofc->type->np_edge * NEdge)) {
	    i = bno / dofc->type->np_edge;
	    phgDofMapEdgeData(dofc->type, e, i, map);
	    loc = (e->edges[i] * dofc->type->np_edge +
		   map[bno % dofc->type->np_edge]) * dofc->dim;
	    memcpy(values, dofc->data_edge + loc, dofc->dim * sizeof(*values));
	    break;
	}
	bno -= m;
	if (bno < (m = dofc->type->np_face * NFace)) {
	    i = bno / dofc->type->np_face;
	    phgDofMapFaceData(dofc->type, e, i, map);
	    loc = (e->faces[i] * dofc->type->np_face +
		   map[bno % dofc->type->np_face]) * dofc->dim;
	    memcpy(values, dofc->data_face + loc, dofc->dim * sizeof(*values));
	    break;
	}
	bno -= m;
	assert(bno < dofc->type->np_elem);
	loc = (e->index * dofc->type->np_elem + bno) * dofc->dim;
	memcpy(values, dofc->data_elem + loc, dofc->dim * sizeof(*values));
	break;
    }

    /* compute c += alpha * A * b */
    A = bufa;
    B = bufb;
    C = values;
    k0 = Kstart;
    if (dofa == NULL) {
	m = DofDim(dofc);
	assert(m == DofDim(dofb));
	if (tb == MAT_OP_N || M0 == 1) {
	    if (alpha0 == 1.) {
		for (i = 0; i < m; i++)
		    *(C++) += *(B++);
	    }
	    else if (alpha0 == -1.) {
		for (i = 0; i < m; i++)
		    *(C++) -= *(B++);
	    }
	    else {
		for (i = 0; i < m; i++)
		    *(C++) += *(B++) * alpha0;
	    }
	}
	else {	/* tb == MAT_OP_T */
	    WARN_UNTESTED
	    if (alpha0 == 1.) {
		for (i = 0; i < M0; i++) {
		    for (j = 0; j < N0; j++)
			*(C++) += B[j * M0];
		    B++;
		}
	    }
	    else if (alpha0 == -1.) {
		for (i = 0; i < M0; i++) {
		    for (j = 0; j < N0; j++)
			*(C++) -= B[j * M0];
		    B++;
		}
	    }
	    else {
		for (i = 0; i < M0; i++) {
		    for (j = 0; j < N0; j++)
			*(C++) += B[j * M0] * alpha0;
		    B++;
		}
	    }
	}
    }
    else if (ta == MAT_OP_N) {
	if (tb == MAT_OP_N || M0 == 1) {
	    if (alpha0 == 1.) {
		for (i = 0; i < M0; i++) {
		    for (j = 0; j < N0; j++) {
			B = bufb + j + k0 * N0;
			tmp = A[0] * B[0];
			for (k = 1; k < Ksize; k++)
			    tmp += A[k] * B[k * N0];
			*(C++) += tmp;
		    }
		    A += incA;
		    if (M0 == K0 && i % Ksize == Ksize - 1)
			k0 += Ksize;
		}
	    }
	    else if (alpha0 == -1.) {
		for (i = 0; i < M0; i++) {
		    for (j = 0; j < N0; j++) {
			B = bufb + j + k0 * N0;
			tmp = A[0] * B[0];
			for (k = 1; k < Ksize; k++)
			    tmp += A[k] * B[k * N0];
			*(C++) -= tmp;
		    }
		    A += incA;
		    if (M0 == K0 && i % Ksize == Ksize - 1)
			k0 += Ksize;
		}
	    }
	    else {
		for (i = 0; i < M0; i++) {
		    for (j = 0; j < N0; j++) {
			B = bufb + j + k0 * N0;
			tmp = A[0] * B[0];
			for (k = 1; k < Ksize; k++)
			    tmp += A[k] * B[k * N0];
			*(C++) += tmp * alpha0;
		    }
		    A += incA;
		    if (M0 == K0 && i % Ksize == Ksize - 1)
			k0 += Ksize;
		}
	    }
	}
	else {	/* tb == MAT_OP_T */
	    if (alpha0 == 1.) {
		for (i = 0; i < M0; i++) {
		    for (j = 0; j < N0; j++) {
			B = bufb + j * K0 + k0;
			tmp = A[0] * B[0];
			for (k = 1; k < Ksize; k++)
			    tmp += A[k] * B[k];
			*(C++) += tmp;
		    }
		    A += incA;
		    if (M0 == K0 && i % Ksize == Ksize - 1)
			k0 += Ksize;
		}
	    }
	    else if (alpha0 == -1.) {
		for (i = 0; i < M0; i++) {
		    for (j = 0; j < N0; j++) {
			B = bufb + j * K0 + k0;
			tmp = A[0] * B[0];
			for (k = 1; k < Ksize; k++)
			    tmp += A[k] * B[k];
			*(C++) -= tmp;
		    }
		    A += incA;
		    if (M0 == K0 && i % Ksize == Ksize - 1)
			k0 += Ksize;
		}
	    }
	    else {
		for (i = 0; i < M0; i++) {
		    for (j = 0; j < N0; j++) {
			B = bufb + j * K0 + k0;
			tmp = A[0] * B[0];
			for (k = 1; k < Ksize; k++)
			    tmp += A[k] * B[k];
			*(C++) += tmp * alpha0;
		    }
		    A += incA;
		    if (M0 == K0 && i % Ksize == Ksize - 1)
			k0 += Ksize;
		}
	    }
	}
    }
    else {	/* ta == MAT_OP_T */
	if (tb == MAT_OP_N || M0 == 1) {
	    if (alpha0 == 1.) {
		for (i = 0; i < M0; i++) {
		    for (j = 0; j < N0; j++) {
			B = bufb + j + k0 * N0;
			tmp = A[0] * B[0];
			for (k = 1; k < Ksize; k++)
			    tmp += A[k * incA] * B[k * N0];
			*(C++) += tmp;
		    }
		    A++;
		    if (M0 == K0 && i % Ksize == Ksize - 1)
			k0 += Ksize;
		}
	    }
	    else if (alpha0 == -1.) {
		for (i = 0; i < M0; i++) {
		    for (j = 0; j < N0; j++) {
			B = bufb + j + k0 * N0;
			tmp = A[0] * B[0];
			for (k = 1; k < Ksize; k++)
			    tmp += A[k * incA] * B[k * N0];
			*(C++) -= tmp;
		    }
		    A++;
		    if (M0 == K0 && i % Ksize == Ksize - 1)
			k0 += Ksize;
		}
	    }
	    else {
		for (i = 0; i < M0; i++) {
		    for (j = 0; j < N0; j++) {
			B = bufb + j + k0 * N0;
			tmp = A[0] * B[0];
			for (k = 1; k < Ksize; k++)
			    tmp += A[k * incA] * B[k * N0];
			*(C++) += tmp * alpha0;
		    }
		    A++;
		    if (M0 == K0 && i % Ksize == Ksize - 1)
			k0 += Ksize;
		}
	    }
	}
	else {	/* tb == MAT_OP_T */
	    if (alpha0 == 1.) {
		for (i = 0; i < M0; i++) {
		    for (j = 0; j < N0; j++) {
			B = bufb + j * K0 + k0;
			tmp = A[0] * B[0];
			for (k = 1; k < Ksize; k++)
			    tmp += A[k * incA] * B[k];
			*(C++) += tmp;
		    }
		    A++;
		    if (M0 == K0 && i % Ksize == Ksize - 1)
			k0 += Ksize;
		}
	    }
	    else if (alpha0 == -1.) {
		for (i = 0; i < M0; i++) {
		    for (j = 0; j < N0; j++) {
			B = bufb + j * K0 + k0;
			tmp = A[0] * B[0];
			for (k = 1; k < Ksize; k++)
			    tmp += A[k * incA] * B[k];
			*(C++) -= tmp;
		    }
		    A++;
		    if (M0 == K0 && i % Ksize == Ksize - 1)
			k0 += Ksize;
		}
	    }
	    else {
		for (i = 0; i < M0; i++) {
		    for (j = 0; j < N0; j++) {
			B = bufb + j * K0 + k0;
			tmp = A[0] * B[0];
			for (k = 1; k < Ksize; k++)
			    tmp += A[k * incA] * B[k];
			*(C++) += tmp * alpha0;
		    }
		    A++;
		    if (M0 == K0 && i % Ksize == Ksize - 1)
			k0 += Ksize;
		}
	    }
	}
    }
}

DOF *
phgDofMM_(MAT_OP transa, MAT_OP transb, int M, int N, int K,
	  FLOAT alpha, DOF *A, int blka, DOF *B, FLOAT beta, DOF **Cptr,
	  const char *srcfile, int srcline)
/******************************************************************************
 * DGEMM-like matrix-matrix operation between individual DOF:
 *
 *	C := alpha * op(A) * op(B) + beta * C
 *
 * (array dimensions: M K x K N = M N)
 *
 * Parameters:
 *
 *  transa - if transa == MAT_OP_T then op(A) = A', otherwise op(A) = A
 *  transb - if transb == MAT_OP_T then op(B) = B', otherwise op(B) = B
 *  M -	specifies  the number of rows of the matrix op(A) and of the matrix C
 *  N -	specifies the number of columns of the matrix op(B) and the number of
 *	columns of the matrix C
 *  K -	specifies the number of columns of the matrix op(A) and the number of
 *	rows of the matrix op(B)
 *  A -	array of dimension [M][K] when transa == MAT_OP_N, and of dimension
 *	[K][M] otherwise
 *  blka - blocking size for A. The following conditions must be satisfied:
 *	(1) If blka <= 0 then (A == NULL and M == K) or DofDim(A) == M * K.
 *	(2) If blka > 0 then M == K and M % blka == 0 and
 *	    DofDim(A) % (blka * blka) == 0 and DofDim(A) / blka == M,
 *	    in this case A is interpreted as a block diagonal matrix with
 *	    M/blka blocks and block size blka,
 *  B -	array of dimension [K][N] when transb == MAT_OP_N, and of dimension
 *	[N][K] otherwise
 *  C -	array of dimension [M][N]
 *
 * Note: all arrays are stored in C (i.e., row-major) order
 *****************************************************************************/
{
    DOF *C = (Cptr == NULL ? NULL : *Cptr);
    INT i, m;
    int ii, jj;
    FLOAT *p, *q;

    MagicCheck(DOF, A);
    MagicCheck(DOF, B);
    MagicCheck(DOF, C);

    assert(A != NULL || M == K);
    assert(A == NULL || (blka <= 0 && DofDim(A) == M * K) ||
	   (blka > 0 && M == K && DofDim(A) == M * blka && M % blka == 0));
    assert(B == NULL || DofDim(B) == K * N);
    assert(C == NULL || DofDim(C) == M * N);

    if (C == NULL) {
	assert(B != NULL && (M * N) % B->type->dim == 0);
	beta = 0.L;
	C = phgDofNew_(B->g, B->type, M * N / B->type->dim, "C", DofNoAction,
			srcfile, srcline);
	C->DB_mask = B->DB_mask;
	if (Cptr != NULL)
	    *Cptr = C;
    }
    else if (beta == 0.) {
	phgDofSetDataByValue(C, (FLOAT)0.);
    }
    else if (beta == -1.) {
	m = DofGetDataCount(C);
	p = DofData(C);
	for (i = 0; i < m; i++, p++)
	    *p = - *p;
    }
    else if (beta != 1.) {
	m = DofGetDataCount(C);
	p = DofData(C);
	for (i = 0; i < m; i++)
	    *(p++) *= beta;
    }

    C->userfunc = DofNoAction;

    if (alpha == 0.)
	return C;

    assert(B != NULL);

    if (A == NULL && B->type == C->type) {
	m = DofGetDataCount(C);
	p = DofData(B);
	q = DofData(C);
	if (transb == MAT_OP_N) {
	    if (alpha == 1.) {
		for (i = 0; i < m; i++)
		    *(q++) += *(p++);
	    }
	    else if (alpha == -1.) {
		for (i = 0; i < m; i++)
		    *(q++) -= *(p++);
	    }
	    else {
		for (i = 0; i < m; i++)
		    *(q++) += *(p++) * alpha;
	    }
	}
	else {
	    WARN_UNTESTED
	    m /= N * M;
	    if (alpha == 1.) {
		for (i = 0; i < m; i++) {
		    for (ii = 0; ii < M; ii++) {
			for (jj = 0; jj < N; jj++)
			    *(q++) += p[jj * M];
			p++;
		    }
		}
	    }
	    else if (alpha == -1.) {
		for (i = 0; i < m; i++) {
		    for (ii = 0; ii < M; ii++) {
			for (jj = 0; jj < N; jj++)
			    *(q++) -= p[jj * M];
			p++;
		    }
		}
	    }
	    else {
		for (i = 0; i < m; i++) {
		    for (ii = 0; ii < M; ii++) {
			for (jj = 0; jj < N; jj++)
			    *(q++) += p[jj * M] * alpha;
			p++;
		    }
		}
	    }
	}
	return C;
    }

    /* pre-evaluation of basis functions at nodal points of C */
    if (C->type->points != NULL) {
	ii = C->type->nbas;
	if (A != NULL && A->type->BasFuncs != NULL && A->type->invariant) {
	    cache_dim = A->type->nbas * A->type->dim;
	    cache_buf = phgAlloc(cache_dim * ii * sizeof(*cache_buf));
	    get_bas_funcs(A, C, C->g->roots, cache_buf);
	}
	if (B->type->BasFuncs != NULL && B->type->invariant) {
	    cache_dimb = B->type->nbas * B->type->dim;
	    cache_bufb = phgAlloc(cache_dimb * ii * sizeof(*cache_bufb));
	    get_bas_funcs(B, C, C->g->roots, cache_bufb);
	}
	if (C->type->BasFuncs != NULL && C->type->invariant &&
		C->type->InitFunc != phgDofInitFuncPoint) {
	    cache_dimc = C->type->nbas * C->type->dim;
	    cache_bufc = phgAlloc(cache_dimc * ii * sizeof(*cache_bufc));
	    get_bas_funcs(C, C, C->g->roots, cache_bufc);
	}
    }

    M0 = M;
    N0 = N;
    K0 = K;
    if (blka == M && M == K)
	blka = 0;
    if (blka <= 0) {
	Kstart = 0;
	Ksize = K;
	incA = (transa == MAT_OP_N ? K : M);
    }
    else {
	Kstart = 0;
	Ksize = blka;
	incA = blka;
    }
    alpha0 = alpha;
    ta = transa;
    tb = transb;
    bufa = (A == NULL ? NULL : phgAlloc(DofDim(A) * sizeof(*bufa)));
    bufb = phgAlloc(DofDim(B) * sizeof(*bufb));
    dofa = A;
    dofb = B;
    dofc = phgDofCopy(C, NULL, NULL, NULL);
    phgDofSetDataByFunction_(C, NULL, mm_func, NULL);
    phgDofFree(&dofc);
    phgFree(bufa);
    phgFree(bufb);

    if (cache_buf != NULL) {
	phgFree(cache_buf);
	cache_buf = NULL;
	cache_dim = 0;
    }
    if (cache_bufb != NULL) {
	phgFree(cache_bufb);
	cache_bufb = NULL;
	cache_dimb = 0;
    }
    if (cache_bufc != NULL) {
	phgFree(cache_bufc);
	cache_bufc = NULL;
	cache_dimc = 0;
    }

    return C;
}

DOF *
phgDofAXPBY_(FLOAT a, DOF *x0, FLOAT b, DOF **y_ptr,
			const char *srcfile, int srcline)
/* y := a * x + b * y.
 * Note: only works when dofvalues are linear function of funcvalues
 * otherwise need to evaluate function values using dof->type->InitFunc
 */
{
#if 0
    int m = DofDim(x0);
    return phgDofMM_(MAT_OP_N, MAT_OP_N, m, 1, m, a, NULL, 0,
					x0, b, y_ptr, srcfile, srcline);
#else
    DOF *x = x0, *y = (y_ptr == NULL ? NULL : *y_ptr);
    INT i, n = 0;
    FLOAT *p;

    MagicCheck(DOF, y);

    assert(y == NULL || y->type != DOF_ANALYTIC);

    if (y == NULL) {
	assert(x0 != NULL && x0->type != DOF_ANALYTIC);
	b = 0.L;
	y = phgDofNew_(x0->g, x0->type, x0->dim, NULL, DofNoAction,
							srcfile, srcline);
	y->DB_mask = x0->DB_mask;
	if (y_ptr != NULL)
	    *y_ptr = y;
    }
    else if (b == 0.) {
	phgDofSetDataByValue(y, (FLOAT)0.);
    }
    else if (b == -1.) {
	n = DofGetDataCount(y);
	p = DofData(y);
	for (i = 0; i < n; i++, p++)
	    *p = - *p;
    }
    else if (b != 1.) {
	n = DofGetDataCount(y);
	p = DofData(y);
	for (i = 0; i < n; i++)
	    *(p++) *= b;
    }

    y->userfunc = DofNoAction;

    if (a == 0.0)
	return y;			/* nothing to do */

    if (y->type != x->type)
	x = phgDofCopy(x0, NULL, y->type, "AXPY tmp");

    if (y->dim != x->dim)
	phgError(1, "%s:%d: phgDofAXPY: incompatible DOFs \"%s\" <-> \"%s\".\n",
			srcfile, srcline, x->name, y->name);

    if ((n = DofGetDataCount(y)) == 0) {
	if (x != x0)
	    phgDofFree(&x);
	return y;
    }

    if (a == -1.0) {		/* y := y - x */
	for (i = 0; i < n; i++) {
	    if (phgDofGetBoundaryType(y, i) == UNREFERENCED)
		continue;
	    y->data[i] -= x->data[i];
	}
    }
    else if (a == 1.0) {	/* y := y + x */
	for (i = 0; i < n; i++) {
	    if (phgDofGetBoundaryType(y, i) == UNREFERENCED)
		continue;
	    y->data[i] += x->data[i];
	}
    }
    else {			/* y := y + a * x */
	for (i = 0; i < n; i++) {
	    if (phgDofGetBoundaryType(y, i) == UNREFERENCED)
		continue;
	    y->data[i] += a * x->data[i];
	}
    }

    if (x != x0)
	phgDofFree(&x);

    return y;
#endif
}

DOF *
phgDofMatVec_(FLOAT alpha, DOF *A, DOF *x, FLOAT beta, DOF **y_ptr,
		const char *srcfile, int srcline)
/* BLAS-like function computing y := alpha A x + beta y */
{
    DOF *y = (y_ptr == NULL ? NULL : *y_ptr);
    INT M;

    if (A == NULL || alpha == 0.)
	return phgDofAXPBY_(alpha, x, beta, y_ptr, srcfile, srcline);

    assert(x != NULL);
    M = DofDim(x);
    assert(y == NULL || M == DofDim(y));
    if (DofDim(A) == 1)
	return phgDofMM_(MAT_OP_N, MAT_OP_N, 1, M, 1,
				alpha, A, 0, x, beta, y_ptr, srcfile, srcline);
    if (DofDim(A) == M)
	return phgDofMM_(MAT_OP_N, MAT_OP_N, M, M, M,
				alpha, A, 1, x, beta, y_ptr, srcfile, srcline);
    assert(DofDim(A) == M * M); 
    return phgDofMM_(MAT_OP_N, MAT_OP_N, M, M, M,
			alpha, A, 0, x, beta, y_ptr, srcfile, srcline);
}
