/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: matvec.c,v 1.142 2007/04/13 07:16:22 zlb Exp $ */

/* Functions for basic mat/vec operations. */

#define NEED_GetVariableArgs

#include "phg.h"
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <math.h>
#include <stdarg.h>

#define vec_compare(v1,v2)  \
  if((v1)->map->nlocal !=(v2)->map->nlocal) \
    phgError(1,"%s:%d the size of the vectors mismatch.",__FILE__,__LINE__);

FLOAT
phgVecNorm2(VEC *src, int which, FLOAT *res)
{
    INT i;
    FLOAT v;

    assert(src->nvec == 1);
    assert(which < src->nvec);

    v = 0.0;
    for (i = 0; i < src->map->nlocal; i++)
	v += src->data[i] * src->data[i];
    v = Sqrt(v);
    if (res != NULL)
	*res = v;

    return v;
}

FLOAT
phgVecNorm1(VEC *src, int which, FLOAT *res)
{
    INT i;
    FLOAT v;

    assert(src->nvec == 1);

    v = 0.0;
    for (i = 0; i < src->map->nlocal; i++)
	v += Fabs(src->data[i]);
    if (res != NULL)
	*res = v;

    return v;
}

FLOAT
phgVecNormInfty(VEC *src, int which, FLOAT *res)
{
    INT i;
    FLOAT v;
    FLOAT cmp;

    assert(src->nvec == 1);

    v = 0.0;
    for (i = 0; i < src->map->nlocal; i++) {
	cmp = Fabs(src->data[i]);
	if (cmp > v)
	    v = cmp;
    }
    if (res != NULL)
	*res = v;

    return v;
}

VEC *
phgVecCopy(VEC *src, VEC **dest_ptr)
{
    VEC *dest = (dest_ptr == NULL ? NULL : *dest_ptr);

    assert(src->nvec == 1);

    if (dest == NULL) {
	dest = phgMapCreateVec(src->map, src->nvec);
	if (dest_ptr != NULL)
	    *dest_ptr = dest;
    }

    MagicCheck(VEC, dest)

    if (dest->nvec * dest->map->nlocal != src->nvec * src->map->nlocal) {
	phgFree(dest->data);
	phgAlloc(dest->map->nlocal * sizeof(*dest->data));
    }
    memcpy(dest->data, src->data, src->map->nlocal * sizeof(*src->data));
    dest->nvec = src->nvec;

    return dest;
}

void
phgMatPack(MAT *mat)
/* packs the matrix and sets up communications for mat-vec operations */
{
    INT i, j, n, *pc, O2Gsize;
    FLOAT *pd;
    MAT_ROW *row;
    size_t nnz, nnz_d, nnz_o;

    if (!mat->assembled)
	phgMatAssemble(mat);

    if (mat->type != PHG_UNPACKED)
	return;

    phgInfo(1, "packing matrix rows.\n");
    mat->type = PHG_PACKED;

    nnz_d = nnz_o = 0;
    for (i = 0, row = mat->rows; i < mat->rmap->nlocal; i++, row++) {
	nnz_d += row->ncols;
    }
    nnz = nnz_d + nnz_o;

    mat->nnz_d = nnz_d;
    mat->nnz_o = nnz_o;

    j = mat->rmap->nlocal;
    mat->packed_cols = phgAlloc((j + nnz) * sizeof(*mat->packed_cols));
    mat->packed_data = phgAlloc(nnz * sizeof(*mat->packed_data));

    /* convert indices and pack data */
    if (mat->cmap->nprocs > 1)
	O2Gsize = mat->localsize - mat->cmap->nlocal;
    else
	O2Gsize = 0;

    pc = mat->packed_cols;
    pd = mat->packed_data;
    for (i = 0, row = mat->rows; i < mat->rmap->nlocal; i++, row++) {
	n = row->ncols;
	*(pc++) = n;
	memcpy(pc, row->cols, n * sizeof(*row->cols));
	memcpy(pd, row->data, n * sizeof(*row->data));
	pc += n;
	pd += n;
	phgFree(row->cols);
	phgFree(row->data);
    }
    assert(pc == mat->packed_cols + mat->rmap->nlocal + mat->nnz_d);
    assert(pd == mat->packed_data + mat->nnz_d);

    phgFree(mat->rows);
    mat->rows = NULL;

}

void
phgMatUnpack(MAT *mat)
/* unpacks matrix */
{
    INT i, n, *pc;
    MAT_ROW *row;
    FLOAT *pd;

    if (mat->type != PHG_PACKED)
	return;

    phgInfo(1, "unpacking matrix rows.\n");
    mat->type = PHG_UNPACKED;

    mat->rows = phgAlloc(mat->rmap->localsize * sizeof(*mat->rows));
    if ((n = mat->rmap->localsize - mat->rmap->nlocal) > 0)
	bzero(mat->rows + mat->rmap->nlocal, n * sizeof(*mat->rows));
    pc = mat->packed_cols;
    pd = mat->packed_data;
    for (i = 0, row = mat->rows; i < mat->rmap->nlocal; i++, row++) {
	n = *(pc++);
	row->alloc = row->ncols = n;
	row->cols = phgAlloc(n * sizeof(*row->cols));
	row->data = phgAlloc(n * sizeof(*row->data));
	memcpy(row->data, pd, n * sizeof(*row->data));
	memcpy(row->cols, pc, n * sizeof(*row->cols));
	pd += n;
	pc += n;
    }
    assert(pc == mat->packed_cols + mat->rmap->nlocal + mat->nnz_d);
    assert(pd == mat->packed_data + mat->nnz_d);

    phgFree(mat->packed_cols);
    mat->packed_cols = NULL;
    phgFree(mat->packed_data);
    mat->packed_data = NULL;

}

void
phgMatSetupDiagonal(MAT *mat)
{
    INT i, k, n, *pc;
    MAT_ROW *row;
    FLOAT *pd;

    if (mat->diag != NULL)
	return;

    assert(mat->type != PHG_DESTROYED);

    mat->diag = phgAlloc(mat->rmap->nlocal * sizeof(*mat->diag));

    if (mat->type == PHG_MATRIX_FREE) {
	/* get the diagonal by multiplying mat with the vector [1 ... 1]^T */
	VEC *x, *y;
	phgFree(mat->diag);
	x = phgMapCreateVec(mat->rmap, 1);
	for (i = 0; i < x->map->nlocal; i++)
	    x->data[i] = 1.0;
	y = phgMatVec(MAT_OP_D, 1.0, mat, x, 0.0, NULL);
	mat->diag = y->data;
	y->data = NULL;
	phgVecDestroy(&x);
	phgVecDestroy(&y);
    }
    else if (mat->type != PHG_PACKED) {
	for (i = 0, row = mat->rows; i < mat->rmap->nlocal; i++, row++) {
	    for (k = 0; k < row->ncols; k++) {
		if (row->cols[k] == i)
		    break;
	    }
	    mat->diag[i] = (k < row->ncols ? row->data[k] : 0.0);
	}
    }
    else {
	pc = mat->packed_cols;
	pd = mat->packed_data;
	for (i = 0; i < mat->rmap->nlocal; i++) {
	    n = *(pc++);
	    for (k = 0; k < n; k++) {
		if (pc[k] == i)
		    break;
	    }
	    mat->diag[i] = (k < n ? pd[k] : 0.0);
	    pc += n;
	    pd += n;
	}
    }
}

VEC *
phgMatVec(MAT_OP op, FLOAT alpha, MAT *A, VEC *x, FLOAT beta, VEC **y_ptr)
/* BLAS gemv like function computing y := alpha op(A) x + beta y.
 * Special values of alpha and beta are taken into account to reduce
 * the number of computations. if A == NULL, then A is considered
 * to be the identity matrix I of size x->nglobal.
 *
 * op should be one of:
 *	MAT_OP_N (multiply x by A),
 *	MAT_OP_T (multiply x by trans(A)),
 *	MAT_OP_D (multiply x by the diagonal part of A)
 *
 * This function can do many kinds of matrix-vector or vector-vector
 * operations, for example:
 *
 *  phgMatVec(MAT_OP_N, 1.0, NULL, x, 0.0, &y)       . y := x  (vector copy)
 *  phgMatVec(MAT_OP_N, -1.0, NULL, x, 0.0, &y)	     . y := - x
 *  phgMatVec(MAT_OP_N, alpha, NULL, x, 0.0, &y)     . y := alpha x
 *  phgMatVec(MAT_OP_N, 0.0, NULL, NULL, -1.0, &y)   . y := - y
 *  phgMatVec(MAT_OP_N, 0.0, NULL, NULL, beta, &y)   . y := beta y
 *  phgMatVec(MAT_OP_N, alpha, NULL, x, beta, &y)    . y := alpha x + beta y
 *  phgMatVec(MAT_OP_N, 1.0, A, x, 0.0, &y)	     . y := A x
 *  phgMatVec(MAT_OP_T, 1.0, A, x, 0.0, &y)          . y := A' x
 *  phgMatVec(MAT_OP_N, -1.0, A, x, 0.0, &y)	     . y := - A x
 */
{

    INT i, j, n, *pc;
    FLOAT *data, *pd, *v, *v0;
    VEC *y = (y_ptr == NULL ? NULL : *y_ptr);

    MagicCheck(VEC, y)

    if (alpha != 0.0) {
	if (x == NULL)
	    phgError(1,
		     "phgMatVec, %s:%d, x must be non NULL when alpha!=0\n",
		     __FILE__, __LINE__);
    }
    else {
	if (y == NULL)
	    phgError(1,
		     "phgMatVec, %s:%d, y must be non NULL when alpha==0\n",
		     __FILE__, __LINE__);
    }

    assert(y == NULL || y->nvec == 1);
    assert(x == NULL || x->nvec == 1);

    /* multiply y by beta */
    if (alpha == 0.0 || A != NULL) {
	if (y == NULL) {
	    /* x must be non NULL */
	    if (A != NULL) {
		if (op != MAT_OP_T)
		    y = phgMapCreateVec(A->rmap, x->nvec);
		else
		    y = phgMapCreateVec(A->cmap, x->nvec);
	    }
	    else
		y = phgMapCreateVec(x->map, x->nvec);

	    if (y_ptr != NULL)
		*y_ptr = y;
	    beta = 0.0;
	}
	else {
	    /* compute y := beta y */
	    if (beta == 0.0) {
		bzero(y->data, y->map->nlocal * sizeof(*y->data));
	    }
	    else if (beta == (FLOAT)(-1.0)) {
		for (i = 0, v = y->data; i < y->map->nlocal; i++, v++)
		    *v = -(*v);
	    }
	    else if (beta != (FLOAT)1.0) {
		for (i = 0, v = y->data; i < y->map->nlocal; i++)
		    *(v++) *= beta;
	    }
	}

	if (alpha == 0.0)
	    return y;
    }
    else {			/* alpha != 0 && A == NULL */
	if (y == NULL) {
	    y = phgMapCreateVec(x->map, x->nvec);
	    if (y_ptr != NULL)
		*y_ptr = y;
	    beta = 0.0;
	}
	else {
	    vec_compare(x, y);
	}
	if (alpha == 1.0) {
	    if (beta == 0.0) {	/* y := x */
		memcpy(y->data, x->data, y->map->nlocal * sizeof(y->data));
	    }
	    else if (beta == (FLOAT)(-1.0)) {	/* y := x - y */
		v0 = x->data;
		for (i = 0, v = y->data; i < y->map->nlocal; i++, v++)
		    *v = *(v0++) - (*v);
	    }
	    else if (beta == (FLOAT)1.0) {	/* y := x + y */
		for (i = 0, v = y->data, v0 = x->data; i < y->map->nlocal; i++)
		    *(v++) += (*(v0++));
	    }
	    else {		/* y := x + beta y */
		for (i = 0, v = y->data, v0 = x->data; i < y->map->nlocal;
		     i++, v++)
		    *v = *(v0++) + beta * (*v);
	    }
	}
	else if (alpha == -1.0) {
	    if (beta == 0.0) {	/* y := - x */
		for (i = 0, v = y->data, v0 = x->data; i < y->map->nlocal; i++)
		    *(v++) = -*(v0++);
	    }
	    else if (beta == (FLOAT)(-1.0)) {	/* y := - x - y */
		v0 = x->data;
		for (i = 0, v = y->data; i < y->map->nlocal; i++, v++)
		    *v = -*(v0++) - (*v);
	    }
	    else if (beta == (FLOAT)1.0) {	/* y := - x + y */
		for (i = 0, v = y->data, v0 = x->data; i < y->map->nlocal; i++)
		    *(v++) += -*(v0++);
	    }
	    else {		/* y := - x + beta y */
		for (i = 0, v = y->data, v0 = x->data; i < y->map->nlocal;
		     i++, v++)
		    *v = -*(v0++) + beta * (*v);
	    }
	}
	else {
	    if (beta == 0.0) {	/* y := alpha x */
		for (i = 0, v = y->data, v0 = x->data; i < y->map->nlocal; i++)
		    *(v++) = alpha * *(v0++);
	    }
	    else if (beta == (FLOAT)(-1.0)) {	/* y := alpha x - y */
		for (i = 0, v = y->data, v0 = x->data; i < y->map->nlocal;
		     i++, v++)
		    *v = alpha * *(v0++) - (*v);
	    }
	    else if (beta == (FLOAT)1.0) {	/* y := alpha x + y */
		for (i = 0, v = y->data, v0 = x->data; i < y->map->nlocal; i++)
		    *(v++) += alpha * (*(v0++));
	    }
	    else {		/* y := alpha x + beta y */
		for (i = 0, v = y->data, v0 = x->data; i < y->map->nlocal;
		     i++, v++)
		    *v = alpha * *(v0++) + beta * (*v);
	    }
	}

	return y;
    }

    /* now we must have alpha != 0 and A != NULL */
    assert(A->type != PHG_DESTROYED);

    if (A->type == PHG_MATRIX_FREE) {
	VEC *b;
	int ret;
	b = phgMapCreateVec(y->map, y->nvec);
	ret = A->mv_func(op, A, x, b);
	if (ret != 0)
	    phgError(1, "phgMatVec: nonzero return (%d) from mv_func, abort.\n",
			ret);
	phgVecAXPBY(alpha, b, 1.0, &y);
	phgVecDestroy(&b);

	return y;
    }

    if (!A->assembled)
	phgMatAssemble(A);

    if (op == MAT_OP_D) {
	assert(x->map->nlocal == y->map->nlocal);
	if (A->diag == NULL)
	    phgMatSetupDiagonal(A);
	pd = A->diag;
	v0 = x->data;
	v = y->data;
	if (alpha == 1.0) {
	    for (i = 0; i < y->map->nlocal; i++)
		*(v++) += *(v0++) * *(pd++);
	}
	else if (alpha == -1.0) {
	    for (i = 0; i < y->map->nlocal; i++)
		*(v++) -= *(v0++) * *(pd++);
	}
	else {
	    for (i = 0; i < y->map->nlocal; i++)
		*(v++) += alpha * *(v0++) * *(pd++);
	}

	return y;
    }

    phgMatPack(A);

    if (alpha == 1.0) {
	data = x->data;
    }
    else if (alpha == -1.0) {
	v = data = phgAlloc(x->map->nlocal * sizeof(*data));
	v0 = x->data;
	for (i = 0; i < A->cmap->nlocal; i++)
	    *(v++) = -*(v0++);
    }
    else {
	v = data = phgAlloc(x->map->nlocal * sizeof(*data));
	v0 = x->data;
	for (i = 0; i < A->cmap->nlocal; i++)
	    *(v++) = alpha * *(v0++);
    }

    if (op == MAT_OP_N) {	/* y += alpha * A * x */
	if (A->cmap->nlocal != x->map->nlocal ||
	    A->rmap->nlocal != y->map->nlocal)
	    phgError(1, "%s:%d: inconsistent matrix-vector.", __FILE__,
		     __LINE__);

	/* multiply with local data */
	pc = A->packed_cols;
	pd = A->packed_data;
	for (i = 0, v = y->data; i < A->rmap->nlocal; i++) {
	    if ((n = *(pc++)) == 0) {
		v++;
		continue;
	    }
	    beta = *(pd++) * data[*(pc++)];
	    for (j = 1; j < n; j++)
		beta += *(pd++) * data[*(pc++)];
	    *(v++) += beta;
	}

    }
    else if (op == MAT_OP_T) {		/* y += alpha * trans(A) * x */
	if (A->cmap->nlocal != y->map->nlocal ||
	    A->cmap->nglobal != y->map->nglobal ||
	    A->rmap->nlocal != x->map->nlocal ||
	    A->rmap->nglobal != x->map->nglobal)
	    phgError(1, "%s:%d: inconsistent mat-vec.", __FILE__, __LINE__);

	/* in-process entries */
	pc = A->packed_cols;
	pd = A->packed_data;
	for (i = 0, v = x->data; i < A->rmap->nlocal; i++) {
	    if ((n = *(pc++)) == 0) {
		v++;
		continue;
	    }
	    beta = *(v++);
	    for (j = 0; j < n; j++) {
		y->data[*(pc++)] += *(pd++) * beta;
	    }
	}

    }
    else {
	phgInfo(-1, "%s: op must be one of MAT_OP_N (%d), MAT_OP_T (%d) and "
		    "MAT_OP_D (%d)", __func__, MAT_OP_N, MAT_OP_T, MAT_OP_D);
	phgError(1, "%s: invalid MAT operation %d!\n", __func__, op);
    }

    if (data != x->data)
	phgFree(data);

    return y;
}

FLOAT
phgVecDot(VEC *x, int which_x, VEC *y, int which_y, FLOAT *res)
{
    INT i;
    FLOAT v;

    assert(x->nvec == 1 && y->nvec == 1 && x->map->nlocal == y->map->nlocal);
    assert(which_x < x->nvec && which_y < y->nvec);

    vec_compare(x, y);
    v = 0.0;
    for (i = 0; i < x->map->nlocal; i++)
	v += x->data[i] * y->data[i];
    if (res != NULL)
	*res = v;

    return v;
}

VEC *
phgVecAXPBY(FLOAT a, VEC *x, FLOAT b, VEC **y_ptr)
{
    int i;
    VEC *y = (y_ptr == NULL ? NULL : *y_ptr);

    MagicCheck(VEC, y)

    /* TODO: allow x == NULL if a == 0, and allow y == NULL if b == 0 */
    assert(x != NULL && y != NULL);
    assert(x->nvec == 1 && y->nvec == 1 && x->map->nlocal == y->map->nlocal);

    vec_compare(x, y);
    for (i = 0; i < x->map->nlocal; i++)
	y->data[i] = a * x->data[i] + b * y->data[i];

    return y;
}

VEC *
phgVecCreate(GRID *g, INT m, INT M, int nvec)
{
    MAP *map = phgMapCreateSimpleMap(g, m, M);
    VEC *v = phgMapCreateVec(map, nvec);

    map->ref_count--;

    return v;
}

void
phgVecDestroy(VEC **vec_ptr)
{
    VEC *vec = *vec_ptr;

    MagicCheck(VEC, vec)

    if (vec == NULL)
	return;

    phgFree(vec->data);
    phgFree(vec->offp_data);
    phgMapDestroy(&vec->map);
    phgFree(vec);

    *vec_ptr = NULL;
    return;
}

MAT *
phgMatCreate(GRID *g, INT m, INT M)
{
    MAP *map = phgMapCreateSimpleMap(g, m, M);

    map->ref_count--;

    return phgMapCreateMat(map, map);
}

MAT *
phgMatCreateNonSquare(GRID *g, INT m, INT M, INT n, INT N)
{
    MAP *rmap = phgMapCreateSimpleMap(g, m, M);
    MAP *cmap = phgMapCreateSimpleMap(g, n, N);

    rmap->ref_count--;
    cmap->ref_count--;

    return phgMapCreateMat(rmap, cmap);
}

void
_phg_matrix_free_setup(MAT *mat, MV_FUNC mv_func, int ndata, void **data)
{

    if (ndata > 0) {
	mat->mv_data = phgAlloc((ndata + 1) * sizeof(*mat->mv_data));
	memcpy(mat->mv_data, data, ndata * sizeof(*data));
	mat->mv_data[ndata] = NULL;
    }

    phgFree(mat->rows);
    mat->rows = NULL;
    mat->mv_func = mv_func;
    mat->type = PHG_MATRIX_FREE;
    mat->assembled = TRUE;
}

MAT *
phgMatCreateMatrixFree(GRID *g, INT m, INT M, MV_FUNC mv_func, void *mv_data,
			...)
{
    MAT *mat = phgMatCreate(g, m, M);
    int ndata;
    void **data;

    if (mat == NULL)
	return NULL;

    GetVariableArgs(ndata, data, mv_data, void *);
    _phg_matrix_free_setup(mat, mv_func, ndata, data);
    phgFree(data);

    return mat;
}

MAT *
phgMatCreateMatrixFreeNonSquare(GRID *g, INT m, INT M, INT n, INT N,
			MV_FUNC mv_func, void *mv_data, ...)
{
    MAT *mat = phgMatCreateNonSquare(g, m, M, n, N);
    int ndata;
    void **data;

    if (mat == NULL)
	return NULL;

    GetVariableArgs(ndata, data, mv_data, void *);
    _phg_matrix_free_setup(mat, mv_func, ndata, data);
    phgFree(data);

    return mat;
}

static void
free_matrix_rows(INT size, MAT_ROW **rows)
{
    MAT_ROW *row;

    if (*rows == NULL)
	return;

    /* free PHG's matrix */
    for (row = *rows; row < *rows + size; row++) {
	/* free PHG matrix row */
	phgFree(row->cols);
	phgFree(row->data);
	row->cols = NULL;
	row->data = NULL;
	row->ncols = row->alloc = 0;
    }
    phgFree(*rows);
    *rows = NULL;
}

static void
free_bdry_eqns(MAT *mat)
{
    INT i;

    if (mat->bdry_eqns == NULL)
	return;

    for (i = 0; i < mat->bdry_eqns_count; i++) {
	phgFree(mat->bdry_eqns[i].LU);
	phgFree(mat->bdry_eqns[i].pvt);
	phgFree(mat->bdry_eqns[i].map);
    }
    phgFree(mat->bdry_eqns);
    mat->bdry_eqns = NULL;
    mat->bdry_eqns_count = mat->bdry_eqns_allocated = 0;
}

void
phgMatFreeMatrix(MAT *mat)
{
    MagicCheck(MAT, mat)

    if (mat == NULL || mat->type == PHG_MATRIX_FREE
	|| mat->type == PHG_DESTROYED)
	return;

    phgInfo(1, "freeing matrix data.\n");
    mat->type = PHG_DESTROYED;

    free_matrix_rows(mat->rmap->localsize, &mat->rows);
    mat->rows = NULL;

    phgFree(mat->diag);
    mat->diag = NULL;
    phgFree(mat->packed_data);
    mat->packed_data = NULL;
    phgFree(mat->packed_cols);
    mat->packed_cols = NULL;

    mat->type = PHG_UNPACKED;

    return;
}

static void
mat_reset(MAT *mat)
{
    MagicCheck(MAT, mat)

    if (mat == NULL || mat->type == PHG_MATRIX_FREE)
	return;

    phgMatFreeMatrix(mat);

    free_matrix_rows(mat->rmap->nlocal, &mat->bdry_rows);
    mat->bdry_rows = NULL;
    free_bdry_eqns(mat);
    phgFree(mat->mv_data);
    mat->mv_data = NULL;
    phgFree(mat->bdry);
    mat->bdry = NULL;

    mat->nnz_d = mat->nnz_o = 0;
    mat->localsize = mat->bdry_localsize = 0;

    mat->assembled = FALSE;
    mat->type = PHG_UNPACKED;

    return;
}

void
phgMatDestroy(MAT **mat_ptr)
{
    MAT *mat = *mat_ptr;

    MagicCheck(MAT, mat)

    if (mat == NULL)
	return;

    *mat_ptr = NULL;

    assert(mat->ref_count >= 0);
    if (mat->ref_count > 0) {
	mat->ref_count--;
	return;
    }

    mat_reset(mat);

    phgMapDestroy(&mat->rmap);
    phgMapDestroy(&mat->cmap);

    phgFree(mat);

    return;
}

MAP *
phgMatGetRowMap(MAT *mat)
{
    mat->rmap->ref_count++;
    return mat->rmap;
}

MAP *
phgMatGetColumnMap(MAT *mat)
{
    mat->cmap->ref_count++;
    return mat->cmap;
}

static void
dup_matrix_rows(INT nrows, MAT_ROW *row0, MAT_ROW **row_ptr)
{
    MAT_ROW *row = *row_ptr;
    INT i;

    if (row0 == NULL || nrows == 0)
	return;

    if (row == NULL)
	*row_ptr = row = phgCalloc(nrows, sizeof(*row));

    for (i = 0; i < nrows; i++, row++, row0++) {
	row->ncols = row->alloc = row0->ncols;
	phgFree(row->cols);
	row->cols = phgAlloc(row->ncols * sizeof(*row->cols));
	memcpy(row->cols, row0->cols, row->ncols * sizeof(*row->cols));
	phgFree(row->data);
	row->data = phgAlloc(row->ncols * sizeof(*row->data));
	memcpy(row->data, row0->data, row->ncols * sizeof(*row->data));
    }
}

MAT *
phgMatDup(MAT *src)
{
    MAT *mat;
    INT i, n;
    size_t size;

    assert(src->type != PHG_DESTROYED);

    if (src->type == PHG_MATRIX_FREE) {
	src->ref_count++;
	return src;
    }

    if (!src->assembled)
	phgMatAssemble(src);

#define DupData(ptr, len)						\
    if (src->ptr != NULL && (size = (len) * sizeof(*src->ptr)) != 0) {	\
	mat->ptr = phgAlloc(size);					\
	memcpy(mat->ptr, src->ptr, size);				\
    }

    mat = phgMapCreateMat(src->rmap, src->cmap);
    mat->type = src->type;
    dup_matrix_rows(src->rmap->nlocal, src->rows, &mat->rows);
    DupData(diag, mat->rmap->nlocal);
    mat->nnz_d = src->nnz_d;
    mat->nnz_o = src->nnz_o;
    n = src->nnz_d + src->nnz_o;
    DupData(packed_cols, n + mat->rmap->nlocal);
    DupData(packed_data, n);
    mat->localsize = src->localsize;
    mat->handle_bdry_eqns = src->handle_bdry_eqns;
    mat->bdry_localsize = src->bdry_localsize;
    DupData(bdry, mat->bdry_localsize);

    mat->bdry_eqns_allocated = mat->bdry_eqns_count = src->bdry_eqns_count;
    if (mat->bdry_eqns_count > 0) {
	mat->bdry_eqns = phgAlloc(mat->bdry_eqns_count *
					sizeof(*mat->bdry_eqns));
	for (i = 0; i < mat->bdry_eqns_count; i++) {
	    n = src->bdry_eqns[i].n;
	    DupData(bdry_eqns[i].LU, n * n);
	    DupData(bdry_eqns[i].pvt, n);
	    DupData(bdry_eqns[i].map, n);
	}
    }

    dup_matrix_rows(mat->rmap->nlocal, src->bdry_rows, &mat->bdry_rows);

    mat->assembled = TRUE;

#undef DupData

    return mat;
}

void
phgMatDumpMATLAB(MAT *A, const char *vn, const char *fn, BOOLEAN dense)
{
    FILE *fp;
    INT i, k;
    FLOAT *a;
    MAT_ROW *row;

    assert(A->type != PHG_DESTROYED);

    if (!dense)
	phgError(1, "%s: sparse format unimplemented.\n", __func__);

    if (dense && A->cmap->nglobal > 3000)
	phgError(1, "%s: matrix too large.\n", __func__);

    if (dense && A->cmap->nprocs > 1)
	phgError(1, "%s: unimplemented for nprocs > 1.\n", __func__);

    a = phgAlloc(A->cmap->nglobal * sizeof(*a));
    if ((fp = fopen(fn, "w+t")) == NULL)
	phgError(1, "%s: cannot open \"%s\" for write.\n", __func__, fn);
    fprintf(fp, "%s = [\n", vn);
    for (i = 0, row = A->rows; i < A->rmap->nlocal; i++, row++) {
	bzero(a, A->cmap->nglobal * sizeof(*a));
	for (k = 0; k < row->ncols; k++)
	    a[row->cols[k]] = row->data[k];
	for (k = 0; k < A->cmap->nglobal; k++) {
	    fprintf(fp, "%0.16lg%s", (double)a[k],
			k < A->cmap->nglobal - 1 ? " " : "\n");
	}
    }
    fprintf(fp, "];\n");
    fclose(fp);
    phgFree(a);
}

VEC *
phgVecRandomize(VEC *v, long int seed)
{
    INT i, n;
    FLOAT *p;

    if (v->nvec == 0 || v->map->nlocal == 0)
	return v;
    if (seed != 0)
	srand48(seed + v->map->rank);
    n = v->nvec * v->map->nlocal;
    if (v->data == NULL)
	v->data = phgAlloc(n * sizeof(*v->data));
    for (i = 0, p = v->data; i < n; i++)
	*(p++) = 2.0 * drand48() - 1.0;

    return v;
}

/*-------------------------------------------------------------------------*/

#define	TRUNK_SIZE	8	/* size of allocation trunks */
#define SORT_COLUMNS	0	/* whether sort columns */

#if SORT_COLUMNS
static INT *cols_tmp;

static int
comp_columns(const void *c0, const void *c1)
{
    return cols_tmp[*((int *)c0)] - cols_tmp[*((int *)c1)];
}
#endif /* SORT_COLUMNS */

static int
add_columns_to_row(MAT_ROW *row, INT ncols, INT *cols, FLOAT *data,
			 BOOLEAN sorted)
/* note: 'sorted' indicates whether the input array 'cols' is sorted */
{
    INT index;
    int i;
#if SORT_COLUMNS
    INT index0, *old_cols;
    FLOAT *old_data;
    INT j, k, *ordering;	/* sorted indices of the array 'cols' */
#else /* SORT_COLUMNS */
    Unused(sorted);
#endif /* SORT_COLUMNS */

    /*for (i = 0; i < ncols; i++)
       if (Fabs(data[i]) < 1e-10)
       data[i] = 0.0; */

    if (ncols <= 0)
	return 0;

#if SORT_COLUMNS
    ordering = phgAlloc(ncols * sizeof(*ordering));
    for (i = 0; i < ncols; i++)
	ordering[i] = i;
    if (!sorted) {
	cols_tmp = cols;
	qsort(ordering, ncols, sizeof(*ordering), comp_columns);
    }
#endif /* SORT_COLUMNS */

    if (row->alloc == 0) {
	/* faster code for a new row */
	row->alloc = ((ncols + TRUNK_SIZE - 1) / TRUNK_SIZE) * TRUNK_SIZE;
	row->cols = phgAlloc(row->alloc * sizeof(*row->cols));
	row->data = phgAlloc(row->alloc * sizeof(*row->data));
	row->ncols = ncols;
#if SORT_COLUMNS
	for (i = 0; i < ncols; i++) {
	    row->cols[i] = cols[ordering[i]];
	    row->data[i] = data[ordering[i]];
	}
	phgFree(ordering);
#else /* SORT_COLUMNS */
	memcpy(row->cols, cols, ncols * sizeof(*cols));
	memcpy(row->data, data, ncols * sizeof(*data));
#endif /* SORT_COLUMNS */
	return 0;
    }

    /* merge new cols with existing cols */

#if SORT_COLUMNS
#if 1
    /* allocate new arrays */
    old_cols = row->cols;
    old_data = row->data;
    row->alloc = (row->ncols <= ncols) ? ncols : row->ncols;
    row->alloc = ((row->alloc + TRUNK_SIZE - 1) / TRUNK_SIZE) * TRUNK_SIZE;
    row->cols = phgAlloc(row->alloc * sizeof(INT));
    row->data = phgAlloc(row->alloc * sizeof(FLOAT));
    i = j = k = 0;
    /* merge arrays */
    while (i < ncols || j < row->ncols) {
	if (k >= row->alloc) {
	    /* need to enlarge the arrays */
	    row->alloc += TRUNK_SIZE;
	    row->cols =
		phgRealloc(row->cols, row->alloc * sizeof(*row->cols));
	    row->data =
		phgRealloc(row->data, row->alloc * sizeof(*row->data));
	}

	if (j >= row->ncols) {
	    row->cols[k] = cols[index = ordering[i++]];
	    row->data[k++] = data[index];
	}
	else if (i >= ncols) {
	    row->cols[k] = old_cols[j];
	    row->data[k++] = old_data[j++];
	}
	else if ((index = cols[ordering[i]]) == (index0 = old_cols[j])) {
	    row->cols[k] = index;
	    row->data[k++] = data[ordering[i++]] + old_data[j++];
	}
	else if (index > index0) {
	    row->cols[k] = index0;
	    row->data[k++] = old_data[j++];
	}
	else {			/* index < index0 */
	    row->cols[k] = index;
	    row->data[k++] = data[ordering[i++]];
	}
    }
    phgFree(old_data);
    phgFree(old_cols);
    phgFree(ordering);
    row->ncols = k;
#else /* 0|1 */
    /* first, count number of distinct cols */
    k = row->ncols + ncols;
    index = cols[ordering[i = 0]];
    index0 = row->cols[j = 0];
    while (i < ncols && j < row->ncols) {
	if (index == index0) {
	    k--;
	    if (++i >= ncols || ++j >= row->ncols)
		break;
	    index = cols[ordering[i]];
	    index0 = row->cols[j];
	}
	else if (index < index0) {
	    if (++i >= ncols)
		break;
	    index = cols[ordering[i]];
	}
	else {
	    if (++j >= row->ncols)
		break;
	    index0 = row->cols[j];
	}
    }

    if (k > row->alloc) {
	/* need to enlarge the arrays */
	row->alloc = ((k + TRUNK_SIZE - 1) / TRUNK_SIZE) * TRUNK_SIZE;
	row->cols = phgRealloc(row->cols, row->alloc * sizeof(*row->cols));
	row->data = phgRealloc(row->data, row->alloc * sizeof(*row->data));
    }

    index = cols[ordering[i = ncols - 1]];
    index0 = row->cols[j = row->ncols - 1];
    row->ncols = k;
    while (--k >= 0) {
	assert(i >= 0 || j >= 0);
	if (i < 0) {
	    if (k != j) {
		row->cols[k] = row->cols[j];
		row->data[k] = row->data[j--];
	    }
	    else {
		j--;
	    }
	    if (j >= 0)
		index0 = row->cols[j];
	}
	else if (j < 0) {
	    row->cols[k] = cols[index = ordering[i--]];
	    row->data[k] = data[index];
	    if (i >= 0)
		index = cols[ordering[i]];
	}
	else if (index == index0) {
	    row->cols[k] = index;
	    row->data[k] = data[ordering[i--]] + row->data[j--];;
	    if (j >= 0)
		index0 = row->cols[j];
	    if (i >= 0)
		index = cols[ordering[i]];
	}
	else if (index > index0) {
	    row->cols[k] = index;
	    row->data[k] = data[ordering[i--]];
	    if (i >= 0)
		index = cols[ordering[i]];
	}
	else {
	    if (k != j) {
		row->cols[k] = index0;
		row->data[k] = row->data[j--];;
	    }
	    else {
		j--;
	    }
	    if (j >= 0)
		index0 = row->cols[j];
	}
    }
#endif /* 0|1 */
#else /* SORT_COLUMNS */
    while (ncols-- > 0) {
	index = *(cols++);
	for (i = 0; i < row->ncols; i++)
	    if (row->cols[i] == index)
		break;
	if (i < row->ncols) {
	    /* add to existing column */
	    row->data[i] += *(data++);
	    continue;
	}
	if (row->ncols >= row->alloc) {
	    row->alloc += TRUNK_SIZE;
	    row->cols =
		phgRealloc(row->cols, row->alloc * sizeof(*row->cols));
	    row->data =
		phgRealloc(row->data, row->alloc * sizeof(*row->data));
	}
	row->cols[row->ncols] = index;
	row->data[row->ncols++] = *(data++);
    }
#endif /* SORT_COLUMNS */

    return 0;
}

MAT *
phgMatAXPBY(FLOAT a, MAT *x, FLOAT b, MAT **py)
/* computes y := a * x + b * y */
{
    MAT *y;
    INT i, j, ncols, nbuffer = 0, *cols = NULL;
    FLOAT *buffer = NULL, *data;
    MAT_ROW *row, *row0;

    assert(a == 0.0 || x != NULL);
    assert(x != NULL || (py != NULL && *py != NULL));

    if (x != NULL && !x->assembled) {
	assert(x->type != PHG_DESTROYED);
	phgMatAssemble(x);
    }

    if (py == NULL || *py == NULL) {
	b = 0.0;
	y = phgMapCreateMat(x->rmap, x->cmap);
	if (py != NULL)
	    *py = y;
    }
    else {
	y = *py;
	assert(y->type != PHG_DESTROYED);
	phgMatUnpack(y);
	if (x != NULL) {
	    assert(x->rmap->nglobal == y->rmap->nglobal &&
		   x->rmap->nlocal == y->rmap->nlocal &&
		   x->cmap->nglobal == y->cmap->nglobal &&
		   x->cmap->nlocal == y->cmap->nlocal &&
		   x->cmap->localsize == y->cmap->localsize);
	    assert((x->cmap->L2Vmap == NULL) == (y->cmap->L2Vmap == NULL));
#if 1
	    if (x->cmap->L2Vmap != NULL)
		assert(!memcmp(x->cmap->L2Vmap, y->cmap->L2Vmap,
				x->cmap->localsize * sizeof(*x->cmap->L2Vmap)));
	    if ((i = x->cmap->localsize - x->cmap->nlocal) > 0) {
		assert(x->cmap->O2Gmap != NULL && y->cmap->O2Gmap != NULL);
		assert(!memcmp(x->cmap->O2Gmap, y->cmap->O2Gmap,
				i * sizeof(*x->cmap->O2Gmap)));
	    }
#endif
	}
    }

    if (x != NULL) {
	phgMatUnpack(x);
    }

    if (a == 0.0 && b == 1.0)
	return y;

    /* TODO: faster code for b == 0.0 */

    y->assembled = FALSE;

    for (i = 0, row = y->rows; i < y->cmap->nlocal; i++, row++) {
	/* compute b * y */
	if (b == 0.0) {
	    row->ncols = row->alloc = 0;
	    phgFree(row->cols);
	    row->cols = NULL;
	    phgFree(row->data);
	    row->data = NULL;
	}
	else if (b == -1.0) {
	    ncols = row->ncols;
	    for (j = 0; j < ncols; j++) {
		row->data[j] = -row->data[j];
	    }
	}
	else if (b != 1.0) {
	    ncols = row->ncols;
	    for (j = 0; j < ncols; j++) {
		row->data[j] *= b;
	    }
	}

	/* add a * x to y */
	if (a == 0.0)
	    continue;
	row0 = x->rows + i;
	ncols = row0->ncols;
	if (ncols == 0)
	    continue;
	data = row0->data;
	if (a != 1.0) {
	    if (nbuffer < ncols) {
		phgFree(buffer);
		buffer = phgAlloc((nbuffer = ncols + 100) * sizeof(*buffer));
	    }
	    if (a == -1.0) {
		for (j = 0; j < ncols; j++)
		    buffer[j] = - *(data++);
	    }
	    else {
		for (j = 0; j < ncols; j++)
		    buffer[j] = a * *(data++);
	    }
	    data = buffer;
	}

	add_columns_to_row(row, ncols, row0->cols, data, FALSE);

	/* TODO: eleminate zero entries */
    }

    phgFree(buffer);
    phgFree(cols);
    free_matrix_rows(y->rmap->nlocal, &y->bdry_rows);
    free_bdry_eqns(y);
    phgMatAssemble(y);

    return y;
}

static void
build_bdry(MAT *mat)
/* builds mat->bdry and computes mat->bdry_localsize */
{
    INT k;

    if (mat->cmap->bdry_nglobal == 0)
	return;

    k = mat->bdry_localsize = mat->cmap->bdry_localsize;
    phgFree(mat->bdry);
    mat->bdry = NULL;
    if (k > 0) {
	mat->bdry = phgAlloc(k * sizeof(*mat->bdry));
	memcpy(mat->bdry, mat->cmap->bdry, k * sizeof(*mat->bdry));
    }

}

static BOOLEAN
collect_local_bdry_equations(MAT *mat, INT i, int map0[],
			     INT **map1, int *n, int *alloc)
{
    static int level = 0;
    INT j, k;
    MAT_ROW *row = mat->rows + i;

    if (*n >= *alloc)
	*map1 = phgRealloc(*map1, (*alloc += 128) * sizeof(**map1));
    map0[i] = *n;
    (*map1)[(*n)++] = i;

    for (j = 0; j < row->ncols; j++) {
	if (map0[k = row->cols[j]] >= 0)
	    continue;
	if (map0[k] == -2) {	/* non boundary entry */
	    static BOOLEAN warned = FALSE;
	    if (!warned) {
		phgWarning("bdry equations involve non bdry unknowns.");
		warned = TRUE;
	    }
	    /* reset map0[] */
	    for (j = 0; j < *n; j++)
		map0[(*map1)[j]] = -1;
	    *n = 0;
	    return FALSE;
	}
	level++;
	if (!collect_local_bdry_equations(mat, k, map0, map1, n, alloc)) {
	    level--;
	    return FALSE;
	}
	level--;
    }

    return TRUE;
}

static void
save_local_bdry_equations(MAT *mat)
/* do LU factorization and save local bdry equations before matrix assembly */
{
    int *map0;
    INT *map1 = NULL;
    int ii, n = 0, alloc = 0;
    MAT_ROW *row, *row1;
    INT i, j;
    double t0 = phgGetTime(NULL);
    int count = 0, nmax = 0;
    BDRY_EQNS *be;

    if (!mat->handle_bdry_eqns)
	return;

    assert(!mat->assembled);

    phgInfo(1, "trying to diagonalize local boundary equations.\n");

    for (ii = 0; ii < mat->rmap->ndof; ii++)
	if (mat->rmap->dofs[ii]->DB_mask != 0)
	    break;
    if (ii >= mat->rmap->ndof)	/* no Dirichlet boundary DOF */
	return;

    map0 = phgAlloc(mat->cmap->localsize * sizeof(*map0));
    for (i = 0; i < mat->cmap->localsize; i++)
	map0[i] = -2;
    for (i = 0; i < mat->cmap->bdry_localsize; i++)
	map0[mat->cmap->bdry[i]] = -1;

    for (i = 0, row = mat->rows; i < mat->rmap->localsize; i++, row++) {
	if (map0[i] == -2)
	    continue;
	/* Note: depending on phgDofDirichletBC, some bdry rows may have
	 * row->ncols == 0 */
	/* skip empty or diagonal rows */
	if (row->ncols <= 0 || (row->ncols == 1 && row->cols[0] == i))
	    continue;
	/* recursively collect connected boundary equations */
	if (!collect_local_bdry_equations(mat, i, map0, &map1, &n, &alloc))
	    continue;
	assert(n > 1);
	if (n > 2048) {
	    phgWarning("%s: bdry eqns too large (%d)\n", __func__, n);
	    /* reset map0[] */
	    for (ii = 0; ii < n; ii++)
		map0[map1[ii]] = -1;
	    n = 0;
	    continue;
	}
	if (n > nmax)
	    nmax = n;
	/* save the boundary equations */
	if (count >= mat->bdry_eqns_allocated) {
	    mat->bdry_eqns_allocated = count + 128;
	    mat->bdry_eqns = phgRealloc(mat->bdry_eqns,
					   mat->bdry_eqns_allocated *
					   sizeof(*mat->bdry_eqns));
	}
	be = mat->bdry_eqns + (count++);
	be->n = n;
	be->LU = phgCalloc(n * n, sizeof(*be->LU));
	be->pvt = phgAlloc(n * sizeof(*be->pvt));
	be->map = phgAlloc(n * sizeof(*be->map));
	for (ii = 0; ii < n; ii++) {
	    row1 = mat->rows + map1[ii];
	    for (j = 0; j < row1->ncols; j++)
		be->LU[map0[row1->cols[j]] + ii * n] = row1->data[j];
	    be->map[ii] = map1[ii];
	}
	if (!phgSolverDenseLU(n, (void *)be->LU, be->pvt))
	    phgError(1, "%s: singular bdry equations (n=%d).\n", __func__, n);

	for (ii = 0; ii < n; ii++) {
	    row1 = mat->rows + map1[ii];
	    if (row1->ncols != 1) {
		phgFree(row1->cols);
		phgFree(row1->data);
		row1->cols = phgAlloc(sizeof(*row1->cols));
		row1->data = phgAlloc(sizeof(*row1->data));
		row1->ncols = row1->alloc = 1;
	    }
	    row1->cols[0] = map1[ii];
	    row1->data[0] = 1.0;
	}

	/* reset map0[] */
	for (ii = 0; ii < n; ii++)
	    map0[map1[ii]] = -1;
	n = 0;
    }
    mat->bdry_eqns_count = count;
    phgInfo(1, "diagonalize: %d eqns, max size = %d, time = %0.4lf\n",
	    count, nmax, (double)(phgGetTime(NULL) - t0));

    phgFree(map0);
    phgFree(map1);
}

static void
save_bdry_columns(MAT *mat)
/* moves all non-diagonal boundary columns to bdry_rows[] */
{
    MAT_ROW *row, *row_bdry;
    INT i, j, k, l;
    BYTE *flags;

    if (!mat->handle_bdry_eqns)
	return;

    assert(!mat->assembled);

    flags = phgCalloc(mat->localsize, sizeof(*flags));
    free_matrix_rows(mat->rmap->nlocal, &mat->bdry_rows);
    mat->bdry_rows = phgCalloc(mat->rmap->nlocal, sizeof(*row));

    for (i = 0; i < mat->bdry_localsize; i++)
	flags[mat->bdry[i]] = 1;

    row_bdry = mat->bdry_rows;
    row = mat->rows;
    for (i = 0; i < mat->rmap->nlocal; i++, row++, row_bdry++) {
	/* count # of bdry columns */
	for (k = 0, j = 0; j < row->ncols; j++) {
	    if (flags[row->cols[j]] != 0)
		k++;
	}
	row_bdry->ncols = row->alloc = k;
	row_bdry->cols = phgAlloc(k * sizeof(*row_bdry->cols));
	row_bdry->data = phgAlloc(k * sizeof(*row_bdry->data));
	for (k = 0, l = 0, j = 0; j < row->ncols; j++) {
	    /* leave diagonal or non boundary entries in mat->rows */
	    if (row->cols[j] == i || flags[row->cols[j]] == 0) {
		if (l != j) {
		    row->cols[l] = row->cols[j];
		    row->data[l] = row->data[j];
		}
		l++;
	    }
	    /* save boundary entries to mat->bdry_rows */
	    if (flags[row->cols[j]] != 0) {
		row_bdry->cols[k] = row->cols[j];
		row_bdry->data[k] = row->data[j];
		k++;
	    }
	}
	row->cols = phgRealloc(row->cols, l * sizeof(*row->cols));
	row->data = phgRealloc(row->data, l * sizeof(*row->data));
	row->ncols = row->alloc = l;
    }
    phgFree(flags);
}

void
phgMatAssemble(MAT *mat)
/* assembles matrix, i.e., move off-process rows to corresponding
 * processes (Alltoallv). */
{
    save_local_bdry_equations(mat);
    mat->localsize = mat->cmap->localsize;

    build_bdry(mat);
    save_bdry_columns(mat);
    mat->assembled = TRUE;

    return;
}

#define dofindex2vecindex(map, dof_ind, vec_ind) {			\
    INT i_ = (dof_ind);							\
    if (map->L2Vmap != NULL && i_ > map->localsize) {			\
	phgInfo(-1, "AddEntries: invalid index: %d\n"			\
		"(note: AddEntries uses local indices)\n", i_);		\
	phgError(1, "AddEntries: abort.\n");				\
    }									\
    (vec_ind) = (map->L2Vmap == NULL ? i_ : map->L2Vmap[i_]);		\
}

void
phgMatAddEntry_(MAT *mat, INT row, INT col, FLOAT value,
		BOOLEAN row_is_global, BOOLEAN col_is_global)
{
    static FLOAT zero = 0.0;
    INT row1, col1;

    if (value == zero)
	return;

    assert(mat->assembled == FALSE);

    col1 = col;
    assert(col1 < mat->cmap->localsize);

    row1 = row;
    assert(row1 >= 0 && row1 < mat->rmap->localsize);

    add_columns_to_row(mat->rows + row1, 1, &col1, &value, FALSE);

#if 0
    assert(!row_is_global && !col_is_global);
    if (col != row && (solver->oem_solver == SOLVER_SUPERLU
			   /*|| solver->oem_solver == SOLVER_SPC */ )) {
	/* enforce symmetry of the sparsity */
	dofindex2vecindex(mat->cmap, col, row1);
	dofindex2vecindex(mat->rmap, row, col1);
	add_columns_to_row(mat->rows + row1, 1, &col1, &zero, FALSE);
    }
#endif /* 0|1 */

    return;
}

void
phgMatAddEntries_(MAT *mat, INT nrows, INT *rows, INT ncols, INT *cols,
		  FLOAT *values, BOOLEAN rows_are_global,
		  BOOLEAN cols_are_global)
{
    static INT allocated = 0;
    static INT *cols1 = NULL;
    static FLOAT *data1 = NULL;
    static FLOAT zero = 0.0;
    INT i, j, n, row1, *cols0;
    FLOAT a, *v;

    FreeAtExit(cols1);
    FreeAtExit(data1);

    if (ncols <= 0 || nrows <= 0)
	return;

    if (ncols == 1 && nrows == 1) {
	phgMatAddEntry_(mat, *rows, *cols, *values, rows_are_global,
			cols_are_global);
	return;
    }

    assert(mat->assembled == FALSE);

    if (allocated < ncols) {
	phgFree(data1);
	phgFree(cols1);
	allocated = ncols + 100;
	data1 = phgAlloc(allocated * sizeof(*data1));
	cols1 = phgAlloc(2 * allocated * sizeof(*cols1));
    }

    /* convert column indices to global ones */
    cols0 = cols;

#if DEBUG
    for (i = 0; i < ncols; i++) {
	assert(cols0[i] < mat->cmap->localsize);
    }
#endif	/* DEBUG */

    v = values;
    for (i = 0; i < nrows; i++) {
	for (j = 0, n = 0; j < ncols; j++) {
	    if ((a = *(v++)) == zero)
		continue;
	    cols1[n] = cols0[j];
	    data1[n++] = a;
	}
	if (n == 0)
	    continue;
	row1 = rows[i];
	if (row1 >= mat->rmap->localsize)
	    phgError(1, "%s: row (%d) >= localsize (%d)\n",
					row1, mat->rmap->localsize);
	add_columns_to_row(mat->rows + row1, n, cols1, data1, FALSE);
    }

#if 0
    /* FIXME: skip boundary DOF */
    if (solver->oem_solver == SOLVER_SUPERLU
	/*|| solver->oem_solver == SOLVER_SPC */ ) {
	/* enforce symmetry of the sparsity */
	INT col;
	assert(!rows_are_global && !cols_are_global);
	for (i = 0; i < nrows; i++) {
	    dofindex2vecindex(mat->rmap, rows[i], col);
	    for (j = 0; j < ncols; j++) {
		if (*(values++) == zero || cols1[j] == col)
		    continue;
		dofindex2vecindex(mat->cmap, cols[j], row1);
		add_columns_to_row(mat->rows + row1, 1, &col, &zero, FALSE);
	    }
	}
    }
#endif /* 0|1 */

    return;
}

static void
solve_local_bdry_equations(MAT *mat, VEC *rhs)
{
    int i, n = 0;
    INT j, k;
    FLOAT *x = NULL;
    BDRY_EQNS *be;

    phgInfo(1, "solving local boundary equations.\n");

    if (mat == NULL || rhs == NULL || mat->bdry_eqns_count == 0)
	return;

    assert(!rhs->assembled);

    be = mat->bdry_eqns;
    for (k = 0; k < mat->bdry_eqns_count; k++, be++) {
	if (n < be->n) {
	    n = be->n;
	    phgFree(x);
	    x = phgAlloc(n * sizeof(*x));
	}
	for (i = 0; i < be->n; i++) {
	    if ((j = be->map[i]) < rhs->map->nlocal)
		x[i] = rhs->data[j];
	    else
		x[i] = rhs->offp_data[j - rhs->map->nlocal];
	}
	phgSolverDenseSV(be->n, (void *)be->LU, be->pvt, 1, (void *)x);
	for (i = 0; i < be->n; i++) {
	    if ((j = be->map[i]) < rhs->map->nlocal)
		rhs->data[j] = x[i];
	    else
		rhs->offp_data[j - rhs->map->nlocal] = x[i];
	}
    }

    phgFree(x);
}

void
phgVecAssemble(VEC *vec)
{
    if (vec->assembled)
	return;

    solve_local_bdry_equations(vec->mat, vec);
    vec->assembled = TRUE;

    return;
}

void
phgVecAddEntry(VEC *vec, int which, INT index, FLOAT value)
{
    phgVecAddEntries(vec, which, 1, &index, &value);
}

void
phgVecAddEntries(VEC *vec, int which, INT n, INT *indices, FLOAT *values)
{
    INT i, row, offp_size;
    FLOAT *data, *offp_data;

    assert(which < vec->nvec/* && vec->assembled == FALSE*/);
    if (n == 0)
	return;

    offp_size = vec->map->localsize - vec->map->nlocal;
    data = vec->data + which * vec->map->nlocal;
    offp_data = vec->offp_data + which * offp_size;
    for (i = 0; i < n; i++, values++, indices++) {
	if (*values == 0.0)
	    continue;
	dofindex2vecindex(vec->map, *indices, row);
	assert(row < vec->map->localsize);
	if (row < vec->map->nlocal)
	    data[row] += *values;
	else
	    offp_data[row - vec->map->nlocal] += *values;
    }

    return;
}

/*--------------------------------------------------------------------------*/

MAT *
phgMatRemoveBoundaryEntries(MAT *mat)
/* removes boundary entries from mat */
{
    INT i, j, k, m, n0;
    MAT_ROW *row, *row1;
    INT *new_index = NULL;
    MAP *map;

    FunctionEntry;

    if (!mat->assembled)
	phgMatAssemble(mat);

    assert(mat->type != PHG_DESTROYED);

    if (mat->type == PHG_PACKED)
	phgMatUnpack(mat);

    if (mat->rmap->bdry_nglobal == 0 && mat->cmap->bdry_nglobal == 0)
	Return mat;

    if (mat->ref_count > 0 && mat->rmap->rank == 0 && mat->cmap->rank == 0)
	phgWarning("matrix referenced by other objects.\n");

    if (mat->rmap->bdry_nglobal != 0) {
	map = phgMapRemoveBoundaryEntries(mat->rmap);
	if (mat->type != PHG_MATRIX_FREE) {
	    /* remove boundary rows */
	    for (i = 0; i < mat->rmap->bdry_nlocal; i++) {
		row = mat->rows + mat->rmap->bdry[i];
		phgFree(row->cols);
		row->cols = NULL;
		phgFree(row->data);
		row->data = NULL;
		row->alloc = 0;
		row->ncols = -1;
	    }
	    row = row1 = mat->rows;
	    for (i = 0; i < mat->rmap->nlocal; i++, row++) {
		if (row->ncols < 0) {
		    row->ncols = 0;
		    continue;
		}
		if (row1 != row) {
		    /* copy row to row1 */
		    row1->cols = row->cols;
		    row1->data = row->data;
		    row1->alloc = row->alloc;
		    row1->ncols = row->ncols;
		    row->ncols = row->alloc = 0;
		}
	 	row1++;
	    }
	    assert(row1 - mat->rows == map->nlocal);
	    mat->rows = phgRealloc(mat->rows, map->localsize * sizeof(*row));
	}

	phgFree(mat->diag);
	mat->diag = NULL;
	free_matrix_rows(mat->rmap->nlocal, &mat->bdry_rows);
	phgMapDestroy(&mat->rmap);
	mat->rmap = map;
    }

    mat->handle_bdry_eqns = FALSE;
    free_bdry_eqns(mat);

    if (mat->cmap->bdry_nglobal == 0)
	Return mat;

    map = phgMapRemoveBoundaryEntries(mat->cmap);

    if (mat->type == PHG_MATRIX_FREE)
	goto end;

    /* remove boundary columns */
    n0 = mat->cmap->partition[mat->cmap->rank];

    /* expand mat->bdry array for faster checking */
    new_index = phgAlloc(mat->localsize * sizeof(*new_index));

    /* Note: the interval [nlocal, localsize) may include
     * unreferenced off-process entries (some of them may have been deleted by
     * save_bdry_columns()), which should be excluded from the communications
     * for getting new indices of off-process entries. So we proceed with the
     * following steps:
     *
     *  Step 1: initialize all entries in new_index to -1.
     *
     *  Step 2: set all boundary entries to -2.
     *
     *  Step 3: set all non boundary entries of new_index in [0, nlocal)
     *          to new local indices for 'mat'.
     *
     *  Step 4: scan rows, copy non boundary entries to 'mat'.
     *          The column indices for entries in [0, nlocal) are
     *          converted to new local indices for 'mat', and the entries
     *          in [nlocal, localsize) are unchanged
     *          (they only exist and require communications when nprocs > 1).
     */

    /* initialize all new_index entries to -1 */
    for (i = 0; i < mat->localsize; i++) {
	new_index[i] = -1;
    }
    /* set boundary entries in new_index to -2 */
    for (i = 0; i < mat->bdry_localsize; i++) {
	assert(mat->bdry[i] >= 0 && mat->bdry[i] < mat->localsize);
	new_index[mat->bdry[i]] = -2;
    }
    /* convert non boundary new_index entries to new local indices for mat */
    for (i = 0, k = 0; i < mat->cmap->nlocal; i++) {
	if (new_index[i] == -1)
	    new_index[i] = k++;
    }
    assert(k == map->nlocal);

    /* new_index >= 0 means a non boundary entry present in mat, and is its new
     * local index; new_index == -1 means an offprocess non boundary entry;
     * new_index == -2 means a boundary entry */

    for (i = 0, row = mat->rows; i < mat->rmap->nlocal; i++, row++) {
	/* count non boundary columns in current row */
	for (j = 0, k = 0; j < row->ncols; j++) {
	    m = row->cols[j];
	    assert(m >= 0 && m < mat->localsize);
	    if (new_index[m] > -2) {
		if (m >= mat->cmap->nlocal) {
		    new_index[m] = 0;
		    row->cols[k] = m;
		}
		else {
		    row->cols[k] = new_index[m];;
		    assert(row->cols[k] < map->nlocal);
		}
		row->data[k++] = row->data[j];
	    }
	}
	row->ncols = k;
	if (k == 0) {
	    phgFree(row->cols);
	    phgFree(row->data);
	    row->cols = NULL;
	    row->data = NULL;
	    row->alloc = 0;
	}
    }

    if (map->nprocs <= 1 || map->nglobal == 0) {
	mat->localsize = map->nlocal;
	goto end;
    }

end:
    phgFree(new_index);
    phgMapDestroy(&mat->cmap);
    mat->cmap = map;

    phgFree(mat->bdry);
    mat->bdry = NULL;
    mat->bdry_localsize = 0;

    Return mat;
}

int
phgMatEigenSolve(MAT *A, MAT *B, int n, int which, FLOAT tau,
		    int *nit, FLOAT *evals, MAP *map, DOF **u, ...)
{
    VEC *vec = NULL;
    int ndof;
    INT i, nglobal;
    DOF ***dofs, **dofs0;
    BOOLEAN remove_bdry;

    GetVariableArgs(ndof, dofs, u, DOF **);
    assert(ndof > 0);
    if (map == NULL) {
	dofs0 = phgAlloc(ndof * sizeof(*dofs0));
	for (i = 0; i < ndof; i++)
	    dofs0[i] = dofs[i][0];
	map = phgMapCreateN(ndof, dofs0);
	phgFree(dofs0);	/* note: map will be freed if dofs0 != NULL */
    }
    else {
	dofs0 = NULL;
	nglobal = 0;
        for (i = 0; i < ndof; i++)
            nglobal += DofGetDataCountGlobal(dofs[i][0]);
	if (nglobal != map->nglobal)
	    phgError(1, "%s: inconsistent map argument!\n", __func__);
    }
    remove_bdry = (map->nglobal - map->bdry_nglobal == A->cmap->nglobal);
    phgMapDofArraysToVecN(map, n, remove_bdry, &vec, ndof, dofs);
    n = phgEigenSolve(A, B, n, which, tau, evals, &vec, nit);
    phgMapVecToDofArraysN(map, vec, remove_bdry, ndof, dofs);
    phgVecDestroy(&vec);
    phgFree(dofs);
    if (dofs0 != NULL)
	phgMapDestroy(&map);

    return n;
}

int
phgMatDumpCSR(MAT *mat, const char *fn)
/*****************************************************************************
 * This function dumps the matrix in CSR format.
 *
 * [From Li Zhongze]
 *
 * Assume a matrix of size 4 has the following form:
 *
 *	a_{00}   a_{01}   0.0      a_{03}
 *	0.0      a_{11}   a_{12}    0.0
 *	a_{20}   0.0      a_{22}   a_{23}
 *	0.0      0.0      a_{32}   a_{33}
 *
 * The sample CSR file format is as follows (0-based C format):
 *
 *	4
 *	0
 *	3
 *	5
 *	8
 *	10
 *	0
 *	1
 *	3
 *	1
 *	2
 *	0
 *	2
 *	3
 *	2
 *	3
 *	a_{00}  
 *	a_{01}
 *	a_{03}
 *	a_{11}
 *	a_{12}
 *	a_{20} 
 *	a_{22}
 *	a_{23}
 *	a_{32}
 *	a_{33}
 *
 * The first line is the size of the matrix, followed by integer arrays ia
 * of size n+1, and ja of size nnz(the total number of nonzero entries in
 * the matrix, nnz=ia[n]-ia[0]). A real array a of size nnz is stored last
 * in the file.  ia is an array of pointer to the start of each row in
 * array ja and a. Array ja stores the column indices (0-based) of entries
 * on each row. Array a stores the corresponding entry values. 
 *****************************************************************************/
{
    FILE *fp;
    INT i, j;
    MAT_ROW *row;
    char *fn1;

    FunctionEntry;

    if (fn == NULL)
	Return 0;		/* do nothing */

    if (phgNProcs > 1) {
	fn1 = phgAlloc(strlen(fn) + 1 + 5);
	sprintf(fn1, "%s.%04d", fn, phgRank);
	fn = fn1;
    }
    else {
	fn1 = NULL;
    }

    if (!mat->assembled)
	phgMatAssemble(mat);

    assert(mat->type != PHG_DESTROYED);

    if (mat->type == PHG_PACKED)
	phgMatUnpack(mat);

    if ((fp = fopen(fn, "w+t")) == NULL) {
	phgError(0, "cannot open file \"%s\"\n", fn);
	Return 1;
    }
    phgFree(fn1);

    /* pass 1: dump matrix header */
    fprintf(fp, "%d\n", mat->rmap->nlocal);
    row = mat->rows;
    for (i = j = 0; i < mat->rmap->nlocal; i++, row++) {
	if (fp != NULL) {
	    fprintf(fp, "%d\n", j);
	    j += row->ncols;
	}
    }
    fprintf(fp, "%d\n", j);

    /* pass 2: dump column indices */
    row = mat->rows;
    for (i = 0; i < mat->rmap->nlocal; i++, row++)
	for (j = 0; j < row->ncols; j++)
	    fprintf(fp, "%d\n", row->cols[j]);

    /* pass 3: dump nonzero matrix entries */
    row = mat->rows;
    for (i = 0; i < mat->rmap->nlocal; i++, row++)
	for (j = 0; j < row->ncols; j++)
	    fprintf(fp, "%0.17lg\n", (double)row->data[j]);

    fclose(fp);
    phgInfo(2, "matrix dumped to file \"%s\".\n", fn);

    Return 0;
}

int
phgVecDump(VEC *vec, const char *fn)
/*****************************************************************************
 * This function dumps a vector to a file.
 *
 * The vector is written as follows (assuming vec size = 4):
 *	4
 *	v_{0}
 *	v_{1}
 *	v_{2}
 *	v_{3}
 *****************************************************************************/
{
    FILE *fp;
    INT i;
    char *fn1;

    FunctionEntry;

    if (fn == NULL)
	Return 0;		/* do nothing */

    if (phgNProcs > 1) {
	fn1 = phgAlloc(strlen(fn) + 1 + 5);
	sprintf(fn1, "%s.%04d", fn, phgRank);
	fn = fn1;
    }
    else {
	fn1 = NULL;
    }

    if (!vec->assembled)
	phgVecAssemble(vec);

    if ((fp = fopen(fn, "w+t")) == NULL) {
	phgError(0, "cannot open file \"%s\"\n", fn);
	Return 1;
    }
    phgFree(fn1);

    fprintf(fp, "%d\n", vec->map->nlocal);
    for (i = 0; i < vec->map->nlocal; i++)
	fprintf(fp, "%0.17lg\n", (double)vec->data[i]);

    fclose(fp);
    phgInfo(2, "RHS dumped to file \"%s\".\n", fn);

    Return 0;
}
