/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: solver-hypre.c,v 1.81 2007/05/22 07:50:47 zlb Exp $ */

/* The HYPRE interface for PHG */

#include "phg.h"

#if USE_HYPRE	/* whole file */

/* HAVE_HYPRE_AMS_SET_PI controls whether explicitly compute the matrix Pi.
 * Note: HAVE_HYPRE_AMS_SET_PI != 0 requires the HYPRE_AMSSetDiscretePi
 * function which does not exist in the standard distribution and is provided
 * by the non oficial patch 'hypre-ams-pi-patch' */
#define HAVE_HYPRE_AMS_SET_PI	0

#include <string.h>
#include <math.h>
#include "HYPRE_utilities.h"
#include "krylov.h"
#include "HYPRE.h"
#include "HYPRE_IJ_mv.h"
#include "HYPRE_parcsr_ls.h"

#ifdef DONT_HAVE_G_COMM
# define g_comm	MPI_COMM_WORLD
#else	/* DONT_HAVE_G_COMM */
# define g_comm	map->comm
#endif	/* DONT_HAVE_G_COMM */

typedef int (*SolveFcn)(HYPRE_Solver solver, HYPRE_ParCSRMatrix A,
			HYPRE_ParVector x, HYPRE_ParVector b);
typedef int (*SetPCFcn)(HYPRE_Solver solver, HYPRE_PtrToSolverFcn solve,
			HYPRE_PtrToSolverFcn setup, HYPRE_Solver precon);
typedef int (*GetItFcn)(HYPRE_Solver solver, int *nits);
typedef int (*GetResidFcn)(HYPRE_Solver solver, double *resid);
typedef int (*DestroyFcn)(HYPRE_Solver solver);
typedef int (*SetTolFcn)(HYPRE_Solver solver, double tol);
typedef int (*SetMaxIterFcn)(HYPRE_Solver solver, int maxiter);

typedef struct {
    int ilower, iupper;
    HYPRE_IJMatrix	A;		/* The coefficient matrix */
    HYPRE_IJVector	B, U;		/* RHS & solution */
    BOOLEAN		assembled;
    HYPRE_Solver	hsolver, precond;

    /* function pointers to be initialized by setup_hypre_solver() */
    SolveFcn	solver_setup, precon_setup; 
    SolveFcn	solver_solve, precon_solve; 
    SetTolFcn	solver_set_tol;
    SetMaxIterFcn solver_set_maxiter;
    DestroyFcn	solver_destroy, precon_destroy;
    SetPCFcn	solver_setpc;
    GetItFcn	solver_get_nits;
    GetResidFcn solver_get_resid;
} OEM_DATA;

/* convenience macro */
#define _t	((OEM_DATA *)solver->oem_data)
#define _pc	((HYPRE_PRECOND *)solver->pc)

typedef enum {
    NONE, PCG, ParaSails, GMRES, BiCGSTAB, Euclid, BoomerAMG,
#if HYPRE_VERSION_MAJOR >= 2
    AMS,
#endif	/* HYPRE_VERSION_MAJOR >= 2 */
    MLI
} SOLVER_ID;

static const char *solver_names[] = {
    "pcg", "gmres", "bicgstab", "boomeramg",
#if HYPRE_VERSION_MAJOR >= 2
    "ams",
#endif	/* HYPRE_VERSION_MAJOR >= 2 */
    NULL
};
static const SOLVER_ID solver_list[] = {
    PCG, GMRES, BiCGSTAB, BoomerAMG,
#if HYPRE_VERSION_MAJOR >= 2
    AMS,
#endif	/* HYPRE_VERSION_MAJOR >= 2 */
};
static int solver_id = 0;		/* default solver = PCG */

static const char *precon_names[] = {
    "none", "parasails", "euclid", "boomeramg",
#if HYPRE_VERSION_MAJOR >= 2
    "ams",
#endif	/* HYPRE_VERSION_MAJOR >= 2 */
    NULL
};
static const SOLVER_ID precon_list[] = {
    NONE, ParaSails, Euclid, BoomerAMG,
#if HYPRE_VERSION_MAJOR >= 2
    AMS,
#endif	/* HYPRE_VERSION_MAJOR >= 2 */
};
static int precon_id = 1;		/* default preconditioner = ParaSails */

static FLOAT pc_rtol = 0.0;		/* preconditioner rtol */
static INT pc_maxit = 1;		/* preconditioner maxit */

/* Note: the output matrices/vectors can be used as input for ams_driver.c
 * from the HYPRE distribution. For example,
 *	maxwell -pre_refines=2 -tol=10 +hypre_two_norm -hypre_dump_matvec=aFEM
 * and
 *	ams_driver -coord
 * should produce identical results. */
static char *dump_matvec = NULL;
static BOOLEAN two_norm = TRUE;	/* whether use two norm */
static INT gmres_kdim = 0/*5*/;	/* <=0 means default */

/* cmdline arguments to be passed to Euclid solver */
static int argc = 0;
static char **argv = NULL;

#if HYPRE_VERSION_MAJOR >= 2

typedef struct{
    HYPRE_IJMatrix G;			/* the discrete gradient matrix */
    HYPRE_IJMatrix Pi;			/* the Hcurl=>(H1)^3 matrix */
    HYPRE_IJMatrix A_alpha;		/* alpha possion matrix */
    HYPRE_IJMatrix A_beta;		/* beta poisson matrix */
    HYPRE_IJVector x, y, z;		/* coordinates of vertices */
} HYPRE_PRECOND;

/* default parameters for AMS solver */
static int	  cycle_type = 0;	/* 3-level cycle type - 1 */
static const char *cycle_names[] = {"01210", "0+1+2", "02120", "010+2",
	"0102010", "1+020", "0201020", "010+020", NULL};
static INT	ams_dim = 3;		/* space dimension */
static INT	rlx_type = 2;		/* relaxation type */
static INT	rlx_sweeps = 1;		/* number of relaxation sweeps */
static INT	amg_coarsen_type = 10;	/* BoomerAMG coarsening type */
static INT	amg_agg_levels = 1;	/* Levels of BoomerAMG agg. coarsening*/
static INT	amg_rlx_type = 3;	/* BoomerAMG relaxation type */
static FLOAT	theta = 0.25;		/* BoomerAMG threshold */
static BOOLEAN  ams_singular = FALSE;	/* singular problem (beta=0) */
static BOOLEAN	ams_h1_method = FALSE;	/* use block-diag Poisson solvers */

#endif	/* HYPRE_VERSION_MAJOR >= 2 */

/* default parameters for Parasails-PCG solver */
static INT	sai_max_levels = 1;
static FLOAT	sai_threshold = 0.1;
static FLOAT	sai_filter = 0.05;
static int	sai_sym = 1;
static const char *sai_sym_list[] = {"nonsym-indefinite", "spd",
					"nonsym-definite", NULL};
static int
RegisterOptions(void)
{
    phgOptionsRegisterTitle("\nHYPRE options:", NULL);
    phgOptionsRegisterKeyword("hypre_solver", "HYPRE solver",
				solver_names, &solver_id);
    phgOptionsRegisterKeyword("hypre_pc", "HYPRE preconditioner",
				precon_names, &precon_id);
    phgOptionsRegisterFloat("hypre_pc_rtol", "Preconditioner rtol", &pc_rtol);
    phgOptionsRegisterInt("hypre_pc_maxit", "Preconditioner maxit", &pc_maxit);
    phgOptionsRegisterNoArg("hypre_two_norm", "Use two-norm (PCG)", &two_norm);
    phgOptionsRegisterFilename("hypre_dump_matvec", "Dump HYPRE "
				"matrices/vectors", &dump_matvec);

    /* cmdline options for Parasails-PCG */
    phgOptionsRegisterTitle("[Options for GMRES solver]", NULL);
    phgOptionsRegisterInt("hypre_gmres_kdim", "Maxi. dimension of Krylov "
				"space (<=0 means use default)", &gmres_kdim);

    /* cmdline options for Parasails-PCG */
    phgOptionsRegisterTitle("[Options for ParaSails solver]", NULL);
    phgOptionsRegisterInt("hypre_sai_max_levels", "Max levels",
				&sai_max_levels);
    phgOptionsRegisterFloat("hypre_sai_threshold", "Threshold", &sai_threshold);
    phgOptionsRegisterFloat("hypre_sai_filter", "Filter", &sai_filter);
    phgOptionsRegisterKeyword("hypre_sai_sym", "Symmetry", sai_sym_list,
					&sai_sym);
#if HYPRE_VERSION_MAJOR >= 2
    /* cmdline options for AMS */
    phgOptionsRegisterTitle("[Options for AMS solver]", NULL);
    phgOptionsRegisterKeyword("hypre_ams_type", " Three-level cycle type",
				cycle_names, &cycle_type);
    phgOptionsRegisterInt("hypre_ams_dim", "Space dimension",
				&ams_dim);
    phgOptionsRegisterFloat("hypre_ams_theta", "BoomerAMG threshold",
				&theta);
    phgOptionsRegisterInt("hypre_ams_rlx", "Relaxation type",
				&rlx_type);
    phgOptionsRegisterInt("hypre_ams_rlxn", "Number of relaxation sweeps",
				&rlx_sweeps);
    phgOptionsRegisterInt("hypre_ams_ctype", "BoomerAMG coarsening type",
				&amg_coarsen_type);
    phgOptionsRegisterInt("hypre_ams_agg",
				"Levels of BoomerAMG agg. coarsening",
				&amg_agg_levels);
    phgOptionsRegisterInt("hypre_ams_amgrlx", "BoomerAMG relaxation type",
				&amg_rlx_type);
    phgOptionsRegisterNoArg("hypre_ams_singular", "Singular problem (beta=0)",
				&ams_singular);
    /*
    phgOptionsRegisterNoArg("hypre_ams_h1", "Use block-diag Poisson solves",
				&ams_h1_method);
    */
#endif	/* HYPRE_VERSION_MAJOR >= 2 */

    /* TODO: cmdline options for setting parameters for PCG, AMG, etc. */

    return 0;
}

static int
Initialize(int *argcp, char ***argvp)
/* duplcate command-line arguments (for use with Euclid preconditioner) */
{
    int i;

    /* TODO:
     *	1. only copy options handled by Euclid, and remove them from
     *	   the arguments list? 
     *	2. change '-oem_options' to '-petsc_options' and '-hypre_options'?
     */

    argv = phgAlloc(((argc = *argcp) + 1) * sizeof(*argv));
    for (i = 0; i < argc; i++)
	argv[i] = strdup((*argvp)[i]);
    argv[i] = NULL;

    return 0;
}

static int
Finalize(void)
{
    int i;

    for (i = 0; i < argc; i++)
	phgFree(argv[i]);

    phgFree(argv);

    argc = 0;
    argv = NULL;

    return 0;
}

static int
Create(SOLVER *solver)
{
    MAP *map = solver->mat->rmap;

    assert(solver->mat->type != PHG_MATRIX_FREE);
    assert(solver->mat->assembled);	/* only works with assembled matrix */

    solver->oem_data = phgAlloc(sizeof(OEM_DATA));
    _t->ilower = map->partition[map->rank];
    _t->iupper = map->partition[map->rank + 1] - 1;
    _t->A = NULL;
    _t->B = _t->U = NULL;
    _t->assembled = FALSE;

    /* create matrix */
    HYPRE_IJMatrixCreate(g_comm, _t->ilower, _t->iupper, _t->ilower,
			 _t->iupper, &_t->A);
    HYPRE_IJMatrixSetObjectType(_t->A, HYPRE_PARCSR);
    HYPRE_IJMatrixInitialize(_t->A);

    return 0;
}

static int
AddMatrixEntries(SOLVER *solver, INT nrows, INT *rows, INT ncols, INT *cols,
			FLOAT *values)
{
    int i, j;
    double *v;

    assert(sizeof(INT) == sizeof(int));

    if (sizeof(FLOAT) == sizeof(double)) {
	for (i = 0; i < nrows; i++) {
	    HYPRE_IJMatrixSetValues(_t->A, 1, &ncols, rows, cols,
							(double *)values);
	    rows++;
	    values += ncols;
	}
    }
    else {
	v = phgAlloc(ncols * sizeof(*v));
	for (i = 0; i < nrows; i++) {
	    for (j = 0; j < ncols; j++)
		v[j] = *(values++);
	    HYPRE_IJMatrixSetValues(_t->A, 1, &ncols, rows, cols, v);
	    rows++;
	    values += ncols;
	}
	phgFree(v);
    }

    return 0;
}

static int
AddRHSEntries(SOLVER *solver, INT n, INT *ni, FLOAT *values)
{
#ifndef DONT_HAVE_G_COMM
    MAP *map = solver->mat->rmap;
#endif	/* DONT_HAVE_G_COMM */
    int j;
    double *v;

    if (_t->B == NULL) {
	/* create RHS vector */
	HYPRE_IJVectorCreate(g_comm, _t->ilower, _t->iupper, &_t->B);
	HYPRE_IJVectorSetObjectType(_t->B, HYPRE_PARCSR);
	HYPRE_IJVectorSetMaxOffProcElmts(_t->B, 0);
	HYPRE_IJVectorInitialize(_t->B);
    }

    if (n <= 0)
	return 0;

    if (sizeof(FLOAT) == sizeof(double)) {
	HYPRE_IJVectorSetValues(_t->B, n, ni, (double *)values);
    }
    else {
	v = phgAlloc(n * sizeof(*v));
	for (j = 0; j < n; j++)
	    v[j] = *(values++);
	HYPRE_IJVectorSetValues(_t->B, n, ni, v);
	phgFree(v);
    }

    return 0;
}

static void
dump_vector(HYPRE_ParVector vec, const char *name)
{
    char fn[128];

    if (dump_matvec == NULL)
	return;
    sprintf(fn , "%s.%s", dump_matvec, name);
    phgInfo(1, "Dumping HYPRE ParVector to file \"%s\"\n", fn);
    HYPRE_ParVectorPrint(vec, fn);
}

static void
dump_matrix(HYPRE_ParCSRMatrix mat, const char *name)
{
    char fn[128];

    if (dump_matvec == NULL)
	return;
    sprintf(fn , "%s.%s", dump_matvec, name);
    phgInfo(1, "Dumping HYPRE ParCSRMatrix to file \"%s\"\n", fn);
    HYPRE_ParCSRMatrixPrint(mat, fn);
}

#if USE_HYPRE && HYPRE_VERSION_MAJOR >= 2

static int
send_data_to_hypre_ijmatrix(MAT *mat, HYPRE_IJMatrix A)
{
    MAP *map = mat->rmap;
    INT i, j, k, n0;
#ifndef DONT_HAVE_G_COMM
    INT r;
#endif	/* DONT_HAVE_G_COMM */
    int nprocs = map->nprocs;
    MAT_ROW *row;

    phgMatAssemble(mat);

    assert(mat->type == PHG_UNPACKED);
    n0 = mat->rmap->partition[map->rank];
    if (nprocs > 1) {
	row = mat->rows;
	for (i = 0; i < mat->rmap->nlocal; i++, row++) {
	    for (k = 0; k < row->ncols; k++)
#ifdef DONT_HAVE_G_COMM
		row->cols[k] = row->cols[k] + n0;
#else	/* DONT_HAVE_G_COMM */
		row->cols[k] = ((r = row->cols[k]) < mat->cmap->nlocal ?
				r + n0 : mat->O2Gmap[r - mat->cmap->nlocal]);
#endif
	}
    }
    
    j = n0;
    row = mat->rows;
    if (sizeof(double) == sizeof(FLOAT)) {
	for (i = 0; i < mat->rmap->nlocal; i++, row++, j++) {
	    if (row->ncols <= 0)
		phgError(1, "phgSolverAssemble: matrix row %d is empty!\n", i);
	    HYPRE_IJMatrixSetValues(A, 1, &(row->ncols), &j, row->cols,
					(double *)row->data);
	    phgFree(row->cols);
	    phgFree(row->data);
	    row->cols = NULL;
	    row->data = NULL;
	    row->ncols = row->alloc = 0;
	} /* end-i-loop */
    }
    else {
	double *data = NULL;
	size_t size = 0;

	for (i = 0; i < mat->rmap->nlocal; i++, row++, j++) {
	    if (row->ncols <= 0)
		phgError(1, "phgSolverAssemble: matrix row %d is empty!\n", i);
	    if (size < row->ncols) {
		phgFree(data);
		data = phgAlloc((size = row->ncols) * sizeof(*data));
	    }
	    for (k = 0; k < row->ncols; k++)
		data[k] = (double)row->data[k];
	    HYPRE_IJMatrixSetValues(A, 1, &(row->ncols), &j, row->cols, data);
	    phgFree(row->data);
	    row->data = NULL;
	    phgFree(row->cols);
	    row->cols = NULL;
	    row->ncols = row->alloc = 0;
	} /* end-i-loop */
	phgFree(data);
    }
    HYPRE_IJMatrixAssemble(A);

    return 0;
}

static void
ams_build_G_xyz(SOLVER *solver, HYPRE_IJMatrix *G, HYPRE_IJVector *x,
		HYPRE_IJVector *y, HYPRE_IJVector *z)
/* builds the gradients matrix G and coordinates vectors x, y, z.
 * (only valid for lowest order elements) */
{
    MAP *map = solver->rhs->map;
    DOF *u = map->dofs[0], *dof;
    GRID *g = u->g;
    SIMPLEX *e;
    SOLVER *solver_vert;
    int ilower_edge = map->partition[map->rank];
    int iupper_edge = map->partition[map->rank + 1] - 1;
    int ilower_vert,iupper_vert;
    int i;
    INT v0, v1;
    INT row, cols[2], ncols = 2;
    double values[2];
    double xyz[Dim];
    COORD *c;

    assert(u->dim==1 && DofTypeDim(u->type)==3 && DofTypeOrder(u->type)<=1);
    assert(u->type == DOF_ND1 || u->type == DOF_HC0 || G == NULL);

    dof = phgDofNew(g, DOF_P1, 1, "test", DofNoData);
    solver_vert = phgSolverCreate(SOLVER_DEFAULT, dof, NULL);
    ilower_vert = solver_vert->mat->rmap->partition[map->rank];
    iupper_vert = solver_vert->mat->rmap->partition[map->rank+1] - 1;

    if (G != NULL) {
	if (*G != NULL)
	    HYPRE_IJMatrixDestroy(*G);
	HYPRE_IJMatrixCreate(g_comm, ilower_edge, iupper_edge, 
				      ilower_vert, iupper_vert, G);
	HYPRE_IJMatrixSetObjectType(*G, HYPRE_PARCSR);
	HYPRE_IJMatrixInitialize(*G);
    }

    if (*x != NULL)
	HYPRE_IJVectorDestroy(*x);
    HYPRE_IJVectorCreate(g_comm, ilower_vert, iupper_vert, x);
    HYPRE_IJVectorSetObjectType(*x, HYPRE_PARCSR);
    HYPRE_IJVectorSetMaxOffProcElmts(*x, 0);
    HYPRE_IJVectorInitialize(*x);

    if (*y != NULL)
	HYPRE_IJVectorDestroy(*y);
    HYPRE_IJVectorCreate(g_comm, ilower_vert, iupper_vert, y);
    HYPRE_IJVectorSetObjectType(*y, HYPRE_PARCSR);
    HYPRE_IJVectorSetMaxOffProcElmts(*y, 0);
    HYPRE_IJVectorInitialize(*y);

    if (*z != NULL)
	HYPRE_IJVectorDestroy(*z);
    HYPRE_IJVectorCreate(g_comm, ilower_vert, iupper_vert, z);
    HYPRE_IJVectorSetObjectType(*z, HYPRE_PARCSR);
    HYPRE_IJVectorSetMaxOffProcElmts(*z, 0);
    HYPRE_IJVectorInitialize(*z);

    values[0]=1.0, values[1]=-1.0;
    ForAllElements(g, e){
	if (G != NULL) {
	    for (i = 0; i < NEdge; i++) {
		row = phgSolverMapE2G(solver, 0, e, i);
		if (row < ilower_edge || row > iupper_edge)
		    continue;
		GetEdgeVertices(e, i, v0, v1);
		cols[0] = phgSolverMapE2G(solver_vert, 0, e, v0);
		cols[1] = phgSolverMapE2G(solver_vert, 0, e, v1);
		HYPRE_IJMatrixSetValues(*G, 1, &ncols, &row, cols, values);
	    }
	}

	for (i = 0; i < NVert; i++) {
	    row = phgSolverMapE2G(solver_vert, 0, e, i);
	    if (row >= ilower_vert && row <= iupper_vert) {
		xyz[0] = (*(c = g->verts + e->verts[i]))[0];
		xyz[1] = (*c)[1];
		xyz[2] = (*c)[2];
		HYPRE_IJVectorSetValues(*x, 1, &row, xyz);
		HYPRE_IJVectorSetValues(*y, 1, &row, xyz + 1);
		HYPRE_IJVectorSetValues(*z, 1, &row, xyz + 2);
	     }
	 }
    }	/* ForAllElements */

    phgDofFree(&dof);
    phgSolverDestroy(&solver_vert);
    if (G != NULL)
	HYPRE_IJMatrixAssemble(*G);
    HYPRE_IJVectorAssemble(*x);
    HYPRE_IJVectorAssemble(*y);
    HYPRE_IJVectorAssemble(*z);
    
    return;
}

static DOF *dof_H1;

static void
ams_func_G(DOF *u, SIMPLEX *e, int bno, const FLOAT *lambda, FLOAT *values)
/* dummy function for computing G.
 * Note: the results are stored as values[u->dim][Dim] */
{
    int i;
    const FLOAT *p, (*J)[Dim + 1];

    Unused(bno);
    /* TODO: cache grad_lambda of the bases of dof_H1 */
    p = dof_H1->type->BasGrads(dof_H1, e, 0, -1, lambda);
    J = (void *)phgGeomGetJacobian(u->g, e);
    for (i = 0; i < u->dim; i++, values += Dim, p += Dim + 1) {
	values[0] = p[0]*J[0][0] + p[1]*J[1][0] + p[2]*J[2][0] + p[3]*J[3][0];
	values[1] = p[0]*J[0][1] + p[1]*J[1][1] + p[2]*J[2][1] + p[3]*J[3][1];
	values[2] = p[0]*J[0][2] + p[1]*J[1][2] + p[2]*J[2][2] + p[3]*J[3][2];
    }
}

#if HAVE_HYPRE_AMS_SET_PI
static void
ams_func_Pi(DOF *u, SIMPLEX *e, int bno, const FLOAT *lambda, FLOAT *values)
/* dummy function for computing Pi.
 * Note: the results are stored as values[u->dim][Dim] */
{
    int i;
    const FLOAT *p;

    Unused(bno);
    /* TODO: cache bases of dof_H1 */
    p = dof_H1->type->BasFuncs(dof_H1, e, 0, -1, lambda);
    for (i = 0; i < u->dim / Dim; i++, values += 3 * Dim, p++) {
	values[0] = *p;
	values[1] = 0.;
	values[2] = 0.;

	values[3] = 0.;
	values[4] = *p;
	values[5] = 0.;

	values[6] = 0.;
	values[7] = 0.;
	values[8] = *p;
    }
}
#endif	/* HAVE_HYPRE_AMS_SET_PI */

static void
ams_build_G_Pi(SOLVER *solver, HYPRE_IJMatrix *G, HYPRE_IJMatrix *Pi)
/* build the gradient matrix G whose elements are determined by:
 *	\grad\phi_j = \sum g_{kj} \psi_k
 * where \phi_j is the j-th basis function of the H1 space,
 * and \psi_k is the k-th basis function of the Hcurl space */
{
    MAP *map = solver->rhs->map;
    DOF *u = map->dofs[0];
    GRID *g = u->g;
    SIMPLEX *e;
    SOLVER *solver_H1;
    DOF *dof_Hcurl;
#if HAVE_HYPRE_AMS_SET_PI
    DOF *dof_Hcurl3;
#endif	/* HAVE_HYPRE_AMS_SET_PI */
    DOF_TYPE *h1_type = NULL;
    int ilower_Hcurl = map->partition[map->rank];
    int iupper_Hcurl = map->partition[map->rank + 1] - 1;
    int ilower_H1, iupper_H1;
    int i, j, N_Hcurl, N_H1, row, ncols, *cols0, *cols;
    FLOAT *p, *floats, *pdata[15];
    double *doubles;

    assert(map->ndof == 1 && u->type->dim == 3);

    if (DofTypeOrder(u->type) <= 4)
	h1_type = DOF_Pn[DofTypeOrder(u->type)];
    else
	h1_type = DOF_HBn[DofTypeOrder(u->type)];

    N_H1 = h1_type->nbas;
    N_Hcurl = u->type->nbas;

    dof_H1 = phgDofNew(g, h1_type, 1, "tmp", DofNoData);
    solver_H1 = phgSolverCreate(SOLVER_DEFAULT, dof_H1, NULL);
    ilower_H1 = solver_H1->mat->rmap->partition[map->rank];
    iupper_H1 = solver_H1->mat->rmap->partition[map->rank + 1] - 1;

    dof_Hcurl = phgDofNew(g, u->type, N_H1, "tmp", DofNoData);
    if (*G != NULL)
	HYPRE_IJMatrixDestroy(*G);
    HYPRE_IJMatrixCreate(g_comm, ilower_Hcurl, iupper_Hcurl, 
	     ilower_H1, iupper_H1, G);
    HYPRE_IJMatrixSetObjectType(*G, HYPRE_PARCSR);
    HYPRE_IJMatrixInitialize(*G);

#if HAVE_HYPRE_AMS_SET_PI
    dof_Hcurl3 = phgDofNew(g, u->type, Dim * N_H1, "tmp3", DofNoData);
    if (*Pi != NULL)
	HYPRE_IJMatrixDestroy(*Pi);
    HYPRE_IJMatrixCreate(g_comm, ilower_Hcurl, iupper_Hcurl, 
	     Dim * ilower_H1, Dim * (iupper_H1 + 1) - 1, Pi);
    HYPRE_IJMatrixSetObjectType(*Pi, HYPRE_PARCSR);
    HYPRE_IJMatrixInitialize(*Pi);
    i = Dim;
#else	/* HAVE_HYPRE_AMS_SET_PI */
    i = 1;
#endif	/* HAVE_HYPRE_AMS_SET_PI */

    floats = phgAlloc(i * N_Hcurl * N_H1 * sizeof(*floats));
    doubles = phgAlloc(i * N_H1 * sizeof(*doubles));
    cols0 = phgAlloc((1 + i) * N_H1 * sizeof(*cols0));
    cols = cols0 + N_H1;

    ForAllElements(g, e) {
	for (j = 0; j < N_H1; j++)
	    cols0[j] = phgSolverMapE2G(solver_H1, 0, e, j);
	/* The gradient matrix */
	p = floats;
	j = dof_Hcurl->dim * dof_Hcurl->type->np_edge;
	for (i = 0; i < NEdge; i++) {
	    u->type->InitFunc(dof_Hcurl, e, EDGE, i, NULL, ams_func_G,
				NULL, p, pdata);
	    pdata[4 + i] = p;
	    p += j;
	}

	j = dof_Hcurl->dim * dof_Hcurl->type->np_face;
	if (j > 0) {
	    for (i = 0; i < NFace; i++) {
		u->type->InitFunc(dof_Hcurl, e, FACE, i, NULL, ams_func_G,
					NULL, p, pdata);
		pdata[10 + i] = p;
		p += j;
	    }
	}

	if (dof_Hcurl->type->np_elem > 0)
	    u->type->InitFunc(dof_Hcurl, e, ELEMENT, 0, NULL, ams_func_G,
				NULL, p, pdata);

	p = floats;
	for (i = 0; i < N_Hcurl; i++) {
	    row = phgSolverMapE2G(solver, 0, e, i);
	    for (j = 0, ncols = 0; j < N_H1; j++, p++) {
		if (Fabs(*p) < 1e-12)
		    continue;
		cols[ncols] = cols0[j];
		doubles[ncols++] = *p;
	    }
	    HYPRE_IJMatrixSetValues(*G, 1, &ncols, &row, cols, doubles);
	}

#if HAVE_HYPRE_AMS_SET_PI
	/* The (H1)^3 matrix */
	p = floats;
	j = dof_Hcurl3->dim * dof_Hcurl3->type->np_edge;
	for (i = 0; i < NEdge; i++) {
	    u->type->InitFunc(dof_Hcurl3, e, EDGE, i, NULL, ams_func_Pi,
				NULL, p, pdata);
	    pdata[4 + i] = p;
	    p += j;
	}

	j = dof_Hcurl3->dim * dof_Hcurl3->type->np_face;
	if (j > 0) {
	    for (i = 0; i < NFace; i++) {
		u->type->InitFunc(dof_Hcurl3, e, FACE, i, NULL, ams_func_Pi,
					NULL, p, pdata);
		pdata[10 + i] = p;
		p += j;
	    }
	}

	if (dof_Hcurl3->type->np_elem > 0)
	    u->type->InitFunc(dof_Hcurl3, e, ELEMENT, 0, NULL, ams_func_Pi,
				NULL, p, pdata);

	p = floats;
	for (i = 0; i < N_Hcurl; i++) {
	    row = phgSolverMapE2G(solver, 0, e, i);
	    for (j = 0, ncols = 0; j < Dim * N_H1; j++, p++) {
		if (Fabs(*p) < 1e-12)
		    continue;
		cols[ncols] = Dim * cols0[j / Dim] + (j % Dim);
		doubles[ncols++] = *p;
	    }
	    HYPRE_IJMatrixSetValues(*Pi, 1, &ncols, &row, cols, doubles);
	}
#endif	/* HAVE_HYPRE_AMS_SET_PI */
    }	/* ForAllElements */

    phgFree(cols0);
    phgFree(floats);
    phgFree(doubles);

    phgSolverDestroy(&solver_H1);
    phgDofFree(&dof_H1);
    phgDofFree(&dof_Hcurl);
    HYPRE_IJMatrixAssemble(*G);
#if HAVE_HYPRE_AMS_SET_PI
    phgDofFree(&dof_Hcurl3);
    HYPRE_IJMatrixAssemble(*Pi);
#endif
    
    return;
}

static void
ams_destroy(SOLVER *solver)
{
    if (solver->pc == NULL)
	return;

    if (_pc->G != NULL) {
	HYPRE_IJMatrixDestroy(_pc->G);
	_pc->G = NULL;
    }
    if (_pc->Pi != NULL) {
	HYPRE_IJMatrixDestroy(_pc->Pi);
	_pc->Pi = NULL;
    }
    if (_pc->x != NULL) {
	HYPRE_IJVectorDestroy(_pc->x);
	_pc->x = NULL;
    }
    if (_pc->y != NULL) {
	HYPRE_IJVectorDestroy(_pc->y);
	_pc->y = NULL;
    }
    if (_pc->z != NULL) {
	HYPRE_IJVectorDestroy(_pc->z);
	_pc->z = NULL;
    }

    if (_pc->A_alpha != NULL) {
	HYPRE_IJMatrixDestroy(_pc->A_alpha);
	_pc->A_alpha = NULL;
    }
    if (_pc->A_beta != NULL) {
	HYPRE_IJMatrixDestroy(_pc->A_beta);
	_pc->A_beta = NULL;
    }
    phgFree(solver->pc);
    solver->pc = NULL;
    solver->pc_initialled = FALSE;
    phgInfo(1, "AMS solver destroyed.\n");

    return;
}

static HYPRE_PRECOND *
ams_init(SOLVER *solver)
{
    HYPRE_PRECOND *data = NULL;

    if (solver->pc != NULL) {
	if (solver->pc_initialled == TRUE)
	    return _pc;
	else
	    phgError(1, "Dirty data for Solver's pc\n");
    }
    else {
	phgInfo(1, "Initializing AMS solver.\n");
	data = phgAlloc(sizeof(HYPRE_PRECOND));
	solver->pc_initialled = TRUE;
	data->G = NULL;
	data->Pi = NULL;
	data->A_alpha = NULL;
	data->A_beta = NULL;
	data->x = NULL;
	data->y = NULL;
	data->z = NULL;
	solver->pc = (void *)data;
	solver->pc_destroy = ams_destroy;
    }
    ams_h1_method = FALSE;

    return data;
}

static void
ams_set_poisson(SOLVER *solver, MAT *alpha, MAT *beta)
{
    MAP *map = solver->rhs->map;
    int ilower_vert, iupper_vert;
    HYPRE_IJMatrix A_alpha=NULL;
    HYPRE_IJMatrix A_beta=NULL;
    
    assert(alpha != NULL);

    if (solver->pc_initialled != TRUE) {
	ams_init(solver);
    }
    else {
	if (_pc->A_alpha != NULL)
	    HYPRE_IJMatrixDestroy(_pc->A_alpha);
	if (_pc->A_beta != NULL)
	    HYPRE_IJMatrixDestroy(_pc->A_beta);
    }

    ilower_vert = alpha->rmap->partition[map->rank];
    iupper_vert = alpha->rmap->partition[map->rank + 1] - 1;
    HYPRE_IJMatrixCreate(g_comm, ilower_vert, iupper_vert, ilower_vert,
				     iupper_vert, &A_alpha);
    HYPRE_IJMatrixSetObjectType(A_alpha, HYPRE_PARCSR);
    HYPRE_IJMatrixInitialize(A_alpha);
    send_data_to_hypre_ijmatrix(alpha, A_alpha);
    _pc->A_alpha = A_alpha;

    if (beta != NULL) {
	ilower_vert = beta->rmap->partition[map->rank];
	iupper_vert = beta->rmap->partition[map->rank + 1] - 1;
	HYPRE_IJMatrixCreate(g_comm, ilower_vert, iupper_vert, ilower_vert,
						   iupper_vert, &A_beta);
	HYPRE_IJMatrixSetObjectType(A_beta, HYPRE_PARCSR);
	HYPRE_IJMatrixInitialize(A_beta);
	send_data_to_hypre_ijmatrix(beta, A_beta);
	_pc->A_beta = A_beta;
    }
    else {
	_pc->A_beta = NULL;
    }

    /* Force AMS use the poisson matrix provided by user */ 
    ams_h1_method = TRUE;

    return;
}

/* user-callable functions */
void
phgSolverHypreAMSSetPoisson(SOLVER *solver, DOF *alpha, DOF *beta)
{
    DOF *tmp, *u = solver->rhs->map->dofs[0];
    MAP *map;
    MAT *ma, *mb;
    int i, j, N;
    GRID *g = solver->rhs->map->dofs[0]->g;
    SIMPLEX *e;
    FLOAT *A, *B, *A0, d;
    INT *I;

    /* TODO: check DOF_TYPE of solver->dofs[] for consistency */
    assert(solver->rhs->map->ndof == 1 && DofTypeDim(u->type) == 3 &&
	   u->dim == 1);
    assert(beta == NULL || beta->dim == 1);

    if (DofTypeOrder(u->type) <= 4)
	tmp = phgDofNew(g, DOF_Pn[DofTypeOrder(u->type)], 1, "H1", DofNoData);
    else
	tmp = phgDofNew(g, DOF_HBn[DofTypeOrder(u->type)], 1, "H1", DofNoData);
    map = phgMapCreate(tmp, NULL);
    ma = phgMapCreateMat(map, map);
    /* FIXME: implement a new function phgSolverDuplicate */
    if (beta != NULL && beta->type == DOF_CONSTANT && *beta->data == 0.)
	mb = NULL;
    else
	mb = phgMapCreateMat(map, map);
    phgMapDestroy(&map);

    N = tmp->type->nbas;	/* number of basis functions in an element */
    A = phgAlloc(N * N * sizeof(*A));
    A0 = phgAlloc(N * N * sizeof(*A0));
    B = phgAlloc(N * N * sizeof(*B));
    I = phgAlloc(N * sizeof(*I));

    /* Note: if beta is constant then the element mass matrices only vary
       by element volumes */
    if (beta == NULL || beta->type == DOF_CONSTANT) {
	e = g->roots;
	d = (beta == NULL ? 1.0 : *beta->data) / phgGeomGetVolume(g, e);
	for (i = 0; i < N; i++)
	    for (j = 0; j <= i; j++)
		A0[i * N + j] = A0[j * N + i] =
		    d * phgQuadBasDotBas(e, tmp, i, tmp, j, -1);
    }

    ForAllElements(g, e) {
	if (beta != NULL && beta->type != DOF_CONSTANT) {
	    /* compute \int beta * phi_i * phi_j */
	    for (i = 0; i < N; i++)
		for (j = 0; j <= i; j++)
		    A[i * N + j] = phgQuadBasABas(e, tmp, i, beta, tmp, j, -1);
	}
	else {
	    d = phgGeomGetVolume(g, e);
	    for (i = 0; i < N; i++)
		for (j = 0; j <= i; j++)
		    A[i * N + j] = A0[i * N + j] * d;
	}
	for (i = 0; i < N; i++) {
	    I[i] = phgMapE2L(ma->rmap, 0, e, i);
	    for (j = 0; j <= i; j++) {
		/* FIXME: use phgQuadGradBasDotGradBas if both alpha and beta
		 * are constant */
		A[j * N + i] = (A[i * N + j] +=
			phgQuadGradBasAGradBas(e, tmp, i, alpha, tmp, j, -1));
		if (mb != NULL)
		    B[j * N + i] = B[i * N + j] =
			phgQuadGradBasAGradBas(e, tmp, i, beta, tmp, j, -1);
	    }
	}
	/* Stiffness matrix */
	for (i = 0; i < N; i++) {
	    BTYPE btype;
	    btype = phgDofGetElementBoundaryType(tmp, e, i * tmp->dim);
	    if (IS_BDRY(btype)) {
#if 0
		phgMatAddEntry(ma, I[i], I[i], 1.0);
		phgMatAddEntry(mb, I[i], I[i], 1.0);
#else
		phgMatAddEntry(ma, I[i], I[i], A[i * N + i]);
		if (mb != NULL)
		    phgMatAddEntry(mb, I[i], I[i], B[i * N + i]);
#endif
	    }
	    else {
		phgMatAddEntries(ma, 1, I + i, N, I, A + i * N);
		if (mb != NULL)
		    phgMatAddEntries(mb, 1, I + i, N, I, B + i * N);
	    }
	}
    }	/* end-ForAllElement */

    phgFree(A);
    phgFree(A0);
    phgFree(B);
    phgFree(I);

    ams_set_poisson(solver, ma, mb);
    phgMatDestroy(&ma);
    phgMatDestroy(&mb);
    phgDofFree(&tmp);
}

void
phgSolverHypreAMSSetConstantPoisson(SOLVER *solver, FLOAT alpha, FLOAT beta)
{
    GRID *g = solver->rhs->map->dofs[0]->g;
    DOF *da, *db;

    if (alpha == 1.0) {
	da = NULL;
    }
    else {
	da = phgDofNew(g, DOF_CONSTANT, 1, "alpha", DofNoAction);
	phgDofSetDataByValue(da, alpha);
    }

    if (beta == 1.0) {
	db = NULL;
    }
    else {
	db = phgDofNew(g, DOF_CONSTANT, 1, "beta", DofNoAction);
	phgDofSetDataByValue(db, beta);
    }

    phgSolverHypreAMSSetPoisson(solver, da, db);

    phgDofFree(&da);
    phgDofFree(&db);
}

#endif	/* USE_HYPRE && HYPRE_VERSION_MAJOR >= 2 */

static void
setup_hypre_solver(SOLVER *solver, SOLVER_ID id, HYPRE_Solver *hsolver,
		   FLOAT rtol, INT maxit,
		   SolveFcn *setup, SolveFcn *solve, DestroyFcn *destroy,
		   SetTolFcn *set_tol, SetMaxIterFcn *set_maxiter,
		   SetPCFcn *set_pc, GetItFcn *get_nits, GetResidFcn *get_resid)
{
    MAP *map = solver->rhs->map;
#if HYPRE_VERSION_MAJOR >= 2
    HYPRE_ParCSRMatrix  parG, parPi, parA_alpha, parA_beta;
    HYPRE_ParVector	par_x, par_y, par_z;
#endif /* HYPRE_VERSION_MAJOR >= 2 */

    switch (id) {
	case PCG:	/* Preconditioned CG solver */
	    HYPRE_ParCSRPCGCreate(g_comm, hsolver);
	    if (set_maxiter == NULL)
		HYPRE_PCGSetMaxIter(*hsolver, maxit);	/* max iterations */
	    else
		*set_maxiter = HYPRE_PCGSetMaxIter;
	    if (set_tol == NULL)
		HYPRE_PCGSetTol(*hsolver, rtol);	/* conv. tolerance */
	    else
		*set_tol = HYPRE_PCGSetTol;
	    HYPRE_PCGSetTwoNorm(*hsolver, two_norm);	/* whether use 2-norm */
	    HYPRE_PCGSetPrintLevel(*hsolver, phgVerbosity + 1);
	    HYPRE_PCGSetLogging(*hsolver, 0);
	    *setup = HYPRE_ParCSRPCGSetup;
	    *solve = HYPRE_ParCSRPCGSolve;
	    *destroy = HYPRE_ParCSRPCGDestroy;
	    if (set_pc != NULL)
		*set_pc = HYPRE_PCGSetPrecond;
	    if (get_nits != NULL)
		*get_nits = HYPRE_PCGGetNumIterations;
	    if (get_resid != NULL)
		*get_resid = HYPRE_PCGGetFinalRelativeResidualNorm;
	    break;
	case GMRES:	/* GMRES */
	    HYPRE_ParCSRGMRESCreate(g_comm, hsolver);
	    if (set_maxiter == NULL)
		HYPRE_ParCSRGMRESSetMaxIter(*hsolver, maxit);
	    else
		*set_maxiter = HYPRE_ParCSRGMRESSetMaxIter;
	    if (set_tol == NULL)
		HYPRE_ParCSRGMRESSetTol(*hsolver, rtol);
	    else
		*set_tol = HYPRE_ParCSRGMRESSetTol;
	    HYPRE_ParCSRGMRESSetPrintLevel(*hsolver, phgVerbosity + 1);
	    HYPRE_ParCSRGMRESSetLogging(*hsolver, 0);
	    if (gmres_kdim > 0)
		HYPRE_ParCSRGMRESSetKDim (*hsolver, gmres_kdim);
	    *setup = HYPRE_ParCSRGMRESSetup;
	    *solve = HYPRE_ParCSRGMRESSolve;
	    *destroy = HYPRE_ParCSRGMRESDestroy;
	    if (set_pc != NULL)
		*set_pc = HYPRE_GMRESSetPrecond/*HYPRE_ParCSRGMRESSetPrecond*/;
	    if (get_nits != NULL)
		*get_nits = HYPRE_ParCSRGMRESGetNumIterations;
	    if (get_resid != NULL)
		*get_resid = HYPRE_ParCSRGMRESGetFinalRelativeResidualNorm;
	    break;
	case BiCGSTAB:	/* BiCGSTAB */
	    HYPRE_ParCSRBiCGSTABCreate(g_comm, hsolver);
	    if (set_maxiter == NULL)
		HYPRE_ParCSRBiCGSTABSetMaxIter(*hsolver, maxit);
	    else
		*set_maxiter = HYPRE_ParCSRBiCGSTABSetMaxIter;
	    if (set_tol == NULL)
		HYPRE_ParCSRBiCGSTABSetTol(*hsolver, rtol);
	    else
		*set_tol = HYPRE_ParCSRBiCGSTABSetTol;
	    HYPRE_ParCSRBiCGSTABSetPrintLevel(*hsolver, phgVerbosity);
	    HYPRE_ParCSRBiCGSTABSetLogging(*hsolver, 0);
	    *setup = HYPRE_ParCSRBiCGSTABSetup;
	    *solve = HYPRE_ParCSRBiCGSTABSolve;
	    *destroy = HYPRE_ParCSRBiCGSTABDestroy;
	    if (set_pc != NULL)
		*set_pc = HYPRE_BiCGSTABSetPrecond;
	    if (get_nits != NULL)
		*get_nits = HYPRE_ParCSRBiCGSTABGetNumIterations;
	    if (get_resid != NULL)
		*get_resid = HYPRE_ParCSRBiCGSTABGetFinalRelativeResidualNorm;
	    break;
	case BoomerAMG:	/* BoomerAMG */
	    HYPRE_BoomerAMGCreate(hsolver);
	    HYPRE_BoomerAMGSetPrintLevel(*hsolver, 2 * phgVerbosity);
	    HYPRE_BoomerAMGSetCoarsenType(*hsolver, 6);	/* Falgout coarsening */
	    HYPRE_BoomerAMGSetRelaxType(*hsolver, 3);	/* GS/J. hybrid relax */
	    HYPRE_BoomerAMGSetNumSweeps(*hsolver, 1);	/* swps on each level */
	    HYPRE_BoomerAMGSetMaxLevels(*hsolver, 20);	/* maxi nbr of levels */
	    if (set_maxiter == NULL)
		HYPRE_BoomerAMGSetMaxIter(*hsolver, maxit);
	    else
		*set_maxiter = HYPRE_BoomerAMGSetMaxIter;
	    if (set_tol == NULL)
		HYPRE_BoomerAMGSetTol(*hsolver, rtol);
	    else
		*set_tol = HYPRE_BoomerAMGSetTol;
	    *setup = HYPRE_BoomerAMGSetup;
	    *solve = HYPRE_BoomerAMGSolve;
	    *destroy = HYPRE_BoomerAMGDestroy;
	    if (set_pc != NULL)
		*set_pc = NULL;
	    if (get_nits != NULL)
		*get_nits = HYPRE_BoomerAMGGetNumIterations;
	    if (get_resid != NULL)
		*get_resid = HYPRE_BoomerAMGGetFinalRelativeResidualNorm;
	    break;
	case ParaSails:	/* ParaSails solver */
	    HYPRE_ParaSailsCreate(g_comm, hsolver);
	    HYPRE_ParaSailsSetParams(*hsolver, sai_threshold, sai_max_levels);
	    HYPRE_ParaSailsSetFilter(*hsolver, sai_filter);
	    HYPRE_ParaSailsSetSym(*hsolver, sai_sym);
	    HYPRE_ParaSailsSetLogging(*hsolver, 0);
	    *setup = HYPRE_ParaSailsSetup;
	    *solve = HYPRE_ParaSailsSolve;
	    *destroy = HYPRE_ParaSailsDestroy;
	    if (set_pc != NULL)
		*set_pc = NULL;
	    if (get_nits != NULL)
		*get_nits = NULL;
	    if (get_resid != NULL)
		*get_resid = NULL;
	    break;
	case Euclid:	/* Euclid solver (ILU) */
	    HYPRE_EuclidCreate(g_comm, hsolver);
	    /* pass options from command line */
	    HYPRE_EuclidSetParams(*hsolver, argc, argv);
	    /* alternatively, pass options from a configuration file
	     * HYPRE_EuclidSetParamsFromFile(*hsolver, "filename"); */
	    *setup = HYPRE_EuclidSetup;
	    *solve = HYPRE_EuclidSolve;
	    *destroy = HYPRE_EuclidDestroy;
	    if (set_pc != NULL)
		*set_pc = NULL;
	    if (get_nits != NULL)
		*get_nits = NULL;
	    if (get_resid != NULL)
		*get_resid = NULL;
	    break;
#if HYPRE_VERSION_MAJOR >= 2
	case AMS:
	    HYPRE_AMSCreate(hsolver);
	    HYPRE_AMSSetDimension(*hsolver, ams_dim);
	    if (set_maxiter == NULL)
		HYPRE_AMSSetMaxIter(*hsolver, maxit);
	    else
		*set_maxiter = HYPRE_AMSSetMaxIter;
	    if (set_tol == NULL)
		HYPRE_AMSSetTol(*hsolver, rtol);
	    else
		*set_tol = HYPRE_AMSSetTol;
	    HYPRE_AMSSetCycleType(*hsolver, cycle_type + 1);
	    HYPRE_AMSSetPrintLevel(*hsolver, phgVerbosity > 1 ? 1 : 0);
	    ams_init(solver);
	    if (map->dofs[0]->type==DOF_ND1 || map->dofs[0]->type==DOF_HC0) {
		/* faster code for the lowest order elements */
		ams_build_G_xyz(solver, &_pc->G, &_pc->x, &_pc->y, &_pc->z);
	    }
	    else {
		ams_build_G_Pi(solver, &_pc->G, &_pc->Pi);
#if !HAVE_HYPRE_AMS_SET_PI
		ams_build_G_xyz(solver, NULL, &_pc->x, &_pc->y, &_pc->z);
#endif	/* !HAVE_HYPRE_AMS_SET_PI */
	    }
	    HYPRE_IJMatrixGetObject(_pc->G, (void **)(void *)&parG);
	    dump_matrix(parG, "G");
	    HYPRE_AMSSetDiscreteGradient(*hsolver, parG);
	    if (_pc->Pi != NULL) {
		HYPRE_IJMatrixGetObject(_pc->Pi, (void **)(void *)&parPi);
		dump_matrix(parPi, "Pi");
#if HAVE_HYPRE_AMS_SET_PI
		HYPRE_AMSSetDiscretePi(*hsolver, parPi);
#endif	/* HAVE_HYPRE_AMS_SET_PI */
	    }
	    else {
		assert(_pc->x != NULL && _pc->y != NULL && _pc->z != NULL);
		HYPRE_IJVectorGetObject(_pc->x, (void **)(void *)&par_x);
		HYPRE_IJVectorGetObject(_pc->y, (void **)(void *)&par_y);
		HYPRE_IJVectorGetObject(_pc->z, (void **)(void *)&par_z);
		dump_vector(par_x, "x");
		dump_vector(par_y, "y");
		dump_vector(par_z, "z");
		HYPRE_AMSSetCoordinateVectors(*hsolver, par_x, par_y, par_z);
	    }

	    /* Poisson matrices */
	    if (ams_h1_method) {
		assert(_pc->A_alpha != NULL);
		HYPRE_IJMatrixGetObject(_pc->A_alpha,
						(void **)(void *)&parA_alpha);
		dump_matrix(parA_alpha, "Aalpha");
		if (!ams_singular && _pc->A_beta != NULL) {
		    HYPRE_IJMatrixGetObject(_pc->A_beta,
						(void **)(void *)&parA_beta);
		    dump_matrix(parA_beta, "Abeta");
		}
		else {
		    parA_beta = NULL;
		}
		HYPRE_AMSSetAlphaPoissonMatrix(*hsolver, parA_alpha);
		HYPRE_AMSSetBetaPoissonMatrix(*hsolver, parA_beta);
	    }
	    if (ams_singular)
		HYPRE_AMSSetBetaPoissonMatrix(*hsolver, NULL);
	    HYPRE_AMSSetSmoothingOptions(*hsolver,
			rlx_type, rlx_sweeps, 1.0, 1.0);
	    HYPRE_AMSSetAlphaAMGOptions(*hsolver, amg_coarsen_type,
			amg_agg_levels, amg_rlx_type, theta);
	    HYPRE_AMSSetBetaAMGOptions(*hsolver, amg_coarsen_type,
			amg_agg_levels, amg_rlx_type, theta);
	    *setup = HYPRE_AMSSetup;
	    *solve = HYPRE_AMSSolve;
	    *destroy = HYPRE_AMSDestroy;
	    if (set_pc != NULL)
		*set_pc = NULL;
	    if (get_nits != NULL)
		*get_nits = HYPRE_AMSGetNumIterations;
	    if (get_resid != NULL)
		*get_resid = HYPRE_AMSGetFinalRelativeResidualNorm;
	    break;
#endif	/* HYPRE_VERSION_MAJOR >= 2 */
#if 0
	case MLI_PCG:		/* Another AMG preconditioner */
	    HYPRE_LSI_MLICreate(g_comm, hsolver);
	    HYPRE_LSI_MLISetParams(*hsolver, "MLI strengthThreshold 0.08");
	    *setup = HYPRE_LSI_MLISetup;
	    *solve = HYPRE_LSI_MLISolve;
	    *destroy = HYPRE_LSI_MLIDestroy;
	    if (set_pc != NULL)
		*set_pc = (void *)HYPRE_LSI_MLISetPrecond;
	    if (get_nits != NULL)
		*get_nits = HYPRE_LSI_MLIGetNumIterations;
	    if (get_resid != NULL)
		*get_resid = HYPRE_LSI_MLIGetFinalRelativeResidualNorm;
	    break;
#endif
	default:
	    *hsolver = NULL;
	    break;
    }
}

static int
Destroy(SOLVER *solver)
{
    if (solver->oem_data == NULL)
	return 0;

#if HYPRE_VERSION_MAJOR >= 2
    ams_destroy(solver); 
#endif	/* HYPRE_VERSION_MAJOR >= 2 */
    if (_t->precond != NULL && _t->precon_destroy != NULL)
	_t->precon_destroy(_t->precond);

    if (_t->hsolver != NULL)
	_t->solver_destroy(_t->hsolver);

    if (_t->A != NULL) {
	HYPRE_IJMatrixDestroy(_t->A);
    }
    if (_t->B != NULL) {
	HYPRE_IJVectorDestroy(_t->B);
    }
    if (_t->U != NULL) {
	HYPRE_IJVectorDestroy(_t->U);
    }

    phgFree(solver->oem_data);
    solver->oem_data = NULL;

    return 0;
}

static int
Assemble(SOLVER *solver)
{
    if (_t->assembled)
	return 0;
    _t->assembled = TRUE;

    HYPRE_IJMatrixAssemble(_t->A);

    /* setup solver */
    phgInfo(1, "setup HYPRE solver \"%s\".\n", solver_names[solver_id]);
    setup_hypre_solver(solver, solver_list[solver_id], &_t->hsolver,
		       0., 0, &_t->solver_setup, &_t->solver_solve,
		       &_t->solver_destroy, &_t->solver_set_tol,
		       &_t->solver_set_maxiter, &_t->solver_setpc,
		       &_t->solver_get_nits, &_t->solver_get_resid);
    if (_t->hsolver == NULL)
	phgError(1, "solver \"%s\" is unavailable or unimplemented.\n",
			solver_names[solver_id]);
    if (_t->solver_setpc != NULL) {
	/* setup preconditioner */
	setup_hypre_solver(solver, precon_list[precon_id], &_t->precond,
			   pc_rtol, pc_maxit, &_t->precon_setup,
			   &_t->precon_solve, &_t->precon_destroy,
			   NULL, NULL, NULL, NULL, NULL);
	if (_t->precond != NULL) {
	    phgInfo(1, "setup HYPRE preconditioner \"%s\".\n",
			precon_names[precon_id]);
	    _t->solver_setpc(_t->hsolver,
			     (HYPRE_PtrToSolverFcn)_t->precon_solve,
			     (HYPRE_PtrToSolverFcn)_t->precon_setup,
			     _t->precond);
	}
    }
    else {
	phgInfo(1, "solver \"%s\" does not support preconditioner.\n",
			solver_names[solver_id]);
	_t->precond = NULL;
	_t->precon_destroy = NULL;
    }

    return 0;
}

static int
Solve(SOLVER *solver, VEC *x, BOOLEAN destroy)
{
    MAP *map = solver->rhs->map;
    INT i, offset, *ni;
    int num_iterations = 1;
    double res_norm = 0., *v = NULL;

    HYPRE_ParCSRMatrix	par_A;
    HYPRE_ParVector	par_B, par_U;

    ni = phgAlloc(map->nlocal * sizeof(*ni));
    offset = map->partition[map->rank];
    for (i = 0; i < map->nlocal; i++)
	ni[i] = i + offset;

    if (sizeof(FLOAT) != sizeof(double))
	v = phgAlloc(map->nlocal * sizeof(*v));

    if (_t->B == NULL) {
	/* copy data to RHS vector */
	if (sizeof(FLOAT) == sizeof(double)) {
	    AddRHSEntries(solver, map->nlocal, ni, (double *)solver->rhs->data);
	}
	else {
	    for (i = 0; i < map->nlocal; i++)
		v[i] = solver->rhs->data[i];
	    AddRHSEntries(solver, map->nlocal, ni, v);
	}
    }

    Assemble(solver);

    /* set up initial solution */
    HYPRE_IJVectorCreate(g_comm, offset, offset + map->nlocal - 1, &_t->U);
    HYPRE_IJVectorSetObjectType(_t->U, HYPRE_PARCSR);
    HYPRE_IJVectorSetMaxOffProcElmts(_t->U, 0);
    HYPRE_IJVectorInitialize(_t->U);
    if (sizeof(FLOAT) == sizeof(double)) {
	HYPRE_IJVectorSetValues(_t->U, map->nlocal, ni, (double *)x->data);
    }
    else {
	for (i = 0; i < map->nlocal; i++)
	    v[i] = x->data[i];
	HYPRE_IJVectorSetValues(_t->U, map->nlocal, ni, v);
    }

    HYPRE_IJMatrixGetObject(_t->A, (void **)(void *)&par_A);
    dump_matrix(par_A, "A");
    HYPRE_IJVectorAssemble(_t->B);
    HYPRE_IJVectorGetObject(_t->B, (void **)(void *)&par_B);
    dump_vector(par_B, "b");
    HYPRE_IJVectorAssemble(_t->U);
    HYPRE_IJVectorGetObject(_t->U, (void **)(void *)&par_U);
    dump_vector(par_U, "x0");

    _t->solver_set_tol(_t->hsolver, solver->rtol);
    _t->solver_set_maxiter(_t->hsolver, solver->maxits);
    if (_t->solver_setup != NULL) {
	_t->solver_setup(_t->hsolver, par_A, par_B, par_U);
	/* only call setup once */
	_t->solver_setup = NULL;
    }
    _t->solver_solve(_t->hsolver, par_A, par_B, par_U);

    if (destroy) {
#if HYPRE_VERSION_MAJOR >= 2
	ams_destroy(solver); 
#endif	/* HYPRE_VERSION_MAJOR >= 2 */
	if (_t->precond != NULL && _t->precon_destroy != NULL) {
	    _t->precon_destroy(_t->precond);
	    _t->precond = NULL;
	    _t->precon_destroy = NULL;
	}
    }

    if (_t->solver_get_nits != NULL)
	_t->solver_get_nits(_t->hsolver, &num_iterations);
    if (_t->solver_get_resid != NULL)
	_t->solver_get_resid(_t->hsolver, &res_norm);

    HYPRE_IJVectorDestroy(_t->B);
    _t->B = NULL;

    if (destroy) { 
	_t->solver_destroy(_t->hsolver);
	_t->hsolver = NULL;
    }

    solver->residual = res_norm;
    solver->nits = num_iterations;

    if (destroy) {
	HYPRE_IJMatrixDestroy(_t->A);
	_t->A = NULL;
    }

    /* copy solution to DOFs */
    if (sizeof(FLOAT) == sizeof(double)) {
	HYPRE_IJVectorGetValues(_t->U, map->nlocal, ni, (double *)x->data);
    }
    else {
	HYPRE_IJVectorGetValues(_t->U, map->nlocal, ni, v);
	for (i = 0; i < map->nlocal; i++)
	    x->data[i] = v[i];
    }
    phgFree(v);
    phgFree(ni);

    HYPRE_IJVectorDestroy(_t->U);
    _t->U = NULL;

    return solver->nits;
}

/*--------------------------------------------------------------------*/

OEM_SOLVER phgSolverHYPRE_ = {
    "HYPRE", RegisterOptions,
    Initialize, Finalize, Create, Destroy, AddMatrixEntries, AddRHSEntries,
    Assemble, Solve
};

#endif		/* USE_HYPRE */
