/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2007 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: solver-phg.c,v 1.9 2007/04/19 13:55:14 tcui Exp $ */

/* The pcg solver.  */

#include "phg.h"
#include "solver-phg.h"
#include <string.h>
#include <math.h>

#define TRUNK_SIZE 8

#define Initialize		NULL
#define Finalize		NULL

typedef struct {
    SOLVER_PC	*pc;
    BOOLEAN	assembled;
} OEM_DATA;

#define _t      ((OEM_DATA *)solver->oem_data)
static int GMRESSteps = 10;

void 
phgSolverGMRESRestartSteps(int Steps)
{
        GMRESSteps = Steps;
}

int
phgSolverJacobi(MAT *A, VEC *b, VEC **px, int max_its, FLOAT rtol,
		FLOAT *res)
{
    VEC *x, *r;
    FLOAT residual = 1.0, rb_norm;
    INT i, nits;

    if (*px == NULL || b == NULL)
	phgError(1, "%s:%d invalid vector handler\n", __FILE__, __LINE__);

    if (A->rows == NULL)
	return 0;

    assert(b->nvec == 1);

    x = *px;
    r = phgMapCreateVec(b->map, 1);

    if (A->diag == NULL)
	phgMatSetupDiagonal(A);

    nits = 0;
    if ((rb_norm = phgVecNorm2(b, 0, NULL)) == 0.0)
	rb_norm = 1.0;
    else
	rb_norm = 1.0 / rb_norm;
    while (residual >= rtol && nits < max_its) {
	phgVecCopy(b, &r);
	phgMatVec(MAT_OP_N, -1.0, A, x, 1.0, &r);
	for (i = 0; i < A->cmap->nlocal; i++)
	    r->data[i] /= A->diag[i];
	nits++;
	residual = phgVecNorm2(r, 0, NULL) * rb_norm;
	phgVecAXPBY(1.0, r, 1.0, &x);
    }			/* end of while-loop */
    phgVecDestroy(&r);
    if (res != NULL)
	*res = residual;

    return nits;
}

void pc_destroy(SOLVER *solver)
{
     phgFree(solver->pc);
}

VEC
*default_pc_proc(SOLVER *solver, VEC *B, VEC *x_1, VEC **u)
{
    FLOAT *old_rhs;
    
    if(B==NULL){
       phgSolverAssemble(solver);      /* assemble matrix and RHS */
       return  NULL;
    }
	
    if (solver->mat->rmap->nlocal != B->map->nlocal){
	phgPrintf("Invalid PC!\n");
	return NULL;
    }

    /* updating RHS */  
    old_rhs = solver->rhs->data;
    solver->rhs->data = B->data;
    /*
    solver->rhs->data=phgAlloc(B->map->nlocal*sizeof(*(B->data)));
    memcpy(solver->rhs->data, B->data, B->map->nlocal * sizeof(*B->data));
    */
 
    /* solving M x = B */
    phgSolverVecSolve(solver,FALSE,*u);
    
    solver->rhs->data = old_rhs;
    return *u;  
}

VEC
*void_pc(SOLVER  *solver, VEC *B, VEC *x_1, VEC **u)
{
    /* no initialization/cleanup needed */
    if (B == NULL)
	return NULL;

    if (*u == NULL)
	phgError(1, "void_pc: invalid vector pointer u.\n");

    phgVecCopy(B, u);

    return *u;
}

VEC
*jacobi_pc(SOLVER *solver, VEC *B, VEC *x_1, VEC **u)
{
    MAT *A;
    VEC *v;
    FLOAT *rhs, *p;
    FLOAT value;
    INT i;

    if(B==NULL){
       return  NULL;
    }

    if (*u == NULL)
	phgError(1, "jacobi_pc: invalid vector pointer u.\n");

#if 0
    /* test the Jacobi solver */
    bzero((*u)->data, B->nlocal * sizeof(*(*u)->data));
    phgSolverJacobi(A, B, u, 1, 0.0, NULL);
    return *u;
#endif

    A = solver->mat;
    if (A->diag == NULL)
	phgMatSetupDiagonal(A);

    v = *u;
    p = v->data;
    rhs = B->data;
    for (i = 0; i < A->rmap->nlocal; i++, rhs++, p++) {
	if ((value = A->diag[i]) == 0.0)
	    phgError(1, "%s:%d, zero diagonal entry.\n", __FILE__, __LINE__);
	*p = *rhs / value;
    }
    p=v->data;
    rhs=B->data;

    return v;
}

void
phgSolverSetPC(SOLVER *solver, SOLVER *pc_solver, PC_PROC  pc_func)
{
   assert(solver != NULL);
   if (solver->pc == NULL)
	solver->pc = phgAlloc(sizeof(SOLVER_PC));
   ((SOLVER_PC *)(solver->pc))->solver = pc_solver;
   ((SOLVER_PC *)(solver->pc))->Solve = pc_func;
   solver->pc_destroy = pc_destroy;
   return;
}

/*---------------------------------------------------------------------------*/

static const char *pc_names[] = {
    "none", "jacobi",
    NULL
};

static const PC_PROC pc_list[] = {
    void_pc, jacobi_pc,
    NULL
};
static int pc_id = 1;	/* default to Jacobi PC */

static int
PCGRegisterOptions(void)
{
    phgOptionsRegisterTitle("\nBuilt-in PCG solver options:", NULL);
    phgOptionsRegisterKeyword("pcg_pc_type", "Preconditioner for built-in PCG",
				pc_names, &pc_id);
    return 0;
}

static int
GMRESRegisterOptions(void)
{
    phgOptionsRegisterTitle("\nBuilt-in GMRES solver options:", NULL);
    phgOptionsRegisterKeyword("gmres_pc_type", "Preconditioner for built-in GMRES",
				pc_names, &pc_id);
    phgOptionsRegisterInt("gmres_restart_step","Restart steps for GMRES solver",
				&GMRESSteps);
    return 0;
}

int
phgSolverPCG__(MAT *A, VEC *b, VEC **x_ptr, SOLVER_PC *pc, int max_its,
	     FLOAT rtol, FLOAT atol, FLOAT *res)
{
    int nits;
    FLOAT residual = 0., rb_norm, tol;
    FLOAT rz_1 = 0.0, rz_2 = 0.0, alpha;
    VEC *r = NULL, *r_1 = NULL, *z = NULL, *Ap = NULL;
    VEC *p = NULL, *p_1 = NULL, *x;

    assert(b->nvec == 1);
    x = *x_ptr;
    nits = 0;
    if (IsZero(rb_norm = phgVecNorm2(b, 0, NULL)))
	rb_norm = 1.0;

    /* initialize preconditioner */
    if (pc != NULL)
	(pc->Solve)(pc->solver, NULL, NULL, NULL);

#if 0
    if (pc == superlu_pc) {
	/*bzero(x->data, x->nlocal * sizeof(*x->data));*/
	r = phgVecCopy(x, NULL);
	z = phgVecCopy(b, NULL);
	while (nits < max_its) {
	    /* TODO: compute (b - a * x) with extended precision */
	    phgMatVec(MAT_OP_N, 1.0, A, r, 0.0, &r_1);
	    phgVecAXPBY(1.0, r_1, -1.0, &z);
	    residual = phgVecNorm2(z, 0, NULL) * rb_norm;
	    phgInfo(1, "iteration %d, residual=%lg\n", nits, (double)residual);
	    if (residual < rb_norm)
		break;
	    superlu_pc(A, z, NULL, &r);
	    nits++;
	    phgVecCopy(x, &r_1);
	    phgVecAXPBY(1.0, r, 1.0, &x);
	    phgVecAXPBY(1.0, x, -1.0, &r_1);
	    if (phgVecNorm1(r_1, 0, NULL) == 0.)
		break;
	}
	phgVecDestroy(&r);
	phgVecDestroy(&r_1);
	phgVecDestroy(&z);
	if (res != NULL)
	    *res = residual;
	/* finalize preconditioner */
	if (pc != NULL)
	    pc(A, NULL, NULL, NULL);
	return nits;
    }
#endif

    r = phgMapCreateVec(x->map, 1);
    r_1 = phgMapCreateVec(x->map, 1);
    z = phgMapCreateVec(x->map, 1);
    Ap = phgMapCreateVec(x->map, 1);
    p = phgMapCreateVec(x->map, 1);
    p_1 = phgMapCreateVec(x->map, 1);

    /* r = b - A * _t_U  */
    if (!IsZero(phgVecNorm1(x, 0, NULL) / A->cmap->nglobal)) {
	phgMatVec(MAT_OP_N, 1.0, A, x, 0.0, &r_1);
	phgVecCopy(r_1, &r);
	phgVecAXPBY(1.0, b, -1.0, &r);
    }
    else {
	phgVecCopy(b, &r);
    }

    residual = phgVecNorm2(r, 0, NULL);
    tol=rtol*rb_norm;
    if(rtol*rb_norm < atol)
       tol=atol;
    while (residual >= tol && nits < max_its) {
	if(pc!=NULL)
	    pc->Solve(pc->solver, r, x, &z);
	else
	    void_pc(NULL, r, x, &z);
	nits++;
	if (nits == 1) {
	    phgVecCopy(z, &p);
	    rz_1 = phgVecDot(r, 0, z, 0, NULL);
	}
	else {
	    rz_2 = rz_1;
	    rz_1 = phgVecDot(r, 0, z, 0, NULL);
	    phgVecCopy(p, &p_1);
	    if (rz_2 == 0.0)
		phgError(1, "%s:%d, the matrix may not be SPD.\n",
						__FILE__, __LINE__);
	    phgVecAXPBY(1.0, z, rz_1 / rz_2, &p);
	}
	phgMatVec(MAT_OP_N, 1.0, A, p, 0.0, &Ap);
	if ((alpha = phgVecDot(p, 0, Ap, 0, NULL)) == 0.0)
	    phgError(1, "%s:%d, the matrix may not be SPD.\n",
						__FILE__, __LINE__);
	alpha = rz_1 / alpha;
	phgVecAXPBY(alpha, p, 1.0, &x);
	phgVecCopy(r, &r_1);
	phgVecAXPBY(-1.0 * alpha, Ap, 1.0, &r);
	residual = phgVecNorm2(r, 0, NULL);
	if (phgVerbosity > 0)
	    phgPrintf("    iter %d, residual %le\n", nits, (double)residual);
    }

    phgVecDestroy(&r);
    phgVecDestroy(&r_1);
    phgVecDestroy(&z);
    phgVecDestroy(&Ap);
    phgVecDestroy(&p_1);
    phgVecDestroy(&p);

    if (res != NULL)
	*res = residual;

    /* finalize preconditioner */
    if (pc != NULL)
	pc->Solve(pc->solver, NULL, NULL, NULL);

    return nits;
}

int
phgSolverGMRES__(MAT *A, VEC *b, VEC **x_ptr, SOLVER_PC *pc, int max_its,
	     FLOAT rtol, FLOAT atol, FLOAT *res)
    /*
     *  for details to the algorithm GMRES(m)-Aya see
     *
     *  R. Barrett, M. Berry, T. F. Chan, J. Demmel, J. Donato, J. Dongarra,
     *  V. Eijkhout, R. Pozo, Ch. Romine, H. van der Vorst:
     *  Templates for the Solution of Linear Systems: Building Blocks
     *  for Iterative Solvers;
     *  SIAM, Philadelphia, 1994
     *
     */
{
    int nits,i,j,k;
    INT nlocal, nglobal;
    FLOAT residual = 0., rb_norm,tol;
    FLOAT tmp,h1,h2;
    FLOAT **H=NULL,*h_data=NULL,*y=NULL,*s=NULL, *c1=NULL,*c2=NULL;
    VEC *r = NULL, *z = NULL, *Av = NULL, *v1 = NULL;
    VEC **v_ptr=NULL, *x;

    assert(b->nvec == 1);
    x = *x_ptr;
    nlocal = x->map->nlocal;
    nglobal = x->map->nglobal;
    /*  allocation of matrix H and Vec v */
    v_ptr=(VEC**)phgAlloc((GMRESSteps+2)*sizeof(*v_ptr));
    H=(FLOAT **)phgAlloc((GMRESSteps+2)*sizeof(*H));
    h_data=(FLOAT*)phgAlloc((GMRESSteps+2)*(GMRESSteps+1)*sizeof(*h_data));
    for(i=0;i<GMRESSteps+2;i++)
    {
    	v_ptr[i]=phgMapCreateVec(x->map, 1);
	H[i]=h_data+i*(GMRESSteps+2);	
    }

    y=(FLOAT *)phgAlloc((GMRESSteps+2)*4*sizeof(*y));
    s=y+GMRESSteps+2;
    c1=y+(GMRESSteps+2)*2;
    c2=y+(GMRESSteps+2)*3;
    z=phgMapCreateVec(x->map, 1);
    Av=phgMapCreateVec(x->map, 1);
     
    /* initialize preconditioner */
    if (pc != NULL)
	pc->Solve(pc->solver, NULL, NULL, NULL);

     /* loop for 'MaxIter' GMRES cycles */
     /* v_ptr[0]= r = b - A * x */
     r=v_ptr[0];
     if (!IsZero(phgVecNorm1(x, 0, NULL) / A->cmap->nlocal)) {
         phgVecCopy(b, &r);
         phgMatVec(MAT_OP_N, -1.0, A, x, 1.0, &r);
     }
     else {
         phgVecCopy(b, &r);
     }

     rb_norm = phgVecNorm2(b, 0, NULL);
     if (IsZero(rb_norm-1e-16))
        rb_norm = 1.0;

     residual = phgVecNorm2(r, 0, NULL);

     tol=rtol*rb_norm;
     if(tol < atol)
	tol = atol;
     nits = 0;
     while (residual >= tol&& nits < max_its) {
	/* M*z = b - A*x */
	if(pc!=NULL)
	   pc->Solve(pc->solver, r, x, &z);
	else
	   void_pc(NULL, r, x, &z);
	
	/* v[1] = z/norm(z) */
	s[1]=phgVecNorm2(z, 0, NULL);
        phgMatVec(MAT_OP_N,1.0/s[1],NULL,z,0.0,&v_ptr[1]);

        /* GMRES iteration */
        i = 0;
	while(i<GMRESSteps&&nits < max_its )
	{
	   i++,nits++;
	   v1=v_ptr[i+1];
	   /* M*v[j+1] = A*v[j] */
	   phgMatVec(MAT_OP_N, 1.0, A, v_ptr[i], 0.0, &Av);
	   if(pc!=NULL)
	      pc->Solve(pc->solver, Av, x, &v1);
	   else
	      void_pc(NULL, Av, x, &v1);

	   /* modified Gram-Schmidt orthogonalization */
	   for(k=1;k<=i;k++){
		/* H[i][k]=<v[i+1],v[k]> */
		H[i][k]=phgVecDot(v1, 0, v_ptr[k], 0, NULL);
		/* v[i+1]=v[i+1]- H[i][k]*v[k] */
		phgMatVec(MAT_OP_N,-H[i][k],NULL,v_ptr[k],1.0,&v1); 
	   } /* end for k-iter */
	   H[i][i+1]= phgVecNorm2(v1, 0, NULL);
	   if(IsZero(H[i][i+1])) {
		phgPrintf("%le happy break",H[i][i+1]);
		i--;nits--;break;
	    }
	   phgMatVec(MAT_OP_N,0.0,NULL,NULL,1.0/H[i][i+1],&v1);

	   /* Q-R algoritm */
	   for (k = 1; k < i; k++) {
 	       h1= c1[k] * H[i][k] + c2[k] * H[i][k+1];
               h2 = - c2[k] * H[i][k] + c1[k] * H[i][k+1];
               H[i][k]=h1;
               H[i][k+1]=h2;
            }
           
	    tmp = Sqrt(H[i][i] * H[i][i] + H[i][i+1] * H[i][i+1]);
	    if(tmp < 1.e-16) tmp=1.e-16;
            c1[i] = H[i][i] / tmp;
            c2[i] = H[i][i+1] /tmp ;
            s[i+1] = - c2[i] * s[i];
            s[i] = c1[i] * s[i];
	    H[i][i] = c1[i]*H[i][i]+c2[i]* H[i][i+1];
    /*        if (Fabs(s[i+1])<rtol*rb_norm||(IsZero(rb_norm) && IsOne(1.0 + s[i+1]))) break; */
	 } /* end-for-while(i) */

	/* solve H * y = s */
        for (j = i; j > 0; j--) {
             y[j] = s[j] / H[j][j];
             for (k = j - 1 ; k > 0; k--)
                s[k] -= H[j][k] * y[j];
        }

	/*  x = x +  V * y   */
	for(j=i;j>0;j--){
	    phgMatVec(MAT_OP_N,y[j],NULL,v_ptr[j],1.0,&x);
	}

        /* r = b - A * x */
        phgVecCopy(b, &r);
        phgMatVec(MAT_OP_N, -1.0, A, x, 1.0, &r);
	residual= phgVecNorm2(r, 0, NULL);

	if (phgVerbosity > 0)
	    phgPrintf("gmres    iter %d, residual %le  rb_norm %le tol %le\n", nits, (double)residual,rb_norm,tol);
    }
    
     if (res != NULL)
       *res = residual;

    phgVecDestroy(&z);
    phgVecDestroy(&Av);

    for(i=0;i<GMRESSteps+2;i++)
	phgVecDestroy(&v_ptr[i]);
    phgFree(v_ptr);
    phgFree(H);
    phgFree(h_data);
    phgFree(y);

    /* finalize preconditioner */
    if (pc != NULL)
	pc->Solve(pc->solver, NULL, NULL, NULL);

    return nits;
}

static int
Create(SOLVER *solver)
{
    phgInfo(2, "Create PCG Solver\n");

    solver->oem_data = phgAlloc(sizeof(OEM_DATA));

    if (solver->pc == NULL)
	phgSolverSetPC(solver,solver,pc_list[pc_id]);

    _t->pc = (SOLVER_PC*) solver->pc;

    if(_t->pc->Solve==NULL){
	if(solver==_t->pc->solver){
          _t->pc->Solve=pc_list[pc_id];
	  phgWarning("Use the %s-pc as PC-Solver's preconditioner \n",pc_names[pc_id]);
	}else{
           _t->pc->Solve=default_pc_proc;
	}
    }
    return 0;
}

static int
Destroy(SOLVER *solver)
{
    if (solver->oem_data == NULL)
	return 0;

    phgInfo(2, "Destroy PCG Solver\n");
    phgFree(solver->oem_data);

    return 0;
}

#define Assemble		NULL
#define AddMatrixEntries	NULL
#define AddRHSEntries		NULL

static int
PCG_Solve(SOLVER *solver, VEC *x, BOOLEAN destroy)
{
    INT nits;
    FLOAT residual = 1.0;
    VEC *B;

    B = phgMapCreateVec(x->map, 0);
    B->data = solver->rhs->data;
    B->nvec = 1;
    nits = phgSolverPCG__(solver->mat, B, &x, _t->pc, solver->maxits,
			  solver->rtol, solver->atol, &residual);
    B->data = NULL;
    if (destroy) {
	phgMatDestroy(&solver->mat);
	solver->sent_to_oem = TRUE;
    }
    phgVecDestroy(&B);

    solver->residual = residual;
    return solver->nits = nits;
}

static int
GMRES_Solve(SOLVER *solver, VEC *x, BOOLEAN destroy)
{
    INT nits;
    FLOAT residual = 1.0;
    VEC *B;

    B = phgMapCreateVec(x->map, 0);
    B->data = solver->rhs->data;
    B->nvec = 1;
    nits = phgSolverGMRES(solver->mat, B, &x, _t->pc, solver->maxits,
			  solver->rtol, &residual);
    B->data = NULL;
    if (destroy) {
	phgMatDestroy(&solver->mat);
	solver->sent_to_oem = TRUE;
    }
    phgVecDestroy(&B);

    solver->residual = residual;
    return solver->nits = nits;
}

/*--------------------------------------------------------------------*/

OEM_SOLVER phgSolverPCG_ = {
    "PCG", PCGRegisterOptions,
    Initialize, Finalize, Create, Destroy, AddMatrixEntries, AddRHSEntries,
    Assemble, PCG_Solve
};

OEM_SOLVER phgSolverGMRES_ = {
    "GMRES", GMRESRegisterOptions,
    Initialize, Finalize, Create, Destroy, AddMatrixEntries, AddRHSEntries,
    Assemble, GMRES_Solve
};
